/* CONFLUEN.C */

/*        Categories database        */
/* M.Fleming, R.Gunther, R.Rosebrugh */
/* Mount Allison University 1994-95  */


#include "confluen.h"


/*
Parameters:
	path1, path2: arrays of integers representing two paths of arrows.

Purpose:
	To determine if there is an overlap between the two paths and to
	return its length; if there is no overlap, returns -1.

Design:
	Compare the beginning of path1 to the end of path2.  If they are
	the same, return the size of the overlap; otherwise return -1.
*/

int overlap(int path1[MAXWORD], int path2[MAXWORD])
{
int i, k;
int size;								/* the length of the smaller of
											path1 and path2				*/
int temp1[MAXWORD], temp2[MAXWORD];		/* temporary variables to store
											portions of path1 and path2 */


if(path_len(path1) > path_len(path2))
	{
	size = path_len(path2);
	}
else
	{
	size = path_len(path1);
	}
for(i=0;i<size;i++)
	{
	copy_path(temp1,path1);
	temp1[size-i] = -1;
	for(k=0;k<size-i;k++)
		{
		temp2[k] = path2[path_len(path2) - (size - k - i)];
		}
	temp2[k] = -1;
	if(equals(temp1,temp2))
		{
		return(size-i);
		}
	}
	return(-1);
}


/*
Parameters:
	w1,w2:	arrays of integers representing paths which are the left and
			right sides of a relation to be added to the category.
	A:      pointer to the current category.

Purpose:
	To add the relation w1 = w2 to the set of relations in the category *A.

Design:
	The number of relations in the category is incremented.  Whichever of
	w1 and w2 has the greater order is stored as the left side of the new
	relation, and the other as the right side.
*/

void add_relation(int w1[MAXWORD], int w2[MAXWORD], category *A)
{
	A->num_rel++;
	A->rel_set = (relation*)realloc(A->rel_set,sizeof(relation) * A->num_rel);
	if (greater(w1,w2))
		{
		copy_path(A->rel_set[A->num_rel-1].lhs,w1);
		copy_path(A->rel_set[A->num_rel-1].rhs,w2);
		}
	else
		{
		copy_path(A->rel_set[A->num_rel-1].lhs,w2);
		copy_path(A->rel_set[A->num_rel-1].rhs,w1);
		}
	printf("\nThe relation %s = ",path_to_string(A->rel_set[A->num_rel-1].lhs,*A));
	printf("%s has been added to the set of relations.\n",path_to_string(A->rel_set[A->num_rel-1].rhs,*A));
}



/*
Parameters:
	rel1, rel2:	two relations from the current category whose left sides
				have an overlap.
	olap_size:  the size of the overlap between rel1 and rel2.
	A:			pointer to the current category.

Purpose:
	To check to see that rel1 and rel2 satisfy the Overlap rule of the
	Knuth-Bendix algorithm and if not, to add a relation to the set
	of relations in the category.  Returns a boolean indicating whether
	or not a relation was added.

Design:
	Construct r1 and r2 as described below in the variable declarations.
	If w1 and w2 (the reduced forms of r1 and r2) are not equal, then
	add the relation w1 = w2 to the category.
*/

bool reduce_overlap(relation rel1, relation rel2, int olap_size, category *A)
{
int r1[MAXWORD];		   			/* arrows in left side of rel1 BEFORE
									   the overlap, plus right side of rel2	*/
int r2[MAXWORD];					/* right side of rel1, plus arrows in
									   left side of rel2 AFTER the overlap	*/
int w1[MAXWORD], w2[MAXWORD];		/* reduced forms of r1 and r2			*/
int temp[MAXWORD];                  /* temporary variable to store path 	*/

copy_path(r2,rel1.rhs);
copy_sub(temp, rel2.lhs, olap_size, path_len(rel2.lhs)-olap_size);
append_path(r2,temp);
copy_sub(r1, rel1.lhs, 0, path_len(rel1.lhs) - olap_size);
append_path(r1, rel2.rhs);
reduce(w1, r1, *A);
reduce(w2, r2, *A);
if (!(equals(w1,w2)))
	{
	add_relation(w1,w2,A);
	return(true);
	}
return(false);
}


/*
Parameters:
	rel1, rel2:	two relations from the current category where the left
				side of rel2 is a subpath of the left side of rel1.
	location:	the location of rel2 inside rel1.
	A:			pointer to the current category.

Purpose:
	To check to see that rel1 and rel2 satisfy the Subpath rule of the
	Knuth-Bendix algorithm and if not, to add a relation to the set of
	relations in the category.  Returns a boolean indicating whether
	or not a relation was added.

Design:
	Construct r1 and r2 as described below in the variable declarations.
	If w1 and w2 (the reduced forms of r1 and r2) are not equal, then
	add the relation w1 = w2 to the category.
*/

bool reduce_subpath(relation rel1, relation rel2, int location, category *A)
{
int r1[MAXWORD];    			/* right side of rel1 */
int r2[MAXWORD];                /* left side of rel1 with left side of rel2
								  (substring) replaced by right side of rel2*/
int w1[MAXWORD], w2[MAXWORD];	/* reduced forms of r1 and r2		*/
int temp[MAXWORD];				/* temporary variable to store path	*/


	copy_path(r1,rel1.rhs);

	copy_sub(r2,rel1.lhs,0,location);
	append_path(r2,rel2.rhs);
	copy_sub(temp,rel1.lhs,location+path_len(rel2.lhs),path_len(rel1.lhs)-(location+path_len(rel2.lhs)));
	append_path(r2,temp);

	reduce(w1,r1,*A);
	reduce(w2,r2,*A);

	if (!equals(w1,w2))
		{
		add_relation(w1,w2,A);
		return(true);
		}
	return(false);
}


/*
Parameter:
	A:		pointer to the current category

Purpose:
	To check all pairs of relations in the category for the Overlap and
	Subpath rules and to add relations to make the set of relations confluent.

Design:
	Using a double for loop, compare all pairs of relations.  Check for
	any overlaps between the left sides of the two relations, and call
	function reduce_overlap to add any relations necessary to make the
	set of relations confluent.  Then, check if the left side of one
	relation is a subpath of the other left side, and call function
	reduce_subpath to add any necessary relations to the category.
*/

void make_confluent(category *A)
{
bool done;							/* set to true when all overlaps and
										subpaths have been checked, and
										the set of relations is confluent */
bool finished;                      /* set to true when all overlaps have
										been checked */
int i,j;
int size;							/* size of an overlap between two paths */
int loc;     						/* location of one path inside another  */
int position;						/* position in one path to begin searching
										for another path as a substring */
int temp1[MAXWORD], temp2[MAXWORD]; /* temporary variables to store paths */


done = false;
while(!done)
	{
	done = true;
	for(i=0;i<A->num_rel;i++)
		{
		for(j=0;j<A->num_rel;j++)
			{
			if((size = overlap(A->rel_set[j].lhs,A->rel_set[i].lhs)) != -1)
				{
				finished = false;
				while(!finished)
					{
					if(reduce_overlap(A->rel_set[i],A->rel_set[j],size,A))
						{
						done = false;
						}
					if(size == 1)
						{
						finished = true;
						}
					else
						{
						size--;
						copy_sub(temp1,A->rel_set[i].lhs,(path_len(A->rel_set[i].lhs) - size),size);
						copy_sub(temp2,A->rel_set[j].lhs,0,size);
						if(!equals(temp1,temp2))
							{
							finished = true;
							}
						}
					}
				}
			position = 0;
			while((loc = search(A->rel_set[i].lhs,A->rel_set[j].lhs,position)) != -1)
				{
				position = loc + 1;
				if(reduce_subpath(A->rel_set[j],A->rel_set[i],loc,A))
					{
					done = false;
					}
				}
			}
		}
	}
	printf("\nThe set of reductions is now confluent...");
}
