/* DAT.C */

/*        Categories database        */
/* M.Fleming, R.Gunther, R.Rosebrugh */
/* Mount Allison University 1994-95  */

#include        <varargs.h>
#include        <stdio.h>
#include        <stdlib.h>
#include        <ctype.h>
#include        <string.h>
#include        "dat.h"
extern int linenum;
char dummyB[MAXWORD];

/*
Parameters:
	None.
Purpose:
	To clear the screen.
Design:
	Clears the screen.
*/
void clearscreen()
{
int i;

	for(i=0;i<25;i++)
		{
		printf("\n");
		}
}



/*
Parameters:
	t:      variable of type data_e indicating the type of error
	c1, c2: characters to be used in display of error message

Purpose:
	To display an error message if an unknown object or arrow has been
	encountered.
*/

void    data_error(data_e t,char c1,char c2)
{
	switch(t)
		{
		case    obj_dom :
				printf("\007unknown domain %c in arrow %c\n",c2,c1);
				break;
		case    obj_cod :
				printf("\007unknown codomain %c in arrow %c\n",c2,c1);
				break;
		case    rel_err :
				printf("\007unknown arrow %c in relation\n",c1);
				break;
		case    fun_lhs :
				fprintf(outf,"\007unknown object %c as domain of object in %c\n",c1,c2);
				break;
		case    fun_rhs :
				fprintf(outf,"\007unknown object %c as codomain of object in %c\n",c1,c2);
				break;
		case    fnc_arr_lhs :
				fprintf(outf,"\007unknown arrow %c as domain of arrow in %c\n",c1,c2);
				break;
		case    fnc_arr_rhs :
				fprintf(outf,"\007unknown arrow %c as codomain of arrow in %c\n",c1,c2);
				break;
		case    X_obj:
				fprintf(outf,"\007incorrect number of objects in %c, expect %d\n",c1,c2);
				break;
		case    X_arr:
				fprintf(outf,"\007incorrect number of arrows in %c, expect %d\n",c1,c2);
				break;
		}
	exit(1);                        /* posible fatal error kill it */
}

/*
Parameters:
	cat_list: a pointer to a category_list
Purpose:
	To initialize a category_list
Design:
	Sets the list head and tail to point to NULL
*/

void    initialize_cat_list(category_list *cat_list)
{
	cat_list->head = NULL;
	cat_list->tail = NULL;
}

/*
Parameters:
	cat_list: a pointer to a category_list
	A: the category to add to the list
Purpose:
	to add a category to a list
Design:
	Adds the category A to the end of the list
*/


void    add_category(category_list *cat_list, category A)
{
struct category_node *temppointer;

	/*makes a category_node out of the category*/

	temppointer = (struct category_node*)malloc(sizeof(struct category_node));
	temppointer->cat = A;
	temppointer->saved = 1;
	temppointer->next = NULL;

	if(cat_list->head == NULL)
		{
		cat_list->head = temppointer;     /*if list is empty sets head and tail to point*/
		cat_list->tail = cat_list->head;  /*to new node*/
		}
	else
		{
		cat_list->tail->next = temppointer;     /*otherwise sets next pointer of last node to point to*/
		cat_list->tail = temppointer;           /*new node and sets tail to point to new node*/
		}
}

/*
Parameters:
	cat_list: a pointer to a category_list
	name: the name of the category to be removed from the list
Purpose:
	To remove the first instance of the category with the given name from the
	list
Design:
	Searches through the list and removes the first category with the given name
*/

void    remove_category(category_list *cat_list, struct category_node *catptr/*char name[MAXWORD]*/)
{
struct  category_node *temp1, *temp2;

	temp1 = cat_list->head;
	temp2 = cat_list->head;
	if(catptr == temp1/*strcmp(temp1->cat.name,name) == 0*/)   /*if first category in list is category to be removed*/
		{
		if(cat_list->tail == cat_list->head)    /*if category to be removed is only category in list*/
			{
			cat_list->tail = NULL;
			}
		cat_list->head = cat_list->head->next;
		free(temp1->cat.rel_set);
		free(temp1);
		}
	else
		{
		while((catptr != temp1) && (temp1->next != NULL))
			{
			temp1 = temp1->next;
			if(catptr == temp1)
				{
				temp2->next = temp1->next;
				if(temp1 == cat_list->tail)
					{
					cat_list->tail = temp2;
					}
				free(temp1->cat.rel_set);
				free(temp1);
				}
			temp2 = temp2->next;
			}
		}
}

/*
Parameters:
	cat_list: a category_list
Purpose:
	To display the names of all the categories in a list
Design:
	Goes through a list and displays the name of each category
*/

void    display_cat_list(category_list cat_list)
{
struct category_node *temppointer;
int     number;

	temppointer = cat_list.head;
	printf("\n\t\t\t\tCURRENT CATEGORIES\n");
	printf("\t\t\t\t------------------\n\n");
	number = 1;
	while(temppointer != NULL)
		{
		printf("\t\t\t\t(%d) %s\n",number,temppointer->cat.name);
		number++;
		temppointer = temppointer->next;
		}
}

/*
Parameters:
	c_list: a category_list
	name: the name of the category node to find
Purpose:
	To find the first category in a list with the given name
Design:
	Returns a pointer to the category_node that contains the first category
	with the given name
*/

struct category_node *find_category_name(category_list c_list, char name[MAXWORD])
{
struct category_node *temppointer;

	temppointer = c_list.head;
	while (temppointer != NULL)
		{
		if (strcmp(temppointer->cat.name,name) == 0)
			{
			return(temppointer);
			}
		temppointer = temppointer->next;
		}
	return(NULL);
}

/*
Parameters:
	c_list: a category_list
	num: the number of the category node in the list to find
Purpose:
	To find the numth category_node in the list
Design:
	Returns a pointer to the numth category_node in the list
*/

struct category_node *find_category_number(category_list c_list, int num)
{
struct category_node *temppointer;
int i;

	temppointer = c_list.head;
	for(i=0;i<num;i++)
		{
		temppointer = temppointer->next;
		}
	return(temppointer);
}

/*
Parameters:
	c_list: a category_list
Purpose:
	To find the number of categories in a list
Design:
	Returns the number of categories in a list
*/

int num_cat(category_list c_list)
{
int i;
struct category_node *temppointer;

	i = 0;
	temppointer = c_list.head;
	while (temppointer != NULL)
		{
		i++;
		temppointer = temppointer->next;
		}
	return(i);
}

/*
Parameter:
	C:      pointer to the current category

Purpose:
	To read in all relations and encode them as arrays of integers.
	Returns the category with the relations stored.
*/

extern category init_rel(category *C)
{
	int num;
	int i,j,k;
	bool gotit;
	char lhs[MAXARR][MAXWORD],rhs[MAXARR][MAXWORD];

	num = readrel(lhs,rhs);
	C->num_rel = num;
	C->rel_set = (relation*)calloc(num,sizeof(relation));
	for (i=0;i<num;i++)
		{
		j = 0;
		while (lhs[i][j]!='\0')                 /* loop relation lhs */
			{
			gotit = false;
			for (k=0;k<C->numarr;k++)
				if (C->arrow[k] == lhs[i][j])   /* know arrow in word */
					{
					C->rel_set[i].lhs[j] = k;               /* encode that arrow */
					gotit = true;
					k = MAXARR;
					}
			if (!gotit)                                             /* unknown arrow in word, give error */
				data_error(rel_err,lhs[i][j],' ');
			j++;
			}
		C->rel_set[i].lhs[j] = -1;                      /* flag end of list */
		j = 0;
		if (rhs[i][j] != '1')                   /* identity */
			while (rhs[i][j]!='\0')    /* loop arrows on rhs */
				{
				gotit = false;
				for (k=0;k<C->numarr;k++)
					if (C->arrow[k] == rhs[i][j])
						{
						C->rel_set[i].rhs[j] = k;
						gotit = true;
						k = MAXARR;
						}
				if (!gotit)
					data_error(rel_err,rhs[i][j],' ');
				j++;
				}
		C->rel_set[i].rhs[j] = -1;
		}
		return(*C);
}


/*
Parameters:
	A:              the current category
	outf:   pointer to the output stream, which may be a file
			or stdout (the screen)

Purpose:
	To print all relations in the category.
*/

void    print_rel(category A,FILE *outf)
{
	int i,j;
	fprintf(outf,"relations\n\t");
	for (i=0;i<A.num_rel-1;i++)
		{
		j = 0;
		while (A.rel_set[i].lhs[j] != -1)
			{
			fprintf(outf,"%c",A.arrow[A.rel_set[i].lhs[j]]);
			j++;
			}
		fprintf(outf," = ");
		j = 0;
		while (A.rel_set[i].rhs[j] != -1)
			{
			fprintf(outf,"%c",A.arrow[A.rel_set[i].rhs[j]]);
			j++;
			}
		if (j == 0)
			fprintf(outf,"1");
		fprintf(outf,",");
		}
	if (A.num_rel != 0)
		{
		j = 0;
		while (A.rel_set[i].lhs[j] != -1)
			{
			fprintf(outf,"%c",A.arrow[A.rel_set[i].lhs[j]]);
			j++;
			}
		fprintf(outf," = ");
		j = 0;
		while (A.rel_set[i].rhs[j] != -1)
			{
			fprintf(outf,"%c",A.arrow[A.rel_set[i].rhs[j]]);
			j++;
			}
		if (j == 0)
			fprintf(outf,"1");
		fprintf(outf,".\n");
		}
	else
		fprintf(outf,".\n");
}


/*
Parameters:
	none

Purpose:
	To read in information about a category from a file and to store the
	category in memory.
*/
extern category init_cat()
{
	category        C;
	bool    gotit;
	char    t1[MAXARR],t2[MAXARR];
	int     i,j;

	linenum = 0;
	strcpy(dummyB,"category");
	readname(C.name,dummyB);
	C.numobj = readobj(C.obj);
	C.numarr = readarr(C.arrow,t1,t2);
	for (i=0;i<C.numarr;i++)
		{
		gotit = false;
		for (j=0;j<C.numobj;j++)                        /* parse the domain */
			if(t1[i] == C.obj[j])
				{
				C.arr[i][0] = j;
				gotit = true;
				j = MAXARR;
				}
		if (!gotit)                                                             /* unknown domain */
			data_error(obj_dom,C.arrow[i],t1[i]);
		gotit = false;
		for (j=0;j<C.numobj;j++)                        /* parse the codomain */
			if(t2[i] == C.obj[j])
				{
				C.arr[i][1] = j;
				gotit = true;
				j = MAXARR;
				}
		if (!gotit)                                                             /* unknown codomain */
			data_error(obj_cod,C.arrow[i],t2[i]);
		}
		C = init_rel(&C);
		swap(&C);       /*make sure left has greater ordering */
		comp_check(C);
	return(C);
}



/* Returns object #i in the category *C */

char    look_obj(category *C,int i)
{
	return(C->obj[i]);
}


/* Returns the name of the category *C */

char    *look_name(category *C)
{
	return(C->name);
}



/* Returns a character array containing arrow #i in the category *C with its
	domain and codomain */

char    *look_arr(category *C,int i,char ret[7])
{
	ret[0] = C->arrow[i];
	ret[1] = ':';
	ret[2] = C->obj[C->arr[i][0]];
	ret[3] = '-';
	ret[4] = '>';
	ret[5] = C->obj[C->arr[i][1]];
	ret[6] = '\0';
	return(ret);
}


/*
Parameters:
	C:              pointer to the current category
	outf:   pointer to the output stream, which may be a file
			or stdout (the screen)

Purpose:
	To print the category's name, objects, arrows and relations
	to the output stream.
*/

void    print_c(category *C,FILE *outf)
{
	int     i;
	char    arr[7];
	fprintf(outf,"category\n\t%s\n",look_name(C));
	fprintf(outf,"objects\n\t");
	for (i=0;i<C->numobj-1;i++)
		fprintf(outf,"%c, ",look_obj(C,i));
	if (C->numobj != 0)
		fprintf(outf,"%c.\n",look_obj(C,i));
	else
		fprintf(outf,".\n");
	fprintf(outf,"arrows\n\t");
	for (i=0;i<C->numarr-1;i++)
		fprintf(outf,"%s, ",look_arr(C,i,arr));
	if (C->numarr != 0)
		fprintf(outf,"%s.\n",look_arr(C,i,arr));
	else
		fprintf(outf,".\n");
	print_rel(*C,outf);
}


/*
Parameters:
	C:      pointer to the current category
	i:      integer representing an object in the category
	a:      array containing all arrows with object i as their domain

Purpose:
	Returns an integer whose value is the number of arrows in the category
	with object i as their domain.  The array a is passed back containing
	all such arrows.
*/

int     all_arr(category *C,int i,int a[MAXARR])
{
	int j,k=0;
	for (j=0;j<C->numarr;j++)
		if (C->arr[j][0] == i)
			{
			a[k] = j;
			k++;
			}
	return(k);
}



/*
Parameters:
	C:      pointer to the current category
	i:      integer representing an object in the category
	a:      array containing all arrows with object i as their domain

Purpose:
	Returns an integer whose value is the number of arrows in the category
	with object i as their codomain.  The array a is passed back containing
	all such arrows.
*/

int     all_arr_cod(category *C,int i,int a[MAXARR])
{
	int j,k=0;
	for (j=0;j<C->numarr;j++)
		if (C->arr[j][1] == i)
			{
			a[k] = j;
			k++;
			}
	return(k);
}

/*
Parameters:
	A:      a pointer to a category
Purpose:
	To go through the set of relations and exchange the left and right side
	of a relation if the right side has a greater order than the left side
Design:
	Changes the set of relations pointed to by A->rel_set so that the left side of
	every relation has a greater order than the rigt side
*/

void    swap(category *A)
{
/*
	temp:   a temporary storage space for a side of a relation
*/

int temp[MAXWORD], i;


	for (i=0;i<=A->num_rel-1;i++)
		{
		if(greater(A->rel_set[i].rhs,A->rel_set[i].lhs))
			{
			copy_path(temp,A->rel_set[i].rhs);
			copy_path(A->rel_set[i].rhs,A->rel_set[i].lhs);
			copy_path(A->rel_set[i].lhs,temp);
			}
		}
}

/*
Parameters:
	A:      a category
Purpose:
	To check that all the relations in a category contain valid left and right
	paths.  This entails checking that the domain and codomain of the left side
	of a relation equal the domain and codomain of the right side of a relation.
	Must also check that in a path the domains and codomains match up.
Design:
	Returns the appropiate error message if there is an error
*/

void comp_check(category A)
{
int i,j;

	for (i=0;i<A.num_rel;i++)
		{
		for (j=0;j<path_len(A.rel_set[i].lhs)-1;j++)
			{
			if ((A.arr[A.rel_set[i].lhs[j]][0]) != (A.arr[A.rel_set[i].lhs[j+1]][1]))
				{
				error(inv_comp,A.arrow[A.rel_set[i].lhs[j]],A.arrow[A.rel_set[i].lhs[j+1]],i+1,"left");
				}
			}
		for (j=0;j<path_len(A.rel_set[i].rhs)-1;j++)
			{
			if ((A.arr[A.rel_set[i].rhs[j]][0]) != (A.arr[A.rel_set[i].rhs[j+1]][1]))
				{
				error(inv_comp,A.arrow[A.rel_set[i].rhs[j]],A.arrow[A.rel_set[i].rhs[j+1]],i+1,"right");
				}
			}

		if ((A.arr[A.rel_set[i].lhs[path_len(A.rel_set[i].lhs)-1]][0]) != (A.arr[A.rel_set[i].rhs[path_len(A.rel_set[i].rhs)-1]][0]) && (A.rel_set[i].rhs[0] != -1))
			{
			error(dom_err,' ',' ',i+1,"");
			}

		if ((A.arr[A.rel_set[i].lhs[0]][1]) != (A.arr[A.rel_set[i].rhs[0]][1])  && (A.rel_set[i].rhs[0] != -1))
			{
			error(cod_err,' ',' ',i+1,"");
			}

		}
}

/*
Parameters:
	e:              an err type which determines the type of error message to display
	c1:     a char which is either empty or the domain of an arrow
	c2:     a char which is either empty or the codomain of an arrow
	n:      an int which is the number of the relation where the error occured
	side:   a string which is either left or right, depending on what side the
			error occured

Purpose:
	To display an error message to the screen and terminate the program

Design:
	Uses switch to display the appropiate error message
*/

void error(err e,char c1,char c2, int n, char side[5])
{
	switch(e)
		{
		case inv_comp :
			printf("\007domain of %c does not equal codomain of %c on %s side of relation # %d\n",c1,c2,side,n);
			break;
		case dom_err :
			printf("\007domain of left path does not equal domain of right path in relation # %d\n",n);
			break;
		case cod_err :
			printf("\007codomain of left path does not equal codomain of right path in relation # %d\n",n);
			break;
		}
	exit(1);
}

/*
Parameters:
	obj:    the name of the object being checked
	A  :    the category

Purpose:
	To determine whether an object is in the category.

Design:
	Returns false if the object is in the category and true if it is not
*/

bool check_obj(char obj ,category A)
{
int i;

	for(i=0;i<A.numobj;i++)
		{
		if(obj == A.obj[i])
			{
			return(false);
			}
		}
	return(true);
}

/*
Parameters:
	arr:    the name of the arrow being checked
	A  :    the category
Purpose:
	To determine whether an arrow is in the category
Design:
	Returns false if the arrow is in the category and true if it is not
*/

bool check_arr(char arr, category A)
{
int i;

	for(i=0;i<A.numarr;i++)
		{
		if(arr == A.arrow[i])
			{
			return(false);
			}
		}
	return(true);
}

/*
Parameters:
	str: the string to check
	A: the current category
Purpose:
	To verify that all arrows in the str are valid arrows of the category A
Design:
	For each arrow of the string checks that it exists in the category A. Returns
	true if this is the case and false otherwise
*/

bool check_string(char str[MAXWORD], category A)
{
int i;

	for (i=0;i<strlen(str);i++)
		{
		if (check_arr(str[i],A))
			{
			printf("\n\007***************ERROR***************\n");
			printf("%c is not an arrow of %s\n",str[i],A.name);
			printf("***********************************\n");
			return(false);
			}
		}
	return(true);
}

/*
Parameters:
	path: the path to be checked
	A: the current category
Purpose:
	To check that the domains and codomains of all the arrows in the path
	match up
Design:
	Checks that all the domains and codomains of the arrows in the path match
	up and returns true if this is the case and false otherwise
*/

bool check_path(int path[MAXWORD], category A)
{
int i,arr,nextarr;

	for (i=0;i<(path_len(path)-1);i++)
		{
		arr = path[i];
		nextarr = path[i+1];
		if ((A.arr[arr][0]) != (A.arr[nextarr][1]))
			{
			printf("\n\007******************************ERROR******************************\n");
			printf(" Invalid Composition. Domain of %c does not equal Codomain of %c\n",A.arrow[arr], A.arrow[nextarr]);
			printf("*****************************************************************\n");
			return(false);
			}
		}
	return(true);
}


/*
Parameters:
	path1:  an array of integers terminated by -1 which is a path
	path2:  an array of integers terminated by -1 which is a path
Purpose:
	To determine if two paths are identical
Design:
	Returns true if path1 is identical to path2 and false otherwise
		note:this function does not reduce them at all
*/

bool    equals(int path1[MAXWORD], int path2[MAXWORD])
{
int i;

if((path_len(path1)) == (path_len(path2)))
	{
	for(i=0;i<path_len(path1);i++)
		{
		if(path1[i] != path2[i])
			{
			return(false);
			}
		}
	return(true);
	}
else
	{
	return(false);
	}
}

/*
Parameters:
	path:   an array of integers terminated by -1 which is a path
Purpose:
	To determine the length of an array of integers terminated by -1
Design:
	Returns the number of integers in the path, does not count -1
*/

int             path_len(int path[MAXWORD])
{
	int length = 0;
	while (path[length] != -1)
		{
		length++;
		}
	return(length);
}

/*
Parameters:
	dest:   an array of integers which will be set to equal source
	source: an array of integers which is copied into dest
Purpose:
	To copy one array of integers to another array of integers
Design:
	Copies source into dest
*/

void    copy_path(int dest[MAXWORD], int source[MAXWORD])
{
int i;

for(i=0;i<=path_len(source);i++)
	{
	dest[i] = source[i];
	}
}

/*
Parameters:
	dest:   a substring of source is copied into dest
	source: a substring of source is copied into dest
	start:  the position in source at which the substring starts
	length: the length of the substring
Purpose:
	To copy a substring of one array into another array
Design:
	Makes dest equal a substring of source
*/

void    copy_sub(int dest[MAXWORD], int source[MAXWORD], int start, int length)
{
int i;

for(i=0;i<length;i++)
	{
	dest[i] = source[start+i];
	}
	dest[i] = -1;
}

/*
Parameters:
	dest:   an array of integers is appended to this array of integers
	source: this array of integers is appended to dest
Purpose:
	To append one array of intgers to another array of integers, all terminated
	by -1
Design:
	Appends source to dest
*/

void append_path(int dest[MAXWORD], int source[MAXWORD])
{
int i, size;

size = path_len(dest);
for(i=0;i<=path_len(source);i++)
	{
	dest[size+i] = source[i];
	}
}

/*
Parameters:
	path1:  an array of integers terminated by -1
	path2:  an array of integers terminated by -1
Purpose:
	To determine if path1 has a greater ordering than path2
Design:
	Returns true if path1 is of greater order than path2
*/

bool    greater(int path1[MAXWORD], int path2[MAXWORD])
{
int j;

	if(path_len(path1) > path_len(path2))
		{
		return(true);
		}
	else
		{
		if ((path_len(path1)) == (path_len(path2)))
			{
			j = 0;
			while  (j < path_len(path1))
				{
				if (path2[j] > path1[j])
					{
					return(true);
					}
				else
					{
					if (path2[j] < path1[j])
						{
						return(false);
						}
					}
				j++;
				}
			}
		}
	return(false);
}

/*
Parameters:
	path:   the path to be converted to a string(function does not change
			 path)
	A:              a category that contains the arrows of the path
Purpose:
	To convert an array of integers, which represent arrows, terminated by -1
	into a path which is a string of characters which are arrows and the string
	is terminated by \0.
Design:
	Returns a string which is the path represented by the integers in the
	integer array
*/

char* path_to_string(int path[MAXWORD], category A)
{
int i;
char temp[MAXWORD];

if(path_len(path) == 0)
	{
	return("1");
	}
else
	{
	for(i=0;i<path_len(path);i++)
		{
		temp[i]= A.arrow[path[i]];
		}
	temp[i] = '\0';
	return(temp);
	}
}

/*
Parameters:
	sub:    an array of integers which is the subpath we are looking for
	target: an array of integers which is being searched for sub
	start:  The place in target where the search is to begin
Purpose:
	To search one path, begining at some starting position, to see if another
	path is a subpath
Design:
	Returns the position where the subpath begins in the path being searched,
	 and -1 if sub is not a subpath of target
*/

int search(int sub[MAXWORD], int target[MAXWORD], int start)
{
int i,j;
bool found;

	found = false;
	for (i=start;i<path_len(target);i++)
		{
		if (sub[0] == target[i])
			{
			found = true;
			for (j=0;j<path_len(sub);j++)
				{
				if (target[i+j] != sub[j])
					{
					found = false;
					}
				}
			}
			if (found)
				{
				break;
				}
		}

	if (found)
		{
		return(i);
		}
	else
		{
		return(-1);
		}
}

/*
Parameters:
	reducedpath:    an array of integers which will be set to equal the
					reduced path
	path:                   the array of integers being reduced
	A       :                       a category
Purpose:
	To reduce a path using the set of relations stored in a category
Design:
	Reduces path and stores its reduced form in reducedpath
*/

void reduce(int reducedpath[MAXWORD], int path[MAXWORD], category A)
{
/*
spot:   the spot where the left side of a relation begins in the path being
		reduced
temp:   a temporary variable used to store the length of a path
diff:   the difference in length between the left side of a relation and the
		right side of the relation
*/

int i,j, spot, temp, diff;

	copy_path(reducedpath,path);
	for (i=0;i<A.num_rel;i++)
		{
		if((spot = search(A.rel_set[i].lhs,reducedpath,0)) != -1)
			{
			for (j=spot;j<spot+path_len(A.rel_set[i].rhs);j++)
				{
				reducedpath[j] = A.rel_set[i].rhs[j - spot];
				}
			temp = path_len(reducedpath);
			diff = path_len(A.rel_set[i].lhs) - path_len(A.rel_set[i].rhs);
			for (j=spot+path_len(A.rel_set[i].lhs);j<temp;j++)
				{
				reducedpath[j - diff] = reducedpath[j];
				}
			reducedpath[j - diff] = -1;
			i = -1;
			}
		}
}

/*
Parameters:
	name:   the name of the object that we want the integer value of
	A       :       the category that contains the object
Purpose:
	To get the integer representation for the character representation of an
	object
Design:
	Returns the integer value of an object, -1 if the object is not in the
	category
*/

int object_number(char name, category A)
{
int i;

for(i=0;i<A.numobj;i++)
	{
	if(A.obj[i] == name)
		{
		return(i);
		}
	}
return(-1);
}

/*
Parameters:
	name:   the name of the arrow that we want the integer value of
	A       :       the category that contains the object
Purpose:
	To get the integer representation for the character representation of an
	arrow
Design:
	Returns the integer value of an arrow, -1 if the arrow is not in the
	category
*/

int arrow_number(char name, category A)
{
int i;

for(i=0;i<A.numarr;i++)
	{
	if(A.arrow[i] == name)
		{
		return(i);
		}
	}
return(-1);
}

/*
Parameters:
	st:     the string that we want to convert to an array of integers terminated
		by -1
	A:      a category that contains the arrows of the string
Purpose:
	To convert a string of characters, where the characters represent arrows, into
	an array of integers, terminated by -1, where the integers represent arrows.
Design:
	Returns an array of integers terminated by -1, and does not change the st.
*/

int* string_to_path(char st[MAXWORD], category A)
{
int i;
int temp[MAXWORD];

if(strlen(st) == 0)
	{
	temp[0] = -1;
	return(temp);
	}
else
	{
	for(i=0;i<strlen(st);i++)
		{
		temp[i]= arrow_number(st[i],A);
		}
	temp[i] = -1;
	return(temp);
	}
}

/*
Parameters:
	A:      a category we wish to make the dual of
Purpose:
	To create the dual of a category
Design:
	Reverses the domain and codomain of all arrows, reverses all of the paths
	that make up the relations, and then makes the necessary swaps to the set
	of relations if some of the right sides now have a greater order than their
	left sides.  Also allows you to save the dual.
*/

void make_dual(category A, category_list c_list)
{
/*
fname:  the name of the file you widh to save the dual under
over, leave:    variables for checking how a person wishes to save their category
temp2, temp3:   variables for storing temporary paths
C:      the new category
*/

char fname[30], over, leave;
FILE *dualfile;
int i,j,temp;
int temp2[MAXWORD], temp3[MAXWORD];
bool done;
category C;

	done = false;
	while(!done)
		{
		printf("(-1 to cancel)\n");
		printf("\nName of file for dual category: ");
		scanf("%30s",fname);
		if (strcmp(fname,"-1") == 0)
			return;
		if((f = fopen(fname,"r")) == NULL)
			{
			fclose(f);
			dualfile = fopen(fname,"w");
			done = true;
			}
		else
			{
			printf("\n\007%s already exists.  Overwrite (Y/N) ? ",fname);
			flush();
			over = getchar();
			if((over == 'y') || (over == 'Y'))
			   {
			   dualfile = fopen(fname,"w");
			   done = true;
			   }
			else
				{
				printf("\nExit to category tools menu (Y/N) ? ");
				flush();
				leave = getchar();
				if ((leave == 'y') || (leave == 'Y'))
					{
					return;
					}
				}
			}
		}

	printf("Enter name of DUAL category>");
	flush();
	gets(C.name);
	C.numobj = A.numobj;
	C.numarr = A.numarr;
	C.num_rel = A.num_rel;
	for(i=0;i<A.numobj;i++)
		{
		C.obj[i] = A.obj[i];
		}
	for(i=0;i<A.numarr;i++)
		{
		C.arrow[i] = A.arrow[i];
		C.arr[i][0] = A.arr[i][1];
		C.arr[i][1] = A.arr[i][0];
		}

	for(i=0;i<A.num_rel;i++)
		{
		for(j=path_len(A.rel_set[i].lhs)-1;j>=0;j--)
			{
			temp2[path_len(A.rel_set[i].lhs)-1-j] = A.rel_set[i].lhs[j];
			}
		temp2[path_len(A.rel_set[i].lhs)] = -1;
		for(j=path_len(A.rel_set[i].rhs)-1;j>=0;j--)
			{
			temp3[path_len(A.rel_set[i].rhs)-1-j] = A.rel_set[i].rhs[j];
			}
		temp3[path_len(A.rel_set[i].rhs)] = -1;

		if(greater(temp2,temp3))
			{
			copy_path(C.rel_set[i].lhs,temp2);
			copy_path(C.rel_set[i].rhs,temp3);
			}
		else
			{
			copy_path(C.rel_set[i].lhs,temp3);
			copy_path(C.rel_set[i].rhs,temp2);
			}
		}

		print_c(&C,dualfile);
		add_category(&c_list,C);
		fclose(dualfile);
}

/*
Parameters:
	obj:    the object to be removed from the category
	A  :    the category which contains the object
Purpose:
	To remove the specified object from the category
Design:
	Removes the object from the category
*/

void remove_obj(int obj, category *A)
{
int i;

	for (i=obj;i<A->numobj - 1;i++)
		{
		A->obj[i] = A->obj[i+1];
		}
	A->numobj--;
	for (i=0;i<A->numarr;i++)                       /* for all arrows in the category */
		{
		if (A->arr[i][0] > obj)                 /* if the domain is an object coming */
			{                                                   /* after the deleted object, then */
			A->arr[i][0]--;                         /* decrease the number representing the domain */
			}
		if (A->arr[i][1] > obj)                 /* same thing for codomain */
			{
			A->arr[i][1]--;
			}
		}
}

/*
Parameters:
	arr:    the arrow to be removed from the category
	A  :    the category which contains the arrow
Purpose:
	To remove the specified arrow from the category
Design:
	Removes the arrow from the category
*/

void remove_arr(int arr, category *A)
{
int i,j;

	for (i=arr;i<A->numarr - 1;i++)
		{
		A->arrow[i] = A->arrow[i+1];
		A->arr[i][0] = A->arr[i+1][0];
		A->arr[i][1] = A->arr[i+1][1];
		}
	A->numarr--;

	for (i=0;i<A->num_rel;i++)
		{
		for (j=0;j<path_len(A->rel_set[i].lhs);j++)
			{
			if (A->rel_set[i].lhs[j] > arr)
				{
				(A->rel_set[i].lhs[j])--;
				}
			}
		for (j=0;j<path_len(A->rel_set[i].rhs);j++)
			{
			if (A->rel_set[i].rhs[j] > arr)
				{
				(A->rel_set[i].rhs[j])--;
				}
			}
		}

}

/*
Parameters:
	rel:    the relation to be removed from the category
	A  :    the category which contains the relation
Purpose:
	To remove the specified relation from the category
Design:
	Removes the relation from the category, and reallocates the storage
	space taken up by the relation
*/

void remove_rel(int rel, category *A)
{
int i;

	for (i=rel;i<A->num_rel - 1;i++)
		{
		A->rel_set[i] = A->rel_set[i+1];
		}
	A->rel_set = (relation*)realloc(A->rel_set,sizeof(relation) * (A->num_rel-1));
	A->num_rel--;
}


/*
Parameters:
	arr:  the arrow being searched for in the path
	path: the path to be searched
Purpose:
	To return true if the given arrow occurs in the given path,
	and false otherwise.
*/

bool arrow_in_path(int arr, int path[MAXWORD])
{
	int count = 0;
	while (path[count] != -1)
		{
		if (path[count] == arr)
			return(true);
		count++;
		}
	return(false);
}



/*
Parameters:
	arr:    the arrow being searched for in the set of relations
	A:              the category which contains the set of relations being searched
Purpose:
	To search the set of relations to see if the given arrow occurs in any of
	the relations
Design:
	Returns true if the arrow does occur in at least one relation, and false
	if it does not occur in any of the relations.
*/

bool arrow_in_rel(int arr, category A)
{
int i,j;

	for (i=0;i<A.num_rel;i++)
		{
		for (j=0;j<path_len(A.rel_set[i].lhs);j++)
			{
			if (A.rel_set[i].lhs[j] == arr)
				{
				return(true);
				}
			}
		for (j=0;j<path_len(A.rel_set[i].rhs);j++)
			{
			if (A.rel_set[i].rhs[j] == arr)
				{
				return(true);
				}
			}
		}
	return(false);
}


cat_functor     init_sf(category A)
{
	cat_functor     X;
	int     num,i,j,k;
	char    lhs[MAXARR];
	int     rhs[MAXARR],r[MAXARR][MAXWORD];
	bool    gotit;

	strcpy(dummyB,"functor");
	readname(X.name,dummyB);
	num = readxobj(X.name,lhs,rhs);
	if (num != A.numobj)
		data_error(X_obj,X.name[0],A.numobj);
	for (i=0;i<num;i++)              /* loop thru' all domains */
		{
		gotit = false;
		for (j=0;j<A.numobj;j++)                /* loop thru' objects of A */
			if (lhs[i] == A.obj[j])         /* correct left hand side */
				{
				X.obj[j] = rhs[i];
				gotit = true;
				j = MAXARR;
				}
		if (!gotit)
			data_error(fun_lhs,lhs[i],X.name[0]);
		}
	num = readxarr(X.name,lhs,r);
	if (num != A.numarr)
		data_error(X_arr,X.name[0],A.numarr);
	for (i = 0;i<A.numarr;i++)                      /* loop thru' A's arrows */
		{
		gotit = false;
		for (j=0;j<num;j++)                             /* loop thru' returned domains */
			if (lhs[j] == A.arrow[i])       /* match */
				{
				for (k=0;k<X.obj[A.arr[i][0]];k++)
					X.arr[i][k] = r[j][k];
				X.arr[i][k] = -1;
				gotit = true;
				j = MAXARR;
				}
		if (!gotit)
			data_error(fnc_arr_lhs,' ',' ');
		}
	return(X);
}

void    print_sf(cat_functor X, FILE *outf, category A)
{
	int i =0,j;
	fprintf(outf,"functor\n\t%s\n",X.name);
	fprintf(outf,"objects\n\t");
	for (i=0;i<A.numobj-1;i++)
		fprintf(outf,"%s(%c) = [%d], ",X.name,look_obj(&A,i),X.obj[i]);
	fprintf(outf,"%s(%c) = [%d].\n",X.name,look_obj(&A,i),X.obj[i]);
	fprintf(outf,"arrows\n\t");
	for (i=0;i<A.numarr-1;i++)
		{
		fprintf(outf,"%s(%c) = (",X.name,A.arrow[i]);
		j = 0;
		while (j<X.obj[A.arr[i][0]] - 1)
			{
			fprintf(outf,"%d,",X.arr[i][j]);
			j++;
			}
		if (X.obj[A.arr[i][0]] != 0)
			fprintf(outf,"%d",X.arr[i][j]);
		fprintf(outf,"), ");
		}
	if (A.numarr != 0)
		{
		fprintf(outf,"%s(%c) = (",X.name,A.arrow[i]);
		j = 0;
		while (j<X.obj[A.arr[i][0]] - 1)
			{
			fprintf(outf,"%d,",X.arr[i][j]);
				j++;
			}
		if (X.obj[A.arr[i][0]] != 0)
			fprintf(outf,"%d",X.arr[i][j]);
		fprintf(outf,").\n");
		}
	else
		fprintf(outf,".\n");
}



void flush()
{
	fflush(stdin);		/* DOS version */
/*	_iob[0]._cnt=0;	*/	/* UNIX version */
}
