/*FUNCTOR.C*/

/*        Categories database        */
/* M.Fleming, R.Gunther, R.Rosebrugh */
/* Mount Allison University 1994-95  */

#include        <string.h>
#include        "functor.h"

/* Global variable */
char dummyC[MAXWORD];

/*
Parameters:
	f_list: a pointer to a functor_list
Purpose:
	To initialize a new functor list
Design:
	Sets the head and tail of the list to point to NULL
*/

void initialize_functor_list(functor_list *f_list)
	{
	f_list->head = NULL;
	f_list->tail = NULL;
	}

/*
Parameters:
	f_list: a pointer to a functor list
	f: a functor_node
Purpose:
	Adds a functor_node to the end of the list
Design:
	Allocates memory for a new functor_node and then adds it to the list
*/

void add_functor(functor_list *f_list, struct functor_node f)
{
struct functor_node *temppointer;

	temppointer = (struct functor_node*)malloc(sizeof(struct functor_node));
	*temppointer = f;
	if(f_list->head == NULL)
		{
		f_list->head = temppointer;     /*if list is empty, point head and tail*/
		f_list->tail = f_list->head;    /*of list to the functor_node being added*/
		}
	else
		{
		f_list->tail->next = temppointer;       /*otherwise, point the next pointer of the*/
		f_list->tail = temppointer;                     /*last node to the new functor_node and set*/
		}                                                                       /*the tail to point to the new functor_node*/
}


/*
Parameters:
	f_list: a pointer to a functor_list
	name: the name of the functor to be removed
Purpose:
	To remove a functor_node from a list
Design:
	Removes a functor_node from a list
*/

void remove_functor(functor_list *f_list, struct functor_node *functorptr)
{
struct  functor_node *temp1, *temp2;

	temp1 = f_list->head;
	temp2 = f_list->head;
	if(functorptr == temp1)     /*if first node in list is one to remove*/
		{
		if(f_list->tail == f_list->head)  /*if it is only node in list, set tail to NULL*/
			{
			f_list->tail = NULL;
			}
		f_list->head = f_list->head->next;      /*set head of list to point to second node of list*/
		free(temp1);
		}
	else
		{
		while((functorptr != temp1) && (temp1->next != NULL))
			{
			temp1 = temp1->next;
			if(functorptr == temp1)
				{
				temp2->next = temp1->next;
				if(temp1 == f_list->tail)         /*as soon as first temp pointer points*/
					{                                                         /*node to be removed, set the next pointer*/
					f_list->tail = temp2;         /*of the node before it to point point to node*/
					}                             /*after one to be removed*/
				free(temp1);
				}
			temp2 = temp2->next;
			}
		}
}

/*
Parameters:
	F:      a functor
	outf: the file to print the functor to
	A: the category which the functor is from
	B: the category which the functor goes to
Purpose:
	To print a functor to a file
Design:
	Prints a functor to a file
*/

void    print_functor(cat_functor F, FILE *outf, category A, category B)
{
	int i =0,j;
	fprintf(outf,"functor\n\t%s\n",F.name);
	fprintf(outf,"objects\n\t");
	for (i=0;i<A.numobj-1;i++)
		fprintf(outf,"%s(%c) = %c, ",F.name,look_obj(&A,i),look_obj(&B,F.obj[i]));
	fprintf(outf,"%s(%c) = %c.\n",F.name,look_obj(&A,i),look_obj(&B,F.obj[i]));
	fprintf(outf,"arrows\n\t");
	for (i=0;i<A.numarr-1;i++)
		{
		fprintf(outf,"%s(%c) = ",F.name,A.arrow[i]);
		j = 0;
		while (F.arr[i][j] != -1)
			{
			fprintf(outf,"%c",B.arrow[F.arr[i][j]]);
			j++;
			}
		if (j==0)
			fprintf(outf,"1");
		fprintf(outf,", ");
		}
	if (A.numarr != 0)
		{
		fprintf(outf,"%s(%c) = ",F.name,A.arrow[i]);
		j = 0;
		while (F.arr[i][j] != -1)
			{
			fprintf(outf,"%c",B.arrow[F.arr[i][j]]);
			j++;
			}
		if (j==0)
			fprintf(outf,"1");
		fprintf(outf,".\n");
		}
	else
		fprintf(outf,".\n");
}

/*
Parameters:
	f_list: a list of functors
Purpose:
	displays the names of all the functors in a list
Design:
	Goes through a list and prints out the name of each functor and returns
	the number of functor_nodes in the list
*/

int display_functor_list(functor_list f_list)
{
struct functor_node *temppointer;
int i;

	temppointer = f_list.head;
	if (temppointer == NULL)
		{
		printf("\007No functors loaded.\n");
		return(0);
		}
	printf("\t\t\t\tList of functors\n");
	printf("\t\t\t\t~~~~~~~~~~~~~~~~\n");
	i = 1;
	while (temppointer != NULL)
		{
		printf("\t\t\t\t(%d)  %s\t",i,temppointer->f.name);
		printf("%s TO ",temppointer->A.name);
		if(temppointer->functor_type == 0)
			{
			printf("%s\n",temppointer->B.name);
			}
		else
			{
			printf("SET\n");
			}
		i++;
		temppointer = temppointer->next;
		}
	return(i);
}

/*
Parameters:
	f_list: a list of functor nodes
	name: the name of the functor being searched for
Purpose:
	To find the functor_node in a list which has a functor with the name that is passed in
Design:
	Returns a pointer the first functor_node in the list which has a functor with the name
	that is passed in
*/


struct functor_node *find_functor_name(functor_list f_list, char name[MAXWORD])
{
struct functor_node *temppointer;

	temppointer = f_list.head;
	while (temppointer != NULL)
		{
		if (strcmp(temppointer->f.name,name) == 0)
			{
			return(temppointer);
			}
		temppointer = temppointer->next;
		}
	return(NULL);
}


/*
Parameters:
	f_list: a list of functor nodes
	num: the number of the functor in the list that is wanted
Purpose:
	To get the numth functor_node in a list
Design:
	Returns a pointer to the numth functor_node in the list
*/


struct functor_node *find_functor_number(functor_list f_list, int num)
{
struct functor_node *temppointer;
int i;

	temppointer = f_list.head;
	for(i=0;i<num;i++)
		{
		temppointer = temppointer->next;
		}
	return(temppointer);
}


/*
Parameters:
	f_list: a list of functor nodes
Purpose:
	To find the number of functor_nodes in a list
Design:
	Returns the number of functor_nodes in a list
*/

int num_functors(functor_list f_list)
{
int i;
struct functor_node *temppointer;

	i = 0;
	temppointer = f_list.head;
	while (temppointer != NULL)
		{
		i++;
		temppointer = temppointer->next;
		}
	return(i);
}

/*
Parameters:
	f: a pointer to a cat_functor
Purpose:
	To get the user to input the name of a functor
Design:
	Stores the name of a functor in the cat_functor that f points to
*/

void    get_functor_name(cat_functor *f)
{
	printf("\nEnter Functor name->");
	flush();
	gets(f->name);
}

/*
Parameters:
	f: a pointer to a cat_functor
	A: the category which the functor is from
	B: the category which the functor is going to
Purpose:
	Gets the user to input what each object of category A goes to under the functor
Design:
	For each object in A the user is prompted to enter which object in B the functor
	takes the object in A to.  The function checks to see that each object the user
	enters is a valid object in B.
*/

void    get_functor_objects(cat_functor *f, category A, category B)
{
char object, x;
int i;

	for(i=0;i<A.numobj;i++)
		{
		clearscreen();
		printf("%s FROM:\n",f->name);
		print_c(&A,stdout);
		printf("\n%s TO:\n",f->name);
		print_c(&B,stdout);
		printf("\n(Object) %c-->",A.obj[i]);
		flush();
		object = getchar();
		if(check_obj(object,B)) /*check to see that object is a valid object of B*/
			{
			printf("\n%c is not an object of %s.\n",object,B.name);
			i--;
			flush();
			getchar();
			}
		else
			{
			f->obj[i] = object_number(object,B);    /*set the value of the subscript which is the*/
			}                                                                               /*object in A to be the object in B that the object*/
		}                                                                                       /*in A goes to*/
}

/*
Parameters:
	path: the path in B that the arrow goes to under the functor
	arrow: the arrow of A that goes to the path under the functor
	f: a category functor
	A: the category that the functor goes from
	B: the category that the functor goes to
Purpose:
	To check that the object which is the domain of the arrow in A goes to the
	object which is the domain of the path in B that the arrow goes to under the functor
Design:
	Returns true if the domains match up, false otherwise
*/

bool    check_domain(int path[MAXWORD], int arrow, cat_functor f, category A, category B)
{
	if(f.obj[A.arr[arrow][0]] != B.arr[path[path_len(path)-1]][0])
		{
		printf("\nERROR...%s has an invalid domain.\n",path_to_string(path,B));
		return(false);
		}
	return(true);
}

/*
Parameters:
	path: the path in B that the arrow goes to under the functor
	arrow: the arrow of A that goes to the path under the functor
	f: a category functor
	A: the category that the functor goes from
	B: the category that the functor goes to
Purpose:
	To check that the object which is the codomain of the arrow in A goes to the
	object which is the codomain of the path in B that the arrow goes to under the functor
Design:
	Returns true if the codomains match up, false otherwise
*/

bool    check_codomain(int path[MAXWORD], int arrow, cat_functor f, category A, category B)
{
	if(f.obj[A.arr[arrow][1]] != B.arr[path[0]][1])
		{
		printf("\nERROR...%s has an invalid codomain.\n",path_to_string(path,B));
		return(false);
		}
	return(true);
}

/*
Parameters:
	f: a category functor
	A: the category that the functor goes from
	B: the category that the functor goes to
Purpose:
	To get the user to input what path of B each generating arrow of category
	A goes to under the functor f
Design:
	Checks to see that each path the user enters is a valid path in B and that
	the domain and codomain of the path correspond to the domain and codomain
	of the arrow under the functor.  Returns true if all the paths are entered
	correctly.
*/

bool get_functor_arrows(cat_functor *f, category A, category B)
{
int i;
char path_string[MAXWORD];
int path[MAXWORD];
int reduced_path[MAXWORD];

	for(i=0;i<A.numarr;i++)
		{
		clearscreen();
		printf("%s FROM:\n",f->name);
		print_c(&A,stdout);
		printf("\n%s TO:\n",f->name);
		print_c(&B,stdout);
		printf("\n{-1 to quit} (Arrow) %c-->",A.arrow[i]);      /*-1 lets the user terminate entry*/
		flush();
		gets(path_string);
		if (strcmp(path_string,"-1") == 0)      /*checks to see if the user has entered -1*/
			{
			return(false);
			}
		if(strcmp(path_string,"1") == 0)
			{
			copy_path(path,string_to_path(path_string,B));
			copy_path(f->arr[i],path);
			}
		else
			{
			if(!(check_string(path_string,B)))      /*checks to see that the string the user enters*/
				{                                                               /*contains only arrows that exist in category B*/
				i--;
				flush();
				getchar();
				}
			else
				{
				copy_path(path,string_to_path(path_string,B));
				if(!(check_path(path,B)))               /*checks to see that domains and*/
					{                                                       /*codomains of each arrow in path match up*/
					i--;
					flush();
					getchar();
					}
				else
					{
					if((!(check_domain(path,i,*f,A,B))) || (!(check_codomain(path,i,*f,A,B))))
						{
						i--;                    /*checks to see that domain and codomain of path*/
						flush();  /*match up with domain and codomain of arrow under*/
						getchar();              /*the functor*/
					}
					else
						{
						reduce(reduced_path,path,B);
						copy_path(f->arr[i],reduced_path);
						}
					}
				}
			}
		}
	return(true);
}

/*
Parameters:
	F: a pointer to a functor_node
Purpose:
	To check that all the relations in the category A hold up under the functor
Design:
	Returns true if the relations are all satisfied and false otherwise
*/

bool check_functor_relations(struct functor_node F)
{
int i,j;
int left_temp[MAXWORD], right_temp[MAXWORD], red1[MAXWORD], red2[MAXWORD];

	for(i=0;i<F.A.num_rel;i++)
		{
		left_temp[0] = right_temp[0] = -1;
		for(j=0;j<path_len(F.A.rel_set[i].lhs);j++)
			{
			append_path(left_temp,F.f.arr[F.A.rel_set[i].lhs[j]]);  /*makes new left and right sides for*/
			}                                                       /*what the original left and right sides*/
		for(j=0;j<path_len(F.A.rel_set[i].rhs);j++)                 /*goes to under the functor*/
			{
			append_path(right_temp,F.f.arr[F.A.rel_set[i].rhs[j]]);
			}
		reduce(red1, left_temp, F.B);
		reduce(red2, right_temp, F.B);          /*checks to see that the new left and right side are equal in B*/
		if ((!(equals(red1, red2))) || ((F.B.arr[left_temp[0]][1] != F.B.arr[right_temp[0]][1]) && (red1[0] != -1)))
			{
			printf("\n\007Relation %s = ",path_to_string(F.A.rel_set[i].lhs,F.A));
			printf("%s does not hold in category ",path_to_string(F.A.rel_set[i].rhs,F.A));
			printf("%s\n",F.B.name);
			flush();
			getchar();
			return(false);
			}
		}
	return(true);
}


/*
Parameters:
	path: the path that you are using on the value
	f: a category functor
	value: the value that you are evaluating
Purpose:
	To get the value that your path gives you when used on value
Design:
	Returns the value
*/

int     function_value(int path[MAXWORD], cat_functor f, int value)
{
int i;

	for(i=(path_len(path)-1);i>=0;i--)
		{
		value = f.arr[path[i]][value-1];
		}
	return(value);
}

/*
Parameters:
	F: functor node
Purpose:
	To check that all the relations hold under your set functor
Design:
	Checks that for each relation in category A the relation holds in SET under
	the functor.  Does this by checking the relation for each element of the domain
	of the left and right hand sides.
*/

bool check_sf_relations(struct functor_node F)
{
int i,j;

	for(i=0;i<F.A.num_rel;i++)
		{
		for(j=1;j<=F.f.obj[F.A.arr[F.A.rel_set[i].lhs[path_len(F.A.rel_set[i].lhs)-1]][0]];j++)
			{
			if((function_value(F.A.rel_set[i].lhs,F.f,j)) != (function_value(F.A.rel_set[i].rhs,F.f,j)))
				{
				printf("\n\007Relation %s = ",path_to_string(F.A.rel_set[i].lhs,F.A));
				printf("%s does not hold in SET.\n",path_to_string(F.A.rel_set[i].rhs,F.A));
				flush();
				getchar();
				return(false);
				}
			}
		}
	return(true);
}

/*
Parameters:
	A: the category that the functor goes from
	B: the category that the functor goes to
Purpose:
	To load a functor from a file.
Design:
	Loads the functor from a file.
*/

cat_functor     init_functor(category A, category B)
{
	cat_functor     F;
	int     num,i,j,k,l;
	bool    gotit;
	char    lhs[MAXARR],rhs[MAXARR],r[MAXARR][MAXWORD];

	strcpy(dummyC,"functor");
	readname(F.name,dummyC);
	num = readfobj(F.name,lhs,rhs);
	if (num != A.numobj)
		data_error(X_obj,F.name[0],A.numobj);
	for (i=0;i<num;i++)
		{
		gotit = false;
		for (j=0;j<A.numobj;j++)                        /* loop thru' all objects of A */
			if (lhs[i] == A.obj[j])                 /* correct left hand side */
				{
				for (k=0;k<B.numobj;k++)   /* loop thru' all objects of B */
					if (B.obj[k] == rhs[i]) /* lookup rhs's number */
						{
						F.obj[j] = k;
						k = MAXARR;
						j = MAXARR;
						gotit = true;
						}
				if (!gotit)                     /* no match on rhs, error message */
					data_error(fun_rhs,rhs[i],F.name[0]);
				}
		if (!gotit)                             /* no match on lhs, error message */
			data_error(fun_lhs,lhs[i],F.name[0]);
		}
	num = readfarr(F.name,lhs,r);
	if (num != A.numarr)
		data_error(X_arr,F.name[0],A.numarr);
	for (j=0;j<num;j++)                                      /* loop thru' returned domains */
		{
		gotit = false;
		for (i = 0;i<A.numarr;i++)               /* loop thru' A's arrows */
			if (lhs[j] == A.arrow[i])        /* match */
				{
				k = 0;
				gotit = true;
				if (r[j][k] != '1')
					{
					while (r[j][k] != '\0')
						{
						gotit = false;
						for (l=0;l<B.numarr;l++)        /* loop thru' B's arrows */
							if (r[j][k] == B.arrow[l])
								{
								F.arr[i][k] = l;
								gotit = true;
								l = MAXARR;
								k++;
								}
						if (!gotit)
							data_error(fnc_arr_rhs,r[j][k],F.name[0]);
						}
					}
				F.arr[i][k] = -1;
				i = MAXARR;
				}
		if (!gotit)
			data_error(fnc_arr_lhs,lhs[j],F.name[0]);
		}
	return(F);
}

/*
Parameters:
	ftype: 0 for a functor to a finite category, 1 for a functor to set.
Purpose:
	To load a functor from disk.
Design:
	Makes a node for the functor you load from disk and returns it.
*/

struct functor_node load_functor(int ftype)
{
bool done;
char choice;
char name[30];
struct functor_node CF;
int i;

	strcpy(CF.f.name,"~");
	done = false;
	while(!done)    /*Get user to enter what type of functor they wish to load*/
		{
		if(ftype == 1)
			{
			CF.functor_type = 1;    /*if a set functor then store 1 for functor type*/
			done = true;
			}
		 if(ftype == 0)
			{
			CF.functor_type = 0;    /*if a functor between finite categories then store 0 for functor type*/
			done = true;
			}
		 }
	clearscreen();
	choice = 'y';
	while((choice == 'y') || (choice == 'Y'))
		{
		flush();          /*Gets the user to enter the name of the functor they wish to load*/
		printf("(-1 to cancel)\n");
		printf("Enter file name for functor you wish to load> ");
		gets(name);
		if(strcmp(name,"-1") == 0)
			return(CF);
		if((f = fopen(name,"r")) == NULL)
			{
			printf("\nError - not able to find file - try again?(y/n)");
			choice = getchar();
			}
		else
			{
			CF.A = init_cat();      /*loads in the category the functor is from into the functor_node*/
			if(CF.functor_type == 0)
				{                               /*if a functor between finite categories then loads in the category*/
				CF.B = init_cat();          /*the functor is going to into the functor_node and loads in the functor*/
				CF.f = init_functor(CF.A,CF.B);
				}
			else
				{
				CF.f = init_sf(CF.A);   /*otherwise just loads in a set functor*/
				}
			CF.saved = 1;
			CF.next = NULL;
			printf("\n                    *****SUCCESS*****\n");
			printf("Hit <return> to go back to main menu...");
			flush();
			gets(name);
			choice = 'n';
			}
		}
		fclose(f);
		return(CF);
}

/*
Parameters:
	F: a functor node
Purpose:
	Saves the functor_node that is passed in to disk
Design:
	Saves the functor_node that is passed in to disk
*/

void save_functor(struct functor_node F)
{
bool done;
char leave, over;
char filename[30];

	done = false;
	while(!done)
		{
		clearscreen();
		printf("\n\nEnter file name for storing functor: ");    /*gets the user to enter the name of the file*/
		flush();                                                                                  /*that they wish to save the category to*/
		gets(filename);
		if((f = fopen(filename,"r")) == NULL)
			{
			fclose(f);
			f = fopen(filename,"w");
			print_c(&(F.A),f);
			if(F.functor_type == 0) /*if it is a functor between finite categories then*/
				{                                       /*save the two categories and the functor to file*/
				fprintf(f,"\n");
				print_c(&(F.B),f);
				fprintf(f,"\n\n");
				print_functor(F.f,f,F.A,F.B);
				}
			else                    /*otherwise, just save one category and the functor to file*/
				{
				fprintf(f,"\n\n");
				print_sf(F.f,f,F.A);
				}
			fclose(f);
			done = true;
			}
		else    /*if the file already exists, ask the user if they wish to overwrite it*/
			{
			printf("\n\007%s already exists.  Overwrite (Y/N) ? ",filename);
			flush();
			over = getchar();
			if((over == 'y') || (over == 'Y'))
				{
				f = fopen(filename,"w");
				print_c(&(F.A),f);
				if(F.functor_type == 0)
					{
					fprintf(f,"\n");
					print_c(&(F.B),f);
					fprintf(f,"\n\n");
					print_functor(F.f,f,F.A,F.B);
					}
				else
					{
					fprintf(f,"\n\n");
					print_sf(F.f,f,F.A);
					}
				fclose(f);
				done = true;
				}
			else
				{
				printf("\nExit to functor menu (Y/N) ? ");
				leave = getchar();
				if ((leave == 'y') || (leave == 'Y'))
					{
					done = true;
					}
				}
			}
		}
}

/*
Parameters:
	F: a category Functor
	A: the category the functor is from
	B: the category the functor goes to
Purpose:
	To make a category_functor node
Design:
	Returns a functor_node
*/

struct functor_node make_functor_node(cat_functor F, category A, category B)
{
struct functor_node temppointer;

	temppointer.f = F;
	temppointer.next = NULL;
	temppointer.saved = 1;
	temppointer.A = A;
	temppointer.B = B;
	return(temppointer);
}

/*
Parameters:
	f: a category functor
	A: the category that the functor goes from
Purpose:
	To get the user to enter what each object of category A goes to in SET
Design:
	Checks to see that each object the user enters is less than or equal to MAXWORD and greater than 0
*/

void get_sf_objects(cat_functor *f, category A)
{
int i, num;
char name[MAXWORD];
bool valid;

	clearscreen();
	print_c(&A,stdout);
	for (i=0;i<A.numobj;i++)
		{
		valid = false;
		while (!valid)
			{
			printf("(Object) %c --> ",A.obj[i]);
			num = atoi(gets(name));
			if ((num >= 0) && (num <= MAXWORD))     /*checks to see that each number the user enters for an*/
				{                                   /*object of SET is >= 0 and less than or equal to MAXWORD*/
				f->obj[i] = num;
				valid = true;
				}
			else
				{
				printf("\007");
				}
			}
		}
}

/*
Parameters:
	f: a category functor
	A: the category that the functor goes from
Purpose:
	Gets the user to enter what each element of the domain of the arrow in SET
	goes to in the codomain of the arrow in SET
Design:
	Returns true if the user enters information for each arrow, false otherwise
*/


bool get_sf_arrows(cat_functor *f, category A)
{
int i,j,num;
char name[MAXWORD];
bool valid;

	for(i=0;i<A.numarr;i++)
		{
		if (f->obj[A.arr[i][0]] != 0)
			{
			clearscreen();
			printf("%s FROM:\n",f->name);
			print_c(&A,stdout);
			printf("%s TO:\n",f->name);
			printf("\tSET\n");
			/*print the domain and codomain of the arrow*/
			printf("\n{-1 to quit} (Arrow) %c : [%d] --> [%d]\n",A.arrow[i],f->obj[A.arr[i][0]],f->obj[A.arr[i][1]]);
			}

		/* print error message if X(codomain) = 0 and X(domain) is not 0 */
		if ((f->obj[A.arr[i][1]] == 0) && (f->obj[A.arr[i][0]] != 0))
			{
			printf("\007\n\nError --- If X(%c) is 0, X(%c) must also be 0.\n",A.obj[A.arr[i][1]],A.obj[A.arr[i][0]]);
			printf("Hit <enter> to continue...");
			flush(); getchar();
			return(false);
			}

		/*for each element of the domain get the user to enter what element it goes to in the codomain*/

		for (j=1;j<=f->obj[A.arr[i][0]];j++)
			{
			valid = false;
			while (!valid)
				{
				printf("\t%d --> ",j);
				num = atoi(gets(name));
				if (num == -1)
					{
					return(false);
					}
				if ((num > 0) && (num <= f->obj[A.arr[i][1]]))  /*check that the elemnt the user enters*/
					{                                           /*is a valid element of the codomain*/
					valid = true;
					f->arr[i][j-1] = num;
					}
				else
					{
					printf("\007");
					}
				}
			}
		}
	return(true);
}

/*
Parameters:
	F: a functor_node
	path: the path the functor is to be used on
	new_path: the new_path that the functor gives you
Purpose:
	To use a functor on a path to get a new path, i.e F(path)
Design:
	performs F(path) on a path, F(path) = new_path
*/

void functor_on_path(struct functor_node F, int path[MAXWORD], int new_path[MAXWORD])
{
int i;

	new_path[0] = -1;
	for(i=0;i<path_len(path);i++)
		{
		append_path(new_path,F.f.arr[path[i]]);
		}
}


int reverse_image_obj(struct functor_node F, int obj, int objarr[MAXWORD])
{
int num,i;

	num = 0;
	for (i=0; i<F.A.numobj; i++)
		{
		if (F.f.obj[i] == obj)
			{
			objarr[num] = i;
			num++;
			}
		}
	return(num);
}


void create_functor(category_list c_list, functor_list *f_list)
{
bool finished;
char name[MAXWORD];
int num;
struct category_node *catptr;
struct functor_node fnode;
cat_functor functor;
category A,B;

	catptr = (struct category_node*)malloc(sizeof(struct category_node));

	clearscreen();
	display_cat_list(c_list);
	finished = false;
	while (!finished)
		{
		printf("(-1 to cancel)\n");
		printf("Functor from> ");
		flush();
		gets(name);
		if(strcmp(name,"-1") == 0)
			{
			return;
			}
		num = atoi(name);
		if ((num >= 1) && (num <= num_cat(c_list)))
			{
			finished = true;
			catptr = find_category_number(c_list,num-1);
			A = catptr->cat;
			}
		}
	finished = false;
	while (!finished)
		{
		printf("Functor to> ");
		flush();
		gets(name);
		if(strcmp(name,"-1") == 0)
			{
			return;
			}
		num = atoi(name);
		if ((num >= 1) && (num <= num_cat(c_list)))
			{
			finished = true;
			catptr = find_category_number(c_list,num-1);
			B = catptr->cat;
			}
		}
	flush();
	get_functor_name(&functor);
	get_functor_objects(&functor,A,B);
	if(get_functor_arrows(&functor,A,B))
		{
		fnode = make_functor_node(functor,A,B);
		fnode.functor_type = 0;
		if(check_functor_relations(fnode))
			{
			add_functor(f_list,fnode);
			f_list->tail->saved = 0;
			}
		}
}



void create_sf(category_list c_list, functor_list *f_list)
{
bool finished;
char name[MAXWORD];
int num;
struct category_node *catptr;
category A,B;
struct functor_node fnode;
cat_functor functor;

	catptr = (struct category_node*)malloc(sizeof(struct category_node));

	clearscreen();
	display_cat_list(c_list);
	finished = false;
	while (!finished)
		{
		printf("(-1 to cancel)\n");
		printf("Functor from> ");
		flush();
		gets(name);
		if(strcmp(name,"-1") == 0)
			{
			return;
			}
		num = atoi(name);
		if ((num >= 1) && (num <= num_cat(c_list)))
			{
			finished = true;
			catptr = find_category_number(c_list,num-1);
			A = catptr->cat;
			}
		}
	flush();
	get_functor_name(&functor);
	get_sf_objects(&functor,A);
	if(get_sf_arrows(&functor,A))
		{
		fnode = make_functor_node(functor,A,B);
		fnode.functor_type = 1;
		if(check_sf_relations(fnode))
			{
			add_functor(f_list,fnode);
			f_list->tail->saved = 0;
			}
		}
}



void remove_func_proc(functor_list *f_list)
{
bool finished;
int num_funcs;
char name[MAXWORD];
int num;
struct functor_node *functorptr;
char ans[MAXWORD];


	functorptr = (struct functor_node*)malloc(sizeof(struct functor_node));

	clearscreen();
	finished = false;
	while (!finished)
		{
		num_funcs = display_functor_list(*f_list);
		if (num_funcs == 0)
			{
			printf("Hit <enter> to continue...");
			flush();
			getchar();
			break;
			}
		else
			{
			printf("(-1 to cancel)\n");
			printf("Functor to remove> ");
			flush();
			gets(name);
			if(strcmp(name,"-1") == 0)
				break;
			num = atoi(name);
			if ((num >= 1) && (num <= num_functors(*f_list)))
				{
				finished = true;
				functorptr = find_functor_number(*f_list,num-1);
				if (functorptr->saved == 0)
					{
					printf("\nSave %s before removing (Y/N)?  ",functorptr->f.name);
					gets(ans);
					if ((ans[0] == 'y') || (ans[0] == 'Y'))
						{
						save_functor(*functorptr);
						}
					}
				remove_functor(f_list, functorptr);
				}
			}
		}
	flush();
}


void save_func_proc(functor_list f_list)
{
bool finished;
char name[MAXWORD];
int num;
int num_funcs;
struct functor_node *functorptr;

	functorptr = (struct functor_node*)malloc(sizeof(struct functor_node));

	clearscreen();
	finished = false;
	while (!finished)
		{
		num_funcs = display_functor_list(f_list);
		if (num_funcs == 0)
			{
			printf("Hit <enter> to continue...");
			flush();
			getchar();
			break;
			}
		else
			{
			printf("(-1 to cancel)\n");
			printf("Functor to save> ");
			flush();
			gets(name);
			if(strcmp(name,"-1") == 0)
				break;
			num = atoi(name);
			if ((num >= 1) && (num <= num_functors(f_list)))
				{
				finished = true;
				functorptr = find_functor_number(f_list,num-1);
				functorptr->saved = 1;
				save_functor(*functorptr);
				}
			}
		}
	flush();
}



void display_func_proc(functor_list f_list)
{
bool finished;
char name[MAXWORD];
int num, num_funcs;
struct functor_node *functorptr;

	functorptr = (struct functor_node*)malloc(sizeof(struct functor_node));

	clearscreen();
	finished = false;
	while (!finished)
		{
		num_funcs = display_functor_list(f_list);
		if (num_funcs == 0)
			{
			printf("Hit <enter> to continue...");
			flush();
			getchar();
			break;
			}
		else
			{
			printf("(-1 to cancel)\n");
			printf("Functor to display> ");
			flush();
			gets(name);
			if (strcmp(name,"-1") == 0)
				break;
			num = atoi(name);
			if ((num >= 1) && (num <= num_functors(f_list)))
				{
				finished = true;
				functorptr = find_functor_number(f_list,num-1);
				clearscreen();
				if(functorptr->functor_type == 0)
					{
					print_functor(functorptr->f,stdout,functorptr->A,functorptr->B);
					}
				else
					{
					print_sf(functorptr->f,stdout,functorptr->A);
					}
				}
			}
		}
	if (num_funcs != 0)
		{
		printf("\nHit <enter> to continue\n");
		flush();
		getchar();
		}
}
