/* INITIAL.C */

/*        Categories database        */
/* M.Fleming, R.Gunther, R.Rosebrugh */
/* Mount Allison University 1994-95  */

#include        "initial.h"


/*
Parameters:
	path:           array of integers containing the path of arrows followed
					so far
	domain:         integer representing the domain of the current arrow in
					the path
	objects:        array of integers containing the objects visited along
					the current path
	object_array:   uses the integers representing the objects as subscripts
					and stores paths to each object as the elements of the
					array
	A:              pointer to the current category
	orig:           a character storing the name of the object being
					tested for an initial object

Purpose:
	A recursive function to determine if object domain is initial.

Design:
	Traces all paths from object domain.  If a path is the first path
	with its codomain, the path is stored in object_array.  Otherwise,
	the path is compared to the path stored for the object which is the
	codomain; if it is not equal to the stored path, the object is not
	initial, a message is displayed, and the function returns false.
	Otherwise, the function returns true.
*/

bool check_initial(int path[], int domain, int objects[], PATHTYPE object_array[MAXARR], category *A, char orig, int maxloop)
{
int i,j, number,current_arrow, codomain;
bool done;
int arrow_array[MAXARR];
int first[MAXWORD], second[MAXWORD];
int loopcount;

done = false;
loopcount = 0;
for(i=0;i<path_len(objects);i++)
	{
	if(objects[i] == domain)
		{
		loopcount++;
		objects[path_len(objects)+1] = -1;
		objects[path_len(objects)] = domain;
		if (loopcount == maxloop)
			{
			done = true;
			i = path_len(objects);
			}
		}
	}
if(!done)
	{
	number = all_arr(A,domain,arrow_array);
	objects[path_len(objects)+1] = -1;
	objects[path_len(objects)] = domain;
	for(i=0;i<number;i++)
		{
		current_arrow = arrow_array[i];
		for((j=path_len(path)+1);j>0;j--)
			{
			path[j] = path[j-1];
			}
		path[0]=current_arrow;
		codomain = A->arr[current_arrow][1];
		if(object_array[codomain][0] == -1)
			{
			copy_path(object_array[codomain],path);
			}
		else
			{
			if(object_array[codomain][0] == -2)
				{
				reduce(first,path,*A);
				if(first[0]!= -1)
					{
					printf("%c NOT INITIAL...%s",orig,path_to_string(path, *A));
					printf(" and 1 are both paths to object %c\n",A->obj[codomain]);
					return(false);
					}
				}
				else
					{
					reduce(first,object_array[codomain],*A);
					reduce(second,path,*A);
					if(!equals(first,second))
						{
						printf("%c NOT INITIAL...%s",orig,path_to_string(path, *A));
						printf(" and %s are both paths to object %c\n",path_to_string(object_array[codomain], *A)
						,A->obj[codomain]);
						return(false);
						}
					}
			}
		if(check_initial(path, codomain, objects, object_array, A, orig, maxloop))
			{
			for(j=1;j<=path_len(path);j++)
				{
				path[j-1] = path[j];
				}
			objects[path_len(objects)-1] = -1;
			}
		else
			{
			return(false);
			}
		}
	}
	return(true);
}


/*
Parameter:
	A:      the current category
Purpose:
	To ask the user for an object in the category and to display a message
	indicating whether or not the object is initial.
Design:
	Asks the user for an object to be tested.  Calls function check_initial
	to determine if the object has more than one path to any other object;
	if so, the object is not initial.  If there is no path to any object
	in the category, the object is not initial and a message is displayed.
	If there is exactly one path to ALL objects, it is displayed that the
	object is initial.
*/

void initial_object(category A, int maxloop)
{
int path[MAXWORD], objects[MAXWORD], objt, i;
PATHTYPE object_array[MAXARR];
char name;
char ch, ob;
int start, end;
int num;

print_c(&A,stdout);
printf("\n\n");
printf("(1) Check all objects.\n");
printf("(2) Check one object.\n\n");
ch = '0';
while (ch != '1' && ch != '2')
	{
	printf("Your choice> ");
	flush();
	ch = getchar();
	}
if (ch == '1')
	{
	start = 0;
	end = A.numobj;
	}
if (ch == '2')
	{
	clearscreen();
	print_c(&A,stdout);
	num = -1;
	printf("\n\n\n");
	while (num == -1)
		{
		printf("Enter name of object> ");
		flush();
		ob = getchar();
		num = object_number(ob,A);
		}
	start = num;
	end = num + 1;
	printf("\n\n");
	}

for (objt=start; objt<end; objt++)
	{
	path[0] = -1;
	objects[0] = -1;
	for(i=0;i<A.numobj;i++)
		{
		object_array[i][0] = -1;
		}
		object_array[objt][0] = -2;
		if(check_initial(path,objt,objects,object_array,&A,A.obj[objt],maxloop))
			{
			for(i=0;i<A.numobj;i++)
				{
				if(object_array[i][0] == -1)
					{
					printf("%c NOT INITIAL...no path to object %c.\n",A.obj[objt],look_obj(&A,i));
					break;
					}
				}
			if(i==A.numobj)
				{
				printf("%c is an initial object!!\n",A.obj[objt]);
				 }
			}
	}
}



