/* LKAN.C */

/*        Categories database        */
/* M.Fleming, R.Gunther, R.Rosebrugh */
/* Mount Allison University 1994-95  */

#include "lkan.h"


/*Global variables*/
coinc_list co_list;
FILE *outfile;
bool to_file;
struct nodetype *e_table[MAXWORD];
struct nodetype *L_table[MAXWORD];
struct nodetype *left_rel[MAXWORD];
struct nodetype *right_rel[MAXWORD];
struct nodetype *left_nat[MAXWORD];
struct nodetype *right_nat[MAXWORD];
struct category_node *catptrA, *catptrB;
struct functor_node *F, *X;


/* left_kan
Parameters:
	c_list:  list of currently loaded categories
	f_list:  list of currently loaded functors
	maxloop: maximum number of times an object can be visited in a path
Purpose:
	This function allows the user to choose the categories and
	functors to be used in computing a Left Kan extension.  It then
	calls the appropriate functions to perform the computation.
*/

void left_kan(category_list c_list,functor_list f_list,int maxloop)
{
int obj, rel, arr;
int arrow_array[MAXARR];
int dom;
int img;
int q;
int least;
int row;
bool done;
char fname[MAXWORD];


	clearscreen();

	/*get user to choose categories A and B*/
	catptrA = get_cat('A',c_list);
	if (catptrA == NULL) return;
	catptrB = get_cat('B',c_list);
	if (catptrB == NULL) return;

	/*get user to choose functor from A to B and make sure that it really is
	  from A to B*/
	F = get_fun(f_list,catptrA,catptrB);
	if (F == NULL) return;

	/*get user to enter functor from A to SET and make sure it really
	is from A to SET*/
	X = get_sf(f_list,catptrA);
	if (X == NULL) return;

	/*get filename for output*/
	printf("\n\n(Hit <enter> if output not to be saved)\n");
	printf("Name of output file:  ");
	flush();
	gets(fname);
	to_file = (fname[0] != '\0');
	if (to_file)
		outfile = fopen(fname,"w");


	clearscreen();

	/* initialize coincidence list to empty */
	co_list.head = co_list.tail = NULL;

	/* initialize all epsilon-tables, L-tables, relation tables and
		naturality tables */
	for (obj=0; obj<F->A.numobj; obj++)
		{
		e_table[obj] = (struct nodetype*)malloc(sizeof(struct nodetype));
		initialize_e_table(X->f.obj[obj], obj);
		}
	for (obj=0; obj<F->B.numobj; obj++)
		{
		L_table[obj] = (struct nodetype*)malloc(sizeof(struct nodetype));
		initialize_L_table(obj, all_arr(&(F->B),obj,arrow_array));
		}
	for (rel=0; rel<F->B.num_rel; rel++)
		{
		left_rel[rel] = (struct nodetype*)malloc(sizeof(struct nodetype));
		right_rel[rel] = (struct nodetype*)malloc(sizeof(struct nodetype));
		initialize_rel_table(rel,path_len(F->B.rel_set[rel].lhs),path_len(F->B.rel_set[rel].rhs));
		}
	for (arr=0; arr<F->A.numarr; arr++)
		{
		left_nat[arr] = (struct nodetype*)malloc(sizeof(struct nodetype));
		right_nat[arr] = (struct nodetype*)malloc(sizeof(struct nodetype));
		dom = F->A.arr[arr][0];
		initialize_nat_tables(path_len(F->f.arr[arr]),X->f.obj[dom],arr);
		}

	/* for each object in category A, define elements in the 
		epsilon table until there are no undefined elements.
		Each time a new element is defined, updates are
		made to L-tables, relation and naturality tables */
	for (obj=0; obj<F->A.numobj; obj++)
		{
		img = F->f.obj[obj];     /* define img as F(obj) */
		row = 1;
		while(undefined(obj))
			{
			least = least_unused(img);
			add_to_e_table(row,least,obj);
			all_arr(&(F->B),img,arrow_array);
			add_L_table_row(least,img,arrow_array);
			update_nat(row,least,obj);
			add_to_all_rel(least,img);
			row++;
			}
		deal_with_coinc();
		}
	printf("Working...\n");
	define_L();   /* define elements in the L-tables */


/* The function renumber() is designed to renumber elements
in all of the sets so that they are consecutive integers.*/

	renumber();

	/* display all tables */
	for(obj=0; obj<F->A.numobj; obj++)
		{
		display_e_table(obj,stdout);
		flush();getchar();
		if (to_file)
			{
			display_e_table(obj,outfile);
			fprintf(outfile,"\n");
			}
		}
	if (to_file)
		fprintf(outfile,"\n\n\n");
	for(obj=0; obj<F->B.numobj; obj++)
		{
		display_L_table(obj,stdout);
		flush();getchar();
		if (to_file)
			{
			display_L_table(obj,outfile);
			fprintf(outfile,"\n");
			}
		}

	if (to_file)
		fclose(outfile);
}



/* initialize_e_table
Parameters:
	XA:   X(A) where A is the object for which the epsilon-table is
			being created.
	obj:  the integer representing the object in category A for which
			the epsilon-table is being created.
Purpose:
	To create an epsilon-table for the object with XA rows.
	The first column contains numbers from 1 to XA.  The second column
	is filled with -1 initially.
*/

void initialize_e_table(int XA, int obj)
{
struct nodetype *temp, *temp2, *temp3;
int i;

	/* if X(obj) is not 0, add a row to the epsilon-table containing
		1 in the first column.  -1 in 2nd column represents undefined */
	if (XA != 0)
		{
		e_table[obj]->data = 1;
		e_table[obj]->down = NULL;
		e_table[obj]->info = obj;
		temp = (struct nodetype*)malloc(sizeof(struct nodetype));
		temp->data = -1;
		temp->down = temp->across = NULL;
		temp->info = F->f.obj[obj];
		e_table[obj]->across = temp;
		}
	else
		{
		e_table[obj] = NULL;
		}

	/* add rows for each element from 2 to XA */
	temp = e_table[obj];
	for (i=2; i<=XA; i++)
		{
		temp2 = (struct nodetype*)malloc(sizeof(struct nodetype));
		temp2->data = i;
		temp2->down = NULL;
		temp2->info = obj;
		temp3 = (struct nodetype*)malloc(sizeof(struct nodetype));
		temp3->data = -1;
		temp3->info = F->f.obj[obj];
		temp3->across = temp3->down = NULL;
		temp2->across = temp3;
		temp->down = temp2;
		temp = temp->down;
		}
}


/* add_to_e_table
Parameters:
	row:  the number to be searched for in the first column of the epsilon-table
	num:  the number to be placed in the second column.
	obj:  the object whose epsilon-table is to be altered.
Purpose:
	To define the element 'num' in the row headed by 'row' in the
	epsilon-table for object 'obj'.
*/

void add_to_e_table(int row, int num, int obj)
{
struct nodetype *temp, *temp2;
int count;

	temp = e_table[obj];
	count = 1;
	while (count < row)     /* keep going down until desired row is found */
		{
		temp = temp->down;
		count++;
		}
		temp->across->data = num;  
}

/* display_e_table
Parameters:
	obj:  the object for which the epsilon-table is to be displayed
	f:    either stdout (the screen) or the output file 
Purpose:
	To display the epsilon-table for object 'obj'
*/

void display_e_table(int obj, FILE *f)
{
struct nodetype *temp, *temp2;

	clearscreen();
	temp = e_table[obj];
	fprintf(f,"  e%c\n",F->A.obj[obj]);
	fprintf(f,"X%c-->L%c\n\n",F->A.obj[obj],F->B.obj[F->f.obj[obj]]);
	while (temp != NULL)
		{
		if (temp->data == -1)
			{fprintf(f," ?");}
		else
			{fprintf(f,"%2d",temp->data);}
		temp2 = temp->across;
		if (temp2 != NULL)
			{
			if (temp2->data == -1)
				{fprintf(f,"    ?");}
			else
				{fprintf(f,"%5d",temp2->data);}
			}
		else
			{
			fprintf(f,"    ?");
			}
		fprintf(f,"\n");
		temp = temp->down;
		}
}


/* undefined
Parameters:
	obj:  the object for which the epsilon-table is to be checked
Purpose:
	To determine if there are any empty entries (-1) in the epsilon-table
	for the given object.  Returns 1 if there are, 0 if the table is complete.
*/

int undefined(int obj)
{
int undef = 0;
struct nodetype *temp;

	temp = e_table[obj];
	while (temp != NULL)
		{
		if (temp->across->data == -1)
			undef = 1;
		temp = temp->down;
		}
	return(undef);
}


/* initialize_L_table
Parameters:
	obj:  the object for which the L-table is to be initialized
	numarr:  the number of arrows out of object 'obj' in category B
Purpose:
	To create an L-table for object 'obj'.  The first node in the tree
	contains the number of arrows out of 'obj'.
*/

void initialize_L_table(int obj, int numarr)
{
	L_table[obj]->data = numarr;
	L_table[obj]->down = L_table[obj]->across = NULL;
	L_table[obj]->info = -1;
}

/* initialize_rel_table
Parameters:
	rel:  integer repr. the relation for which the table is being created.
	numleft:  the number of arrows in the left side of the relation.
	numright:  "     "   "     "    "  "  right  "   "  "      "
Purpose:
	To initialize a relation-table.  left_rel is a tree storing the
	information for the left side of the relation, and right_rel
	for the right side.  The first node in each contains the number
	of arrows in that particular path.
*/

void initialize_rel_table(int rel, int numleft, int numright)
{
	left_rel[rel]->data = numleft;
	left_rel[rel]->down = left_rel[rel]->across = NULL;
	left_rel[rel]->info = -1;
	right_rel[rel]->data = numright;
	right_rel[rel]->down = right_rel[rel]->across = NULL;
	right_rel[rel]->info = -1;
}


/* add_L_table_row
Parameters:
	newnum:  the number which is to head the new row.
	obj:     the object for which the L-table is to be altered.
	arrows:  array of integers containing all arrows out of 'obj'.
Purpose:
	To add a row to the L-table for object 'obj'.  The first column
	stores the element 'newnum' while each other column is initialized
	to -1.  The info portion of the node is used to store the object
	which is the codomain of the arrow.
*/

void add_L_table_row(int newnum, int obj, int arrows[MAXWORD])
{
int numarr, i;
struct nodetype *temp, *temp2;

	numarr = L_table[obj]->data;

	temp = L_table[obj];
	while (temp->down != NULL)
		temp = temp->down;
	temp2 = (struct nodetype*)malloc(sizeof(struct nodetype));
	temp2->data = newnum;
	temp2->down = temp2->across = NULL;
	temp2->info = obj;
	temp->down = temp2;
	temp = temp->down;
	for (i=0; i<numarr; i++)
		{
		temp2 = (struct nodetype*)malloc(sizeof(struct nodetype));
		temp2->data = -1;
		temp2->down = temp2->across = NULL;
		temp2->info = F->B.arr[arrows[i]][1];
		temp->across = temp2;
		temp = temp->across;
		}
}


/* add_rel_table_row
Parameters:
	rel:     integer repr. the relation-table to be altered.
	newnum:  the number to head the new row.
	obj:     the object which is the domain of both left and right sides.
	left:    the left side of the relation (array of integers).
	right:   the right side of the relation.
Purpose:
	To add a row to a relation-table.  The first column of the row
	contains the element 'newnum'.  Each other column is initialized
	to -1.  The info node contains the codomain of the arrow represented
	by that column.
*/

void add_rel_table_row(int rel, int newnum, int obj, int left[MAXWORD], int right[MAXWORD])
{
int numarr, i;
struct nodetype *temp, *temp2;

	/* add to left side first */

	numarr = left_rel[rel]->data;   /* find # of arrows in left path */

	temp = left_rel[rel];
	while (temp->down != NULL)
		temp = temp->down;      /* move down until last row of table */
	temp2 = (struct nodetype*)malloc(sizeof(struct nodetype));
	temp2->data = newnum;
	temp2->down = temp2->across = NULL;
	temp2->info = obj;
	temp->down = temp2;
	temp = temp->down;
	for (i=0; i<numarr; i++)   /* go across row for each arrow in path */
		{
		temp2 = (struct nodetype*)malloc(sizeof(struct nodetype));
		temp2->data = -1;
		temp2->down = temp2->across = NULL;
		temp2->info = F->B.arr[left[i]][1];   /* info field = cod. of current arrow */
		temp->across = temp2;
		temp = temp->across;
		}

	/* now the same on the right side */
	
	numarr = right_rel[rel]->data;

	temp = right_rel[rel];
	while (temp->down != NULL)
		temp = temp->down;
	temp2 = (struct nodetype*)malloc(sizeof(struct nodetype));
	temp2->data = newnum;
	temp2->down = temp2->across = NULL;
	temp2->info = obj;
	temp->down = temp2;
	temp = temp->down;
	for (i=0; i<numarr; i++)
		{
		temp2 = (struct nodetype*)malloc(sizeof(struct nodetype));
		temp2->data = -1;
		temp2->down = temp2->across = NULL;
		temp2->info = F->B.arr[right[i]][1];
		temp->across = temp2;
		temp = temp->across;
		}
}


/* display_L_table
Parameters:
	obj:  the object for which the L-table is to be displayed
	f:    either stdout (the screen) or the output file
Purpose:
	To display the L-table for object 'obj'
*/

void display_L_table(int obj, FILE *f)
{
struct nodetype *temp, *temp2;
int numarr, i;
int arrow_array[MAXARR];
int count;

	numarr = all_arr(&(F->B),obj,arrow_array);
	clearscreen();

	fprintf(f,"       ");
	for (i=0; i<numarr; i++)
		fprintf(f,"  L%c",F->B.arrow[arrow_array[i]]);
	fprintf(f,"\n");
	fprintf(f,"  L%c   ",F->B.obj[obj]);
	for (i=0; i<numarr; i++)
		fprintf(f,"  L%c",F->B.obj[F->B.arr[arrow_array[i]][1]]);
	fprintf(f,"\n");

	count = 0;
	temp = L_table[obj]->down;
	while (temp != NULL)
		{
		temp2 = temp;
		fprintf(f,"\n");
		while (temp2 != NULL)
			{
			if (temp2->data == -1)
				{fprintf(f,"   ?");}
			else
				{fprintf(f,"%4d",temp2->data);}
			if (temp2 == temp) fprintf(f,"   ");
			temp2 = temp2->across;
			}
		temp = temp->down;
		count++;
		}
	fprintf(f,"\n");
}


/* least_unused
Parameters:
	obj:  the object for which the L-table is to be checked.
Purpose:
	To find the next undefined element in L(obj).  Simply returns
	one more than the highest element currently defined.
*/

int least_unused(int obj)
{
int max;
struct nodetype *temp;
int used[MAXWORD];
int count;
int current;
int i;
bool done;

	max = 0;
	temp = L_table[obj]->down;
	while (temp != NULL)
		{
		if (temp->data > max)
			max = temp->data;
		temp = temp->down;
		}
	return(max+1);
}


/* define_L
Purpose:
	To define elements in the L-tables and to fill in all consequences
	in the relation tables and naturality tables.  Also calls
	appropriate functions to deal with any coincidences which arise.
*/

void define_L()
{
int arrow_array[MAXARR], arrow_arr2[MAXWORD];
int i,j,count;
int cod;
int gen;
struct nodetype *temp, *temp2;
bool finished;
int least;
bool done;

finished = false;
while (!finished)
	{
	finished = true;
	for (i=0; i<F->B.numobj; i++)    /* for each object in B */
		{
		all_arr(&(F->B),i,arrow_array);   /* get all arrows out of object i */
		temp = L_table[i]->down;

		/* move down until a row is found without undefined elements */
		if (temp != NULL)
			{
			if (temp->across != NULL)
				{
				done = false;
				while(!done)
					{
					if (temp != NULL)
						{
						if (temp->across->data == -1)
							{done = true;}
						else
							{temp = temp->down;}
						}
					else
						{
						done = true;
						}
					}
				}
			}

		temp2 = temp;
		count = -1;
		while (temp2 != NULL)
			{
			if (temp2->data == -1)
				{
				gen = arrow_array[count];  /* gen is an arrow from array */
				cod = F->B.arr[gen][1];    /* cod is codomain of gen */
				least = least_unused(cod); /* find least unused elt. of L(cod) */
				temp2->data = least;
				temp2->info = cod;
				all_arr(&(F->B),cod,arrow_arr2);
				add_L_table_row(least,cod,arrow_arr2); /* add row to L-table  */
				add_to_all_rel(least,cod);   /* and all appropriate relations */
				finished = false;
				}
			count++;
			temp2 = temp2->across;
			}
		}
	fill_conseq_rel();
	fill_conseq_nat();
	deal_with_coinc();
	}
}


/* display_rel_table
Parameters:
	num:   integer representing the relation-table to be displayed.
	f:     either stdout (the screen) or the output file.
Purpose:
	To display the relation-table for relation 'num'.
*/

void display_rel_table(int num, FILE *f)
{
int leftnum, rightnum, i;
char dom;
struct nodetype *temp1, *temp2, *temp3, *temp4;


	clearscreen();
	leftnum = left_rel[num]->data;
	rightnum = right_rel[num]->data;

	dom = F->B.obj[B.arr[F->B.rel_set[num].lhs[leftnum-1]][0]];

	fprintf(f,"     ");
	for(i=leftnum-1; i>=0; i--)
		fprintf(f,"L%c   ",F->B.arrow[F->B.rel_set[num].lhs[i]]);
	fprintf(f,"          ");
	for (i=rightnum-1; i>=0; i--)
		fprintf(f,"L%c   ",F->B.arrow[F->B.rel_set[num].rhs[i]]);
	fprintf(f,"\n");

	fprintf(f,"   L%c",dom);
	for (i=leftnum-1; i>=0; i--)
		fprintf(f,"-->L%c",F->B.obj[B.arr[F->B.rel_set[num].lhs[i]][1]]);
	fprintf(f,"     ");

	fprintf(f,"   L%c",dom);
	for (i=rightnum-1; i>=0; i--)
		fprintf(f,"-->L%c",F->B.obj[B.arr[F->B.rel_set[num].rhs[i]][1]]);

	fprintf(f,"\n");

	temp1 = left_rel[num]->down;
	temp3 = right_rel[num]->down;

	while (temp1 != NULL)
		{
		temp2 = temp1;
		fprintf(f,"\n");
		while (temp2 != NULL)
			{
			if (temp2->data == -1)
				{fprintf(f,"    ?");}
			else
				{fprintf(f,"%5d",temp2->data);}
			temp2 = temp2->across;
			}
		fprintf(f,"     ");
		temp4 = temp3;
		while (temp4 != NULL)
			{
			if (temp4->data == -1)
				{fprintf(f,"    ?");}
			else
				{fprintf(f,"%5d",temp4->data);}
			temp4 = temp4->across;
			}
		temp1 = temp1->down;
		temp3 = temp3->down;
		}
}



/* initialize_nat_tables
Parameters:
	numarrows:  the number of arrows in the image of arrow 'gen' under F.
	XA:         X(A) where A is the domain of 'gen'.
	gen:        the arrow for which the naturality table is being created.
Purpose:
	To create left and right naturality tables for the arrow 'gen'.
	The table will contain XA rows.  The first column contains numbers
	from 1 to XA.  The left table contains 2 additional columns, while
	the right table has numarrows+1 more.  Each of these columns is
	initialized to -1, with the info portion of the node storing
	the codomain of the arrow represented by that column.
*/

void initialize_nat_tables(int numarrows, int XA, int gen)
{
struct nodetype *temp, *temp2, *temp3;
struct nodetype *nat;
int i,j,k,start;
int cod;
int numarr;

	for (k=0; k<2; k++)
		{
		/* left table has 2 columns, right has numarrows + 1 */
		switch(k)
			{ 
			case 0: nat = left_nat[gen]; numarr = 2; break;
			case 1: nat = right_nat[gen]; numarr = numarrows+1;
			}
		nat->data = numarr;
		nat->down = nat->across = NULL;
		temp = nat;

		/* make a row for each element in XA */
		for (i=1; i<=XA; i++)
			{
			temp2 = (struct nodetype*)malloc(sizeof(struct nodetype));
			temp2->data = i;
			temp2->across = temp2->down = NULL;
			temp2->info = F->A.arr[gen][0];
			temp->down = temp2;

			start = 0;
			if (k == 0)
				{
				temp3 = (struct nodetype*)malloc(sizeof(struct nodetype));
				temp3->data = X->f.arr[gen][i-1];
				temp3->across = temp3->down = NULL;
				temp3->info = F->A.arr[gen][1];
				temp2->across = temp3;
				temp2 = temp2->across;
				start = 1;
				}


			for (j=start; j<numarr; j++)
				{
				if ((k==1) && (j==0)) 
					{
					cod = F->f.obj[F->A.arr[gen][0]];
					}
				else
					{
					if (k==1) cod = F->B.arr[F->f.arr[gen][numarr-start-2]][1];
					if (k==0) cod = F->f.obj[F->A.arr[gen][1]];
					}
				temp3 = (struct nodetype*)malloc(sizeof(struct nodetype));
				temp3->data = -1;
				temp3->across = temp3->down = NULL;
				temp3->info = cod;
				temp2->across = temp3;
				temp2 = temp2->across;
				}

			temp = temp->down;
			}
		}

}


/* display_nat_table
Parameters:
	gen:  the arrow for which the naturality table is to be displayed.
	f:    either stdout (the screen) or the output file.
Purpose:
	To display the naturality table for arrow 'gen' in category A.
*/

void display_nat_table(int gen, FILE *f)
{
char dom, cod;
char domimg, codimg;
int i, temp;
struct nodetype *temp1, *temp2, *temp3, *temp4;

	clearscreen();
	dom = F->A.obj[F->A.arr[gen][0]];
	cod = F->A.obj[F->A.arr[gen][1]];

	fprintf(f,"     X%c",F->A.arrow[gen]);
	fprintf(f,"   e%c",F->A.obj[F->A.arr[gen][1]]);
	fprintf(f,"         ");

	fprintf(f,"   e%c",F->A.obj[F->A.arr[gen][0]]);
	temp = 0;
	while (F->f.arr[gen][temp] != -1)
		{
		fprintf(f,"   L%c",F->B.arrow[F->f.arr[gen][temp]]);
		temp++;
		}
	fprintf(f,"\n");

	codimg = F->B.obj[F->f.obj[F->A.arr[gen][1]]];
	domimg = F->B.obj[F->f.obj[F->A.arr[gen][0]]];
	fprintf(f,"   X%c",dom);
	fprintf(f,"-->X%c",cod);
	fprintf(f,"-->L%c",codimg);

	fprintf(f,"    ");
	fprintf(f,"   X%c",dom);
	fprintf(f,"-->L%c",domimg);
	for (i=path_len(F->f.arr[gen]) - 2; i>=0; i--)
		fprintf(f,"-->L%c",F->B.obj[F->B.arr[F->f.arr[gen][i]][1]]);
	if (F->f.arr[gen][0] != -1)
		fprintf(f,"-->L%c\n",codimg);


	temp1 = left_nat[gen]->down;
	temp3 = right_nat[gen]->down;

	while (temp1 != NULL)
		{
		temp2 = temp1;
		fprintf(f,"\n");
		while (temp2 != NULL)
			{
			if (temp2->data == -1)
				{fprintf(f,"    ?");}
			else
				{fprintf(f,"%5d",temp2->data);}
			temp2 = temp2->across;
			}
		fprintf(f,"    ");
		temp4 = temp3;
		while (temp4 != NULL)
			{
			if (temp4->data == -1)
				{fprintf(f,"    ?");}
			else
				{fprintf(f,"%5d",temp4->data);}
			temp4 = temp4->across;
			}
		temp1 = temp1->down;
		temp3 = temp3->down;
		}
}


/* add_to_all_rel
Parameters:
	num:    the number which is to head the new row.
	obj:    the object being searched for as the domain.
Purpose:
	To add a row to all relation tables which have 'obj' as their domain.
*/

void add_to_all_rel(int num, int obj)
{
int q;
int lrel[MAXWORD], rrel[MAXWORD];
int lastarr;

	for (q=0; q<F->B.num_rel; q++)
		{
		lastarr = path_len(F->B.rel_set[q].lhs) - 1;
		if (F->B.arr[F->B.rel_set[q].lhs[lastarr]][0] == obj)
			{
			reverse(F->B.rel_set[q].lhs,lrel);
			reverse(F->B.rel_set[q].rhs,rrel);
			add_rel_table_row(q,num,obj,lrel,rrel);
			}
		}
}


/* update_nat
Parameters:
	preim: \  the information epsilon(preim) = img has just been
	img:   /  added to the epsilon-table for object 'obj'
	obj:    the object being searched for.
Purpose:
	To go through all naturality tables and to update any which are
	affected by the information epsilon(preim) = img for object 'obj'.
*/

void update_nat(int preim, int img, int obj)
{
int i,dom,cod;
struct nodetype *temp, *temp2;

	for (i=0; i<F->A.numarr; i++)    /* for each arrow in A */
		{
		dom = F->A.arr[i][0];
		cod = F->A.arr[i][1];

		if (cod == obj)      /* if obj passed in is codomain of current arrow... */
			{            /* update the third column of appropriate row */
			temp = left_nat[i]->down;
			while (temp != NULL)
				{
				temp2 = temp->across;
				if (temp2->data == preim)
					{
					temp2->across->data = img;
					}
				temp = temp->down;
				}
			}

		if (dom == obj)    /* if obj passed in is domain of current arrow... */
			{          /* update second column */
			temp = right_nat[i]->down;
			while (temp != NULL)
				{
				if (temp->data == preim)
					{
					temp->across->data = img;
					}
				temp = temp->down;
				}
			}
		check_coinc(left_nat[i],right_nat[i]);  /* check for coincidences */
		}
}



/* check_coinc
Parameters:
	left:   the left side of the table being checked
	right:  the right side....
Purpose:
	To check a table to see if a coincidence occurs.  This happens
	if the rightmost node in the left and right sides are not equal.
	If a coincidence occurs, it is added to the coincidence list.
*/

void check_coinc(struct nodetype *left, struct nodetype *right)
{
int leftnum, rightnum;
struct nodetype *temp1, *temp2, *temp3, *temp4;
int i;

	leftnum = left->data;      /* find number of columns in each */
	rightnum = right->data;    /* side of table */

	temp1 = left->down;
	temp2 = right->down;

	while (temp1 != NULL)
		{
		temp3 = temp1;
		temp4 = temp2;

		/* move to last column of each table */
		for (i=1; i<=leftnum; i++)
			temp3 = temp3->across;
		for (i=1; i<=rightnum; i++)
			temp4 = temp4->across;

		/* if the two values are both defined and are different, */
		/* add them to the coincidence table */
		if ((temp3->data != temp4->data) && (temp3->data != -1) && (temp4->data != -1))
			add_coinc(temp3->data,temp4->data,temp3->info);
		temp1 = temp1->down;
		temp2 = temp2->down;
		}

}


/* add_coinc
Parameters:
	left, right:  the two numbers which form the coincidence.
	obj:  the object for which the coincidence occurs in L(obj).
Purpose:
	To add the coincidence left-right(obj) to the coincidence list.
*/

void add_coinc(int left,int right,int obj)
{
struct coinc_node *temp;
int dummy;

	if (left > right)
		{
		dummy = left;
		left = right;
		right = dummy;
		}

	if (!already_coinc(left,right,obj))
		{
		temp = (struct coinc_node*)malloc(sizeof(struct coinc_node));
		temp->x = left;
		temp->y = right;
		temp->obj = obj;
		temp->next = NULL;
		if (co_list.head == NULL)  /* if list was empty */
			{
			co_list.head = co_list.tail = temp;
			}
		else
			{     /* add to end of list */
			co_list.tail->next = temp;
			co_list.tail = temp;
			}
		}
}


/* reverse
Parameters:
	rel:  a path of arrows
	newrel:  the reverse of that path
Purpose:
	Returns a path which is the reverse of the input path (terminated by -1)
	e.g.  1,2,0,1,3,-1   becomes  3,1,0,2,1,-1.
*/

void reverse(int rel[MAXWORD], int newrel[MAXWORD])
{
int len, i;

	len = path_len(rel);
	for (i=0; i<len; i++)
		newrel[i] = rel[len-1-i];
	newrel[i] = -1;
}


/* fill_conseq_rel
Purpose:
	Goes through all relation tables and fills in any spaces for which
	the information is available in the L-tables.
*/

void fill_conseq_rel()
{
int i,j,q,dom,img;
int arrow_array[MAXARR];
struct nodetype *temp1, *temp2, *temp3, *temp4, *temp5;
int leftpath[MAXWORD], rightpath[MAXWORD];

	for (i=0; i<F->B.num_rel; i++)   /* for each relation in B */
		{
		reverse(F->B.rel_set[i].lhs,leftpath);
		reverse(F->B.rel_set[i].rhs,rightpath);
		temp1 = left_rel[i]->down;
		temp2 = right_rel[i]->down;
		while (temp1 != NULL)
			{
			temp3 = temp1;
			for (j=0; j<path_len(leftpath); j++)  /* each arr. in leftpath */
				{
				dom = F->B.arr[leftpath[j]][0]; /* domain */
				temp4 = L_table[dom]->down;
				if (temp3->data != -1)
					{
					/* find correct row of L-table */
					while (temp4->data != temp3->data)
						temp4 = temp4->down;
					temp5 = temp4->across;
					all_arr(&(F->B),dom,arrow_array);
					q=0;
					
					/* go across to correct arrow in L-table */
					while(arrow_array[q] != leftpath[j])
						{
						temp5 = temp5->across;
						q++;
						}

					/* get info from L-table and */
					/* place it in relation table */
					img = temp5->data;
					temp3->across->data = img;
					}
				temp3 = temp3->across;
				}

			/* same thing for right side */
			temp3 = temp2;
			for (j=0; j<path_len(rightpath); j++)
				{
				dom = F->B.arr[rightpath[j]][0];
				temp4 = L_table[dom]->down;
				if (temp3->data != -1)
					{
					while (temp4->data != temp3->data)
						temp4 = temp4->down;
					temp5 = temp4->across;
					all_arr(&(F->B),dom,arrow_array);
					q=0;
					while(arrow_array[q] != rightpath[j])
						{
						temp5 = temp5->across;
						q++;
						}
					img = temp5->data;
					temp3->across->data = img;
					}
				temp3 = temp3->across;
				}
			temp1 = temp1->down;
			temp2 = temp2->down;
			}
		check_coinc(left_rel[i],right_rel[i]);  /* check for coincidences */
		}
}


/* fill_conseq_nat
Purpose:
	Goes through all naturality tables and fills in any spaces for which
	the information is available in the L-tables.
*/

void fill_conseq_nat()
{
int i,j,q,dom,img;
int arrow_array[MAXARR];
struct nodetype *temp1, *temp2, *temp3, *temp4;
int path[MAXWORD];

	for (i=0; i<F->A.numarr; i++)   /* for each arrow in A */
		{
		reverse(F->f.arr[i],path);
		temp1 = right_nat[i]->down;
		while (temp1 != NULL)
			{
			temp2 = temp1->across;
			for (j=0; j<path_len(path); j++)   /* each arr. in path */
				{
				dom = F->B.arr[path[j]][0]; /* domain */
				temp3 = L_table[dom]->down;
				if (temp2->data != -1)
					{
					/* find correct row of L-table */
					while (temp3->data != temp2->data)
						{
						temp3 = temp3->down;
						}
					temp4 = temp3->across;
					all_arr(&(F->B),dom,arrow_array);
					q=0;

					/* go across to correct arrow in L-table */
					while(arrow_array[q] != path[j])
						{
						temp4 = temp4->across;
						q++;
						}

					/* get info from L-table and place it */
					/* in naturality table */
					img = temp4->data;
					temp2->across->data = img;
					}
				temp2 = temp2->across;
				}
			temp1 = temp1->down;
			}
		check_coinc(left_nat[i], right_nat[i]);  /* check for coincidences */
		}
}


/* deal_with_coinc
Purpose:
	Takes the first coincidence in the coincidence list and updates
	all tables accordingly.  check_coinc_2() adds to the list any
	further coincidences which are created.  update_e() and update_L()
	replace y with x in all appropriate epsilon-tables and L-tables,
	and fill in all consequences in the relation and naturality     tables.
	The rows containing y are then deleted from the L-table for object
	'obj' and for all relation tables with 'obj' as domain.
*/

void deal_with_coinc()
{
struct coinc_node *temp;
int x,y,obj;

	if (co_list.head == NULL)
		return;
	while (num_coincs() != 0)
		{
		temp = co_list.head;   /* deal with first coinc. in list */
		x = temp->x;
		y = temp->y;
		obj = temp->obj;
		check_coinc_2(obj,x,y);  /* check if any new coincs are added
						by dealing with this one */
		update_e(x,y,obj);      /* update epsilon-tables */
		update_L(x,y,obj);      /* update L-tables */
		update_coinc_list(x,y,obj);     /* update coincidence list */
		delete_row(obj,y);      /* delete rows from L-tables and */
		delete_rel_row(y,obj);  /* relation tables */
		}
}


/* update_coinc_list
Parameters:
	x,y:  the two integers forming the coincidence
	obj:  the object for which the coincidence has occurred.
Purpose:
	To replace y by x in all subsequent coincidences in the list,
	to remove any redundancies occurring in the list, and to delete
	the coincidence x-y(obj) from the list.
*/

void update_coinc_list(int x, int y, int obj)
{
/* replace all occurrences of y in the coincidence list with x */
struct coinc_node *temp, *temp2;
int dummy;

	/*display_coincs();
	flush(); getchar();*/

	temp = co_list.head->next;   /* temp is second coinc. in list */
	temp2 = co_list.head;        /* temp2 is a trailing pointer */

	while (temp != NULL)
		{
		/* delete coinc. if identical to one being dealt with */
		if ((temp->x == x) && (temp->y == y) && (temp->obj == obj))
			{
			temp2->next = temp->next;
			if (temp == co_list.tail)
				co_list.tail = temp2;
			}

		/* replace y with x in coincidence and sort if necessary */
		if ((temp->x == y) && (temp->obj == obj))
			temp->x = x;
		if ((temp->y == y) && (temp->obj == obj))
			{
			temp->y = x;
			if (temp->x > temp->y)
				{
				dummy = temp->x;
				temp->x = temp->y;
				temp->y = dummy;
				}
			}
		temp = temp->next;
		temp2 = temp2->next;
		}

	remove_redundancies();

	/* delete first coincidence from list now that it has been dealt with */
	if ((co_list.head->x == x) && (co_list.head->y == y) && (co_list.head->obj == obj))
		{
		if (co_list.tail == co_list.head)
			co_list.tail = NULL;
		co_list.head = co_list.head->next;
		}

	/*display_coincs();
	flush(); getchar();*/
}



/* display_coincs
Purpose:
	To display a list of all coincidences.
*/

void display_coincs()
{
struct coinc_node *temp;

	printf("Coincidences... ");
	temp = co_list.head;
	while (temp != NULL)
		{
		printf("%d-%d(%d)  ",temp->x,temp->y,temp->obj);
		temp = temp->next;
		}
	printf("\n");
}



/* delete_row
Purpose:
	Deletes the row beginning with the integer 'row' from the L-table
	for object 'obj'.
*/

void delete_row(int obj, int row)
{
struct nodetype *temp, *temp2;

	temp = L_table[obj];
	while (temp->down->data != row)
		{
		temp = temp->down;
		}
	temp2 = temp->down;
	temp->down = temp->down->down;
	free_row(temp2);
	free(temp2);
}


/* check_coinc_2
Paramters:
	obj:  object for which a coincidence has occurred.
	x,y:  the two integers forming the coincidence.
Purpose:
	Given a coincidence x-y(obj), this function checks to see if any
	further coincidences are created by identifying x with y in the
	L-table for 'obj'.
*/

void check_coinc_2(int obj, int x, int y)
{
struct nodetype *temp, *temp2;

	temp = temp2 = L_table[obj]->down;
	while (temp->data != x)
		temp = temp->down;      /* go to row containing x */
	while (temp2->data != y)
		temp2 = temp2->down;    /* go to row containing y */

	/* go across each row, and if any entries differ in the two
	rows, add a new coincidence to the list */

	temp = temp->across;
	temp2 = temp2->across;
	while (temp!= NULL)
		{
		if ((temp->data != temp2->data) && (temp->data != -1) && (temp2->data != -1))
			add_coinc(temp->data,temp2->data,temp->info);
		temp = temp->across;
		temp2 = temp2->across;
		}

}



/* update_e
Parameters:
	x,y,obj:  information for the coincidence x-y(obj)
Purpose:
	To update any epsilon-tables affected by the coincidence x-y(obj).
	Also calls update_nat to update naturality tables.
*/

void update_e(int x, int y, int obj)
{
int i;
struct nodetype *temp;
int preim;

	for (i=0; i<F->A.numobj; i++)   /* for each object in A */
		{
		temp = e_table[i];
		while (temp != NULL)
			{
			preim = temp->data;
			if ((temp->across->data == y) && (temp->across->info == obj))
				{
				temp->across->data = x;
				update_nat(preim,x,i);  /* update the naturality tables */
				}
			temp = temp->down;
			}
		}
}


/* update_L
Parameters:
	x,y,obj:  information for the coincidence x-y(obj)
Purpose:
	To update and L-tables affected by the coincidence x-y(obj).
	Also calls fill_conseq_rel and fill_conseq_nat to update
	the appropriate relation and naturality tables.
*/

void update_L(int x, int y, int obj)
{
int i;
struct nodetype *temp, *temp2;
int arrow_array[MAXARR];
int gen, count;


	for (i=0; i<F->B.numobj; i++)
		{
		all_arr(&(F->B),i,arrow_array);
		temp = L_table[i]->down;
		while (temp != NULL)
			{
			count = 0;
			temp2 = temp->across;
			while (temp2 != NULL)
				{
				gen = arrow_array[count];
				if ((temp2->data == y) && (temp2->info == obj))
					{
					temp2->data = x;
					}
				temp2 = temp2->across;
				count++;
				}
			temp = temp->down;
			}
		}
	fill_conseq_rel();
	fill_conseq_nat();
}


/* num_coincs
Purpose:
	To search the coincidence list and return the number of coincidences
	which have not yet been dealt with.
*/

int num_coincs()
{
int num;
struct coinc_node *temp;

	temp = co_list.head;
	num = 0;
	while (temp != NULL)
		{
		num++;
		temp = temp->next;
		}
	return(num);
}



/* remove_redundancies
Purpose:
	To remove any redundant information from the coincidence list.
*/

void remove_redundancies()
{
struct coinc_node *temp, *temp2, *temp3;
int x1, x2, y1, y2, obj1, obj2;

	/* temp and temp2 are used as pointers to pairs of coincs in the list */

	temp = co_list.head;
	while (temp != NULL)
		{
		x1 = temp->x;
		y1 = temp->y;
		obj1 = temp->obj;
		temp2 = temp->next;
		temp3 = temp;
		while (temp2 != NULL)
			{
			x2 = temp2->x;
			y2 = temp2->y;
			obj2 = temp2->obj;

			/* if second coincidence is redundant (e.g. 3-3)
			or if identical to the first one, remove it */
			if ((x2 == y2) || ((x1 == x2) && (y1 == y2) && (obj1 == obj2)))
				{
				temp3->next = temp2->next;
				if (temp2 == co_list.tail)
					co_list.tail = temp3;
				}
			else
				{
				temp3 = temp3->next;
				}
			temp2 = temp2->next;
			}
		temp = temp->next;
		}

	/* remove first coincidence if x = y */
	if (co_list.head->x == co_list.head->y)
		{
		if (co_list.head == co_list.tail)
			co_list.tail = NULL;
		co_list.head = co_list.head->next;
		}
}


/* delete_rel_row
Parameters:
	row:  the integer heading the row to be deleted.
	obj:  the object for which that element has been deleted.
Purpose:
	To delete the row headed by the integer 'row' from all relations
	which have 'obj' as their domain.
*/

void delete_rel_row(int row, int obj)
{
int i,dom,leftside[MAXWORD];
struct nodetype *temp, *temp2, *temp3, *temp4;

	for (i=0; i<F->B.num_rel; i++)
		{
		copy_path(leftside,F->B.rel_set[i].lhs);

		/* dom is the domain of the left path */
		dom = F->B.arr[leftside[path_len(leftside)-1]][0];
		if (dom == obj)
			{
			temp = left_rel[i];
			temp2 = right_rel[i];
			/* move down to appropriate row */
			while (temp->down->data != row)
				{
				temp = temp->down;
				temp2 = temp2->down;
				}
			temp3 = temp->down;
			temp4 = temp2->down;
			temp->down = temp->down->down;   /* skip over row */
			temp2->down = temp2->down->down; /* being deleted */
			free_row(temp3);
			free(temp3);
			free_row(temp4);
			free(temp4);
			}
		}
}



/* already_coinc
Parameters:
	left,right,obj:   information for the coincidence left-right(obj).
Purpose:
	Returns true if the given coincidence is already in the coinc. list,
	false otherwise.
*/

bool already_coinc(int left, int right, int obj)
{
struct coinc_node *temp;

	temp = co_list.head;
	while (temp != NULL)
		{
		if ((temp->x == left) && (temp->y == right) && (temp->obj == obj))
			return(true);
		temp = temp->next;
		}
	return(false);
}



void renumber()
{
int i,row;
struct nodetype *temp;

	for (i=0; i<F->B.numobj; i++)
		{
		row = 1;
		temp = L_table[i]->down;
		while (temp != NULL)
			{
			if (temp->data != row)
				{
				renum(e_table,F->A.numobj,temp->data,row,i,1);
				renum(L_table,F->B.numobj,temp->data,row,i,0);
				temp->data = row;
				}
			temp = temp->down;
			row++;
			}
		}
}


void renum(struct nodetype *table[MAXWORD], int max, int old, int newnum, int obj, int etable)
{
int i;
struct nodetype *temp, *temp2;

	for (i=0; i<max; i++)
		{
		temp = table[i];
		if (!etable)
			temp = temp->down;
		while (temp != NULL)
			{
			temp2 = temp->across;
			while (temp2 != NULL)
				{
				if (temp2->data == old && temp2->info == obj)
					temp2->data = newnum;
				temp2 = temp2->across;
				}
			temp = temp->down;
			}
		}
}




void free_row(struct nodetype *ptr)
{
	if (ptr->across != NULL)
		free_row(ptr->across);
	free(ptr->across);
}
