/* MENU.C */

/*        Categories database        */
/* M.Fleming, R.Gunther, R.Rosebrugh */
/* Mount Allison University 1994-95  */


#include        <stdio.h>
#include        <string.h>
#include        "menu.h"

category_list c_list;
cat_functor functor;
functor_list f_list;
struct category_node *catptr;
category A,B;
struct big_list beta_list;



/*   top_menu
Purpose:
	To display the top-level menu with choices to go to the
	category menu, functor menu, category tools menu, or the
	right Kan extension feature.
*/

void top_menu(int maxloop)
{
bool done;
char choice;

	initialize_cat_list(&c_list);
	initialize_functor_list(&f_list);
	done = false;
	while(!done)
		{
		clearscreen();
		printf("\t\t\t     Categories Database\n");
		printf("\t\t\t     ~~~~~~~~~~~~~~~~~~~\n\n");
		printf("\t\t\t(1)  Category Menu\n");
		printf("\t\t\t(2)  Functor Menu\n");
		printf("\t\t\t(3)  Category Tools\n");
		printf("\t\t\t(4)  Right Kan Extension\n");
		printf("\t\t\t(5)  Left Kan Extension\n");
		printf("\t\t\t(6)  Change maximum order of endomorphisms\n");
		printf("\t\t\t        (current value = %d)\n\n",maxloop);
		printf("\t\t\t(0)  Quit\n\n\n\n\n\n");
		printf("\n\n\n\nYour choice... ");
		flush();
		choice = getchar();
		switch(choice)
			{
			case '0':
				save_unsaved(c_list,f_list);
				done = true;
				break;

			case '1':
				cat_menu();
				break;

			case '2':
				functor_menu();
				break;

			case '3':
				tools_choice(c_list,maxloop);
				break;

			case '4':
				right_kan(c_list, f_list, maxloop);
				break;

			case '5':
				left_kan(c_list,f_list,maxloop);
				break;

			case '6':
				maxloop = change_maxloop(maxloop);
			}
		}

}


/*   cat_menu
Purpose:
	Displays a menu with all of the category options:
		create, load, edit, display, save, exit
	Then calls the function corresponding to your choice.
*/

void cat_menu()
{
bool done, finished;
char choice, sav, name[MAXWORD];
category D;
struct nodetype *tree1, *tree2, *B_tree;
int path[MAXWORD], alpha[MAXWORD], beta[MAXWORD], objects[MAXWORD], i, num;
struct functor_node *functorptr;
char ans[MAXWORD];


	done = false;
	while(!done)
		{
		clearscreen();
		printf("\t\t\t     CATEGORY MENU\n");
		printf("\t\t\t     ~~~~~~~~~~~~~\n\n");
		printf("\t\t\t(1)  Create category\n");
		printf("\t\t\t(2)  Load category\n");
		printf("\t\t\t(3)  Edit category\n");
		printf("\t\t\t(4)  Display category\n");
		printf("\t\t\t(5)  List current categories\n");
		printf("\t\t\t(6)  Save category\n");
		printf("\t\t\t(7)  Remove category\n\n");
		printf("\t\t\t(0)  Back to main menu\n\n\n\n\n\n");
		printf("\n\nYour choice... ");
		flush();
		choice = getchar();
		switch(choice)
			{
			case '0':
				done = true;
				break;

			case '1':
/*                              A.num_rel = 0;*/
				create_category(&A);
				if (strcmp(A.name,"-1") != 0)
					{
					add_category(&c_list,A);
					c_list.tail->saved = 0;
					}
				flush();
				break;

			case '2':
				A = load_category();
				if (strcmp(A.name,"~") != 0)
					{
					add_category(&c_list,A);
					}
				flush();
				break;

			case '3':
				edit_choice(c_list);
				break;

			case '4':
				display_choice(c_list);
				break;

			case '5':
				display_cats(c_list);
				break;

			case '6':
				save_choice(c_list);
				break;

			case '7':
				remove_choice(&c_list);

			}
		}


}


/*   ask_for_cat
Parameters:
	c_list:     the list of currently loaded categories.
	operation:  a string representing what is to be done with the
				category chosen.
	finished:       a boolean storing whether or not the operation is completed.
Purpose:
	To ask the user to choose the category that is to be manipulated.
	Returns a pointer to the selected category.
*/

struct category_node *ask_for_cat(category_list c_list, char operation[15], bool *finished)
{
char name[MAXWORD];
int num;

	display_cat_list(c_list);
	printf("(-1 to cancel)\n");
	printf("\nCategory to %s > ",operation);
	flush();
	gets(name);
	num = atoi(name);
	if ((num >= 1) && (num <= num_cat(c_list)))
		{
		catptr = find_category_number(c_list,num-1);
		*finished = true;
		return(catptr);
		}
	if (num == -1)
		{
		*finished = true;
		}
	return(NULL);
}


/*   get_cat_name
Parameters:
	A:      a category
Purpose:
	Gets the name for a category and stores it in the category.
Design:
	Stores a name in a category.
*/

void get_cat_name(category *A)
{
char catname[30];

	flush();
	printf("Category name: ");
	gets(catname);
	strcpy(A->name,catname);
	}


/*   get_obj
Parameters:
	A:      a category
Purpose:
	Allows the user to input objects and stores them in the category.
Design:
	Checks to see that duplicate objects are not entered and the user can
	get a listing of all objects by typing @.
*/

void get_obj(category *A)
{
char ob;

	clearscreen();
	flush();
	printf("\nEnter @ to display all objects");
	printf("\n\nEnter object name (Hit <enter> when finished) : ");
	ob = getchar();
	flush();
	while (ob != '\n')
		{
		if (ob == '@')
			{
			clearscreen();
			display_object(*A);
			}
		else
			{
			if(check_obj(ob,*A))
				{
				A->obj[A->numobj] = ob;
				A->numobj++;
				}
			else
				{
				printf("\n\007***********************ERROR***********************\n");
				printf("    Duplicate object name. Object not added\n");
				printf("**********************************************\n");
				}
			}
			printf("Object: ");
			ob = getchar();
			flush();
		}
}

/*   get_arr
Parameters:
	A:      a category
Purpose:
	Allows the user to input arrows and stores them in the category
Design:
	Checks to see that the domain and codomain of the arrows are objects of
	the category and also makes sure that arrows with duplicate names aren't
	entered.
*/

void get_arr(category *A)
{
char ar,dom,cod;

	flush();
	printf("\nEnter @ to display all arrows");
	printf("\n\nEnter arrow name (Hit <enter> when finished) : ");
	ar = getchar();
	flush();
	while (ar != '\n')
		{
		if (ar == '@')
			{
			clearscreen();
			display_arrow(*A);
			}
		else
			{
			printf("Domain of %c: ",ar);
			dom = getchar();
			flush();
			printf("Codomain of %c: ",ar);
			cod = getchar();
			flush();
			if((check_obj(dom,*A)))
				{
				printf("\n\007***********************ERROR***********************\n");
				printf("   %c is not an object of %s. Arrow not added\n",dom, A->name);
				printf("***************************************************\n");
				}
			 if((check_obj(cod,*A)))
				{
				printf("\n\007***********************ERROR***********************\n");
				printf("   %c is not an object of %s. Arrow not added\n",cod, A->name);
				printf("***************************************************\n");
				}
			 if(!(check_arr(ar,*A)))
				{
				printf("\n\007***********************ERROR***********************\n");
				printf("      Duplicate arrow name. Arrow not added\n");
				printf("***************************************************\n");
				}
			 if((!(check_obj(dom,*A))) && (!(check_obj(cod,*A))) && (check_arr(ar,*A)))
				{
				A->arrow[A->numarr] = ar;
				A->arr[A->numarr][0] = object_number(dom,*A);
				A->arr[A->numarr][1] = object_number(cod,*A);
				A->numarr++;
				}
			}
			printf("\nArrow: ");
			ar = getchar();
			flush();
		}
}

/*   get_rel
Parameters:
	A:      category
Purpose:
	Allows the user to input as many relations as they want to
Design:
	Allocates space for the new relations and checks them for valid composition
	as well as valid left and right domains and codomains.
*/

void get_rel(category *A)
{
char left[MAXWORD], right[MAXWORD];
bool add;
int leftrel[MAXWORD], rightrel[MAXWORD], j;

	flush();
	left[0] = '!';
	while (strlen(left) != 0)
		{
		add = true;
		printf("\nEnter @ for left side to display all arrows and relations");
		printf("\n\nEnter left side of equation : ");
		gets(left);
		printf("Enter right side of equation : ");
		gets(right);
		if (left[0] == '@')
			{
			clearscreen();
			display_arrow(*A);
			print_rel(*A,stdout);
			}
		if (strcmp(right,"1") == 0)
			{
			right[0] = '\x0';
			}
		if((strlen(left) != 0) && (left[0] != '@'))
			{
				if(!(check_string(left,*A)))
					{
					add = false;
					}
				if(!(check_string(right,*A)))
					{
					add = false;
					}
				if(add)
					{
					copy_path(leftrel,string_to_path(left, *A));
					copy_path(rightrel,string_to_path(right, *A));
					for (j=0;j<path_len(leftrel)-1;j++)
						{
						if ((A->arr[leftrel[j]][0]) != (A->arr[leftrel[j+1]][1]))
							{
							printf("\n\007******************************ERROR******************************\n");
							printf(" Invalid Composition. Domain of %c does not equal Codomain of %c\n",left[j], left[j+1]);
							printf("*****************************************************************\n");
							add = false;
							}
						}
					for (j=0;j<path_len(rightrel);j++)
						{
						if(check_arr(A->arrow[rightrel[j]],*A))
							{
								printf("\n\007***************ERROR***************\n");
								printf("%c is not an arrow of %c",A->arrow[rightrel[j]],A->name);
								printf("***********************************");
								add = false;
							}
						}
					for (j=0;j<path_len(rightrel)-1;j++)
						{
						if ((A->arr[rightrel[j]][0]) != (A->arr[rightrel[j+1]][1]))
							{
							printf("\n\007***********************ERROR*************************************\n");
							printf(" Invalid Composition.  Domain of %c does not equal Codomain of %c\n",right[j], right[j+1]);
							printf("******************************************************************\n");
							add = false;
							}
						}

					if ((A->arr[leftrel[path_len(leftrel)-1]][0]) != (A->arr[rightrel[path_len(rightrel)-1]][0]) && (rightrel[0] != -1))
						{
						printf("\n\007************************************ERROR************************************\n");
						printf("   Invalid Comp. Domain of left side does not equal Domain of right side.\n");
						printf("*****************************************************************************\n");
						add = false;
						}

					if ((A->arr[leftrel[0]][1]) != (A->arr[rightrel[0]][1])  && (rightrel[0] != -1))
						{
						printf("\n\007*************************************ERROR*************************************\n");
						printf("   Invalid Comp. Codomain of left side does not equal Codomain of right side.\n");
						printf("*******************************************************************************\n");
						add = false;
						}

					if (((A->arr[leftrel[0]][1]) != (A->arr[leftrel[path_len(leftrel)-1]][0])) && (rightrel[0] == -1))
						{
						printf("\n\007*************************************ERROR*************************************\n");
						printf("          %s cannot equal identity, not an endomorphism\n",left);
						printf("*******************************************************************************\n");
						add = false;
						}
					}

				if(add)
					{
					if(A->num_rel == 0)
						{
						A->rel_set = (relation*)calloc(1,sizeof(relation));
						}
					else
						{
						A->rel_set = (relation*)realloc(A->rel_set,sizeof(relation) * (A->num_rel+1));
						}
					if(greater(rightrel,leftrel))
						{
						copy_path(A->rel_set[A->num_rel].lhs,rightrel);
						copy_path(A->rel_set[A->num_rel].rhs,leftrel);
						}
					else
						{
						copy_path(A->rel_set[A->num_rel].rhs,rightrel);
						copy_path(A->rel_set[A->num_rel].lhs,leftrel);
						}
					A->num_rel++;
					}
			}
		}
}
/*   create_category
Parameters:
	A:      category
Purpose:
	Allows a user to create a category
Design:
	Uses the functions get_cat_name, get_obj, and get_rel, to allow the
	user to create a category
*/

void create_category(category *A)
{
char sv;
char filename[30];

	A->numarr = 0;
	A->numobj = 0;
	A->num_rel = 0;
	clearscreen();
	printf("(-1 to cancel)\n");
	get_cat_name(A);
	if(strcmp(A->name,"-1") == 0)
		return;
	get_obj(A);
	get_arr(A);
	get_rel(A);
}

/*   save_category
Parameters:
	A:      category
Purpose:
	Allows the user to save a category
Design:
	Asks the user for verication before overwriting a file and saves a category
*/

void save_category(category *A)
{
bool done;
char leave, over;
char filename[30];

	done = false;
	while(!done)
		{
		clearscreen();
		printf("\n\nEnter file name for storing category: ");
		flush();
		gets(filename);
		if((f = fopen(filename,"r")) == NULL)
			{
			fclose(f);
			f = fopen(filename,"w");
			print_c(A,f);
			fclose(f);
			done = true;
			}
		else
			{
			printf("\n\007%s already exists.  Overwrite (Y/N) ? ",filename);
			flush();
			over = getchar();
			if((over == 'y') || (over == 'Y'))
			   {
			   f = fopen(filename,"w");
			   print_c(A,f);
			   fclose(f);
			   done = true;
			   }
			else
				{
				printf("\nExit to category menu (Y/N) ? ");
				leave = getchar();
				if ((leave == 'y') || (leave == 'Y'))
					{
					done = true;
					}
				}
			}
		}
}

/*   load_category
Parameters:
	None.
Purpose:
	Allows a user to load in a category from disk. If there are errors in
	the category, the program will terminate.
Design:
	Gets a category from disk.
*/

category load_category()
{
char choice;
char name[30];
category C;

	strcpy(C.name,"~");
	clearscreen();
	choice = 'y';
	while((choice == 'y') || (choice == 'Y'))
		{
		flush();
		printf("(-1 to cancel)\n");
		printf("Enter file name for category you wish to load> ");
		gets(name);
		if(strcmp(name,"-1") == 0)
			return(C);
		printf("Trying to open file...\n");
		if((f = fopen(name,"r")) == NULL)
			{
			printf("\nError - not able to find file - try again?(y/n)");
			choice = getchar();
			printf("\n");
			}
		else
			{
			C = init_cat();
			printf("\n                    *****SUCCESS*****\n");
			printf("Hit <return> to go back to category menu...");
			flush();
			gets(name);
			choice = 'n';
			}
		}
		fclose(f);
		return(C);
}

/*   display_object
Parameters:
	A:      category
Purpose:
	Displays all of the objects in a category onto the screen.
Design:
	Displays objects onto screen.
*/

void display_object(category A)
{
int i;
char obj;

	clearscreen();
	printf("objects\n");
	printf("\t");
	for(i=0;i<((A.numobj)-1);i++)
		{
		printf("%c,",look_obj(&A,i));
		}
	if(A.numobj!=0)
		{
		printf("%c.\n\n",look_obj(&A,i));
		}
	else
		{
		printf("\n\n");
		}
}

/*    display_arrow
Parameters:
	A:      category
Purpose:
	Displays all of the arrows in a category onto the screen.
Design:
	Displays arrows onto screen.
*/

void display_arrow(category A)
{
int i;
char arr[7];

	clearscreen();
	printf("arrows\n");
	printf("\t");
	for(i=0;i<((A.numarr)-1);i++)
		{
		printf("%s,",look_arr(&A,i,arr));
		}
	if(A.numarr!=0)
		{
		printf("%s.\n\n\n",look_arr(&A,i,arr));
		}
	else
		{
		printf("\n\n");
		}
}

/*   ask_for_obj
Parameters:
	A:      category
Purpose:
	Asks the user for the object they wish to remove from the category
Design:
	Returns the name of the object if there are no arrows in the category
	dependent on it, otherwise gives an error message and returns \t.
*/

int ask_for_obj(category A)
{
char choice[5];
int i,n,OArray[MAXARR];


	n = -2;
	while ((n > A.numobj) || (n <= 0))
		{
		clearscreen();
		if (A.numobj == 0)
			{
			printf("No objects in category\n");
			flush();
			getchar();
			return(-1);
			}
		for (i=1;i<=A.numobj;i++)
			{
			printf("(%2d)  %c\n",i,A.obj[i-1]);
			}
		printf("\nObject to be removed (-1 to cancel): ");
		flush();
		gets(choice);
		n = atoi(choice);
		if (n == -1)
			{
			return(-1);
			}
		if ((n <= A.numobj) && (n > 0))
			{
			if ((all_arr(&A,n-1,OArray) == 0) && (all_arr_cod(&A,n-1,OArray) == 0))
				{
				return(n-1);
				}
			else
				{
				printf("\n\007Object has arrows dependent on it.  Unable to remove.\n");
				flush();
				getchar();
				return(-1);
				}
			}
		}
}

/*   ask_for_arr
Parameters:
	A:      category
Purpose:
	Asks the user for the arrow they wish to remove from the category
Design:
	Returns the name of the arrow if there are no relations in the category
	dependent on it, otherwise gives an error message and returns \t.
*/

int ask_for_arr(category A)
{
char choice[5];
char arr[7];
int i,n,OArray[MAXARR];


	n = -2;
	while ((n > A.numarr) || (n <= 0))
		{
		clearscreen();
		if (A.numarr == 0)
			{
			printf("No arrows in category\n");
			flush();
			getchar();
			return(-1);
			}
		for (i=1;i<=A.numarr;i++)
			{
			printf("(%2d)  %s\n",i,look_arr(&A,i-1,arr));
			}
		printf("\nArrow to be removed (-1 to cancel): ");
		flush();
		gets(choice);
		n = atoi(choice);
		if (n == -1)
			{
			return(-1);
			}
		if ((n <= A.numarr) && (n > 0))
			{
			if (!arrow_in_rel(n-1,A))
				{
				return(n-1);
				}
			else
				{
				printf("\n\007Arrow has relations dependent on it.  Unable to remove.\n");
				flush();
				getchar();
				return(-1);
				}
			}
		}
}


/*   ask_for_rel
Parameters:
	A:      category
Purpose:
	Asks the user which relation they wish to remove.
Design:
	Displays all of the relations onto the screen with their number in front
	of them and gets the user to choose the number of the relation they wish
	to remove.  If they choose an invalid number returns -1, otherwise returns
	the number of the relation they wish to remove.
*/

int ask_for_rel(category A)
{
char choice[5];
int i,n;


	clearscreen();
	if (A.num_rel == 0)
		{
		printf("No relations in category\n");
		flush();
		getchar();
		return(-1);
		}
	for(i=1;i<=A.num_rel;i++)
		{
		printf("%2d. %s=",i,path_to_string(A.rel_set[i-1].lhs,A));
		printf("%s\t\t\t",path_to_string(A.rel_set[i-1].rhs,A));
		if (i%2 == 0)
			{
			printf("\n");
			}
		}
	printf("\n\nYour choice (-1 to quit)...");
	flush();
	gets(choice);
	n = atoi(choice);
	if ((n <= A.num_rel) && (n > 0))
		{
		return(n-1);
		}
	return(-1);
}


/*  edit_category
Parameters:
	A:      category
Purpose:
	Displays a category edit menu with the following options:
		change name, add objects, add arrows, add relations, remove objects,
		remove arrows, remove relations, display category, exit to main menu
Design:
	Displays the menu and calls the appropiate function
*/

void edit_category(category *A)
{
bool done;
char choice;
int ob,arr,rel;

	done = false;
	while(!done)
		{
		clearscreen();
		printf("\t\t\t\tCATEGORY EDIT MENU\n");
		printf("\t\t\t\t~~~~~~~~~~~~~~~~~~\n");
		printf("\t\t\t\t(1) Change Name\n");
		printf("\t\t\t\t(2) Add Objects\n");
		printf("\t\t\t\t(3) Add Arrows\n");
		printf("\t\t\t\t(4) Add Relations\n");
		printf("\t\t\t\t(5) Remove Objects\n");
		printf("\t\t\t\t(6) Remove Arrows\n");
		printf("\t\t\t\t(7) Remove Relations\n");
		printf("\t\t\t\t(8) Display Category\n\n");
		printf("\t\t\t\t(0) Exit to Category Menu\n\n\n\n\n\n\n\n");
		printf("Your choice... ");
		flush();
		choice = getchar();
		switch(choice)
			{
			case '0':
				done = true;
				break;

			case '1':
				clearscreen();
				get_cat_name(A);
				break;

			case '2':
				get_obj(A);
				break;

			case '3':
				display_arrow(*A);
				get_arr(A);
				break;

			case '4':
				clearscreen();
				display_arrow(*A);
				print_rel(*A,stdout);
				get_rel(A);
				break;

			case '5':
				ob = -2;
				while(ob != -1)
					{
					clearscreen();
					ob = ask_for_obj(*A);
					if (ob != -1)
						{
						remove_obj(ob,A);
						}
					}
				break;

			case '6':
				arr = -2;
				while(arr != -1)
					{
					clearscreen();
					arr = ask_for_arr(*A);
					if (arr != -1)
						{
						remove_arr(arr,A);
						}
					}
				break;

			case '7':
				rel = -2;
				while(rel != -1)
					{
					rel = ask_for_rel(*A);
					if (rel != -1)
						{
						remove_rel(rel,A);
						}
					}
				break;


			case '8':
				clearscreen();
				print_c(A,stdout);
				printf("\nHit <enter> to continue\n");
				flush();
				getchar();
				break;
			}
		}
}


/*   ask_for_paths
Parameters:
	p1, p2: two arrays of integers which will contain the paths the user enters
	A:      category
Purpose:
	To get two paths from the user
Design:
	Gets two paths from the user and converts them to arrays of integers
	terminated by -1 and then stores them in p1 and p2.
*/

int ask_for_paths(int p1[MAXWORD], int p2[MAXWORD], category *A)
{
char first[30], second[30];
int q;


	p1[0] = p2[0] = -2;
	printf("(-1 to cancel)\n");
	printf("First Path: ");
	flush();
	gets(first);
	if (strcmp(first,"-1") == 0)
		return(1);
	for (q=0; q<strlen(first); q++)
		{
		if (first[q] != '1')
			{
			if (check_arr(first[q],*A))
				{
				printf("\n\007*** Invalid path ... %c is not an arrow in %s. ***\n",first[q],A->name);
				return(0);
				}
			}
		}
	copy_path(p1, string_to_path(first, *A));
	if (!check_path(p1,*A))
		{
		p1[0] = -2;
		return(0);
		}

	printf("Second Path: ");
	flush();
	gets(second);
	if (strcmp(second,"-1") == 0)
		return(1);
	for (q=0; q<strlen(second);q++)
		{
		if (second[q] != '1')
			{
			if (check_arr(second[q],*A))
				{
				printf("\n\007*** Invalid path ... %c is not an arrow in %s. ***\n",second[q],A->name);
				return(0);
				}
			}
		}
	copy_path(p2, string_to_path(second,*A));
	if (!check_path(p2,*A))
		{
		p2[0] = -2;
		return(0);
		}
	return(1);
}


/*   category_tools
Parameters:
	A:      category
Purpose:
	Displays a menu of category tools with the following options:
		make confluent, initial object, equals, make dual,
		display category, exit to main menu
Design:
	Displays the menu and calls the appropiate function
*/

void category_tools(category* A, int maxloop)
{
bool done, finished;
char choice, name[MAXWORD];
int p1[MAXWORD], p2[MAXWORD], redp1[MAXWORD], redp2[MAXWORD], num;
struct category_node *catptrA, *catptrB;
struct functor_node *functorptrAB, *functorptrASET;
int obj,i,levels;
int beta[MAXWORD], path[MAXWORD], objects[MAXWORD];
char betastr[MAXWORD];
bool check;
struct nodetype *B_tree, *product_tree, *RofB_tree;
int count;
bool success;
int ids;
char idarray[MAXWORD];
int rhoarr[MAXWORD];


	done = false;
	while (!done)
		{
		clearscreen();
		printf("\t\t\t\t    CATEGORY TOOLS\n");
		printf("\t\t\t\t    ~~~~~~~~~~~~~~\n\n");
		printf("\t\t\t\t(1) Make Confluent\n");
		printf("\t\t\t\t(2) Initial Object?\n");
		printf("\t\t\t\t(3) Equality of Composites\n");
		printf("\t\t\t\t(4) Make Dual\n");
		printf("\t\t\t\t(5) Sum?\n");
		printf("\t\t\t\t(6) Display Category\n\n");
		printf("\t\t\t\t(0) Exit to Main Menu\n\n\n\n\n\n\n");
		printf("Your choice... ");
		flush();
		choice = getchar();

		switch(choice)
			{
			case '0':
				done = true;
				break;

			case '1':
				make_confluent(A);
				printf("\nHit <enter> to continue\n");
				flush();
				getchar();
				catptrA->saved = false;
				break;

			case '2':
				clearscreen();
				initial_object(*A,maxloop);
				printf("\nHit <enter> to continue\n");
				flush();
				getchar();
				break;

			case '3':
				equality_test(A);
				break;

			case '4':
				clearscreen();
				make_dual(*A,c_list);
				flush();
				break;

			case '5':
				clearscreen();
				print_c(A,stdout);
				printf("\n\n");
				sum(*A,maxloop);
				flush();
				break;

			case '6':
				clearscreen();
				print_c(A,stdout);
				printf("\nHit <enter> to continue\n");
				flush();
				getchar();
			}
		}
}

/*   functor_menu
Parameters:
	None
Purpose:
	Displays a menu of functor operations with the following options:
		Create functor, Create functor to SET, Load functor,
		Load functor to SET, Remove functor, Save functor, Display functor,
		List current functors.
Design:
	Displays the menu and calls the appropiate function
*/

void functor_menu()
{
bool done, finished;
char name[MAXWORD];
int num;
char choice;
struct functor_node *functorptr, fnode;
int num_funcs;
char ans[MAXWORD];

	done = false;
	while (!done)
		{
		clearscreen();
		printf("\t\t\t\t     FUNCTOR MENU\n");
		printf("\t\t\t\t     ~~~~~~~~~~~~\n\n");
		printf("\t\t\t\t(1)  Create functor\n");
		printf("\t\t\t\t(2)  Create functor to SET\n");
		printf("\t\t\t\t(3)  Load functor\n");
		printf("\t\t\t\t(4)  Load functor to SET\n");
		printf("\t\t\t\t(5)  Remove functor\n");
		printf("\t\t\t\t(6)  Save functor\n");
		printf("\t\t\t\t(7)  Display functor \n");
		printf("\t\t\t\t(8)  List current functors\n\n");
		printf("\t\t\t\t(0)  Exit to main menu\n\n\n\n\n\n\n");
		printf("Your choice... ");
		flush();
		choice = getchar();
		switch(choice)
			{
			case '0':
				done = true;
				break;

			case '1':
				create_functor(c_list,&f_list);
				break;


			case '2':
				create_sf(c_list,&f_list);
				break;


			case '3':
				fnode = load_functor(0);
				if (strcmp(fnode.f.name,"~") != 0)
					add_functor(&f_list,fnode);
				flush();
				break;

			case '4':
				fnode = load_functor(1);
				if (strcmp(fnode.f.name,"~") != 0)
					add_functor(&f_list,fnode);
				flush();
				break;

			case '5':
				remove_func_proc(&f_list);
				break;

			case '6':
				save_func_proc(f_list);
				break;

			case '7':
				display_func_proc(f_list);
				break;

			case '8':
				clearscreen();
				display_functor_list(f_list);
				printf("Hit <enter> to continue...");
				flush();
				getchar();

			}
		}
}


/*   save_unsaved
Parameters:
	c_list: List of currently loaded categories
	f_list: List of currently loaded functors
Purpose:
	To ask the user if each unsaved category and functor is to be saved
	before quitting the program.
*/

void save_unsaved(category_list c_list, functor_list f_list)
{
struct category_node *catptr;
int i;
struct functor_node *functorptr;
char sav;

	clearscreen();
	for(i=0;i<num_cat(c_list);i++)
		{
		catptr = find_category_number(c_list,i);
		if(catptr->saved == 0)
			{
			printf("\007\nCategory %s not saved.  Save(Y/N)?",catptr->cat.name);
			flush();
			sav = getchar();
			if((sav == 'Y') || (sav == 'y'))
				{
				save_category(&(catptr->cat));
				catptr->saved = 1;
				}
			}
		}
	for(i=0;i<num_functors(f_list);i++)
		{
		functorptr = find_functor_number(f_list,i);
		if(functorptr->saved == 0)
			{
			printf("\007\nFunctor %s not saved.  Save(Y/N)?",functorptr->f.name);
			flush();
			sav = getchar();
			if((sav == 'Y') || (sav == 'y'))
				{
				save_functor(*functorptr);
				functorptr->saved = 1;
				}
			}
		}
}

/*   edit_choice
Parameters:
	c_list: List of currently loaded categories
Purpose:
	Provided that any categories are loaded, this function will ask the user
	to select the desired category from the list, and then call the function
	edit_category to display the editing options.
*/

void edit_choice(category_list c_list)
{
struct category_node *catptr;
bool finished;


	clearscreen();
	finished = false;
	while (!finished)
		{
		if (c_list.head == NULL)
			{
			printf("\007No categories loaded\n\n");
			printf("Hit <enter> to continue\n");
			flush();
			getchar();
			finished = true;
			}
		else
			{
			catptr = ask_for_cat(c_list,"edit",&finished);
			if (catptr != NULL)
				{
				catptr->saved = 0;
				edit_category(&(catptr->cat));
				}
			flush();
			}
		}
}


/*   display_choice
Parameters:
	c_list: List of currently loaded categories
Purpose:
	Provided that any categories are loaded, this function will ask the user
	to select the desired category from the list, and then call the function
	print_c to display the chosen category.
*/

void display_choice(category_list c_list)
{
bool finished;
struct category_node *catptr;


	clearscreen();
	finished = false;
	while (!finished)
		{
		if (c_list.head == NULL)
			{
			printf("\007No categories loaded\n\n");
			finished = true;
			}
		else
			{
			catptr = ask_for_cat(c_list,"display",&finished);
			if (catptr != NULL)
				{
				clearscreen();
				print_c(&(catptr->cat),stdout);
				}
			}
		}
	printf("\nHit <enter> to continue\n");
	flush();
	getchar();
}

/*  display_cats
Parameters:
	c_list: List of currently loaded categories
Purpose:
	To call function display_cat_list to list all available categories,
	if any are loaded.
*/

void display_cats(category_list c_list)
{
	clearscreen();
	if (c_list.head == NULL)
		{
		printf("\007No categories loaded\n\n");
		printf("Hit <enter> to continue");
		}
	else
		{
		display_cat_list(c_list);
		}
	flush();
	getchar();
}

/*   save_choice
Parameters:
	c_list: List of all currently loaded categories.
Purpose:
	Provided that any categories are loaded, this function will ask the user
	to select the desired category from the list, and then call the function
	save_category to save it.
*/

void save_choice(category_list c_list)
{
bool finished;
struct category_node *catptr;

	clearscreen();
	finished = false;
	while (!finished)
		{
		if (c_list.head == NULL)
			{
			printf("\007No categories loaded\n\n");
			finished = true;
			}
		else
			{
			catptr = ask_for_cat(c_list,"save",&finished);
			if (catptr != NULL)
				{
				catptr->saved = 1;
				save_category(&(catptr->cat));
				printf("\nCategory saved...\n");
				}
			}
		}
	printf("Hit <enter> to continue\n");
	flush();
	getchar();
}


/*   remove_choice
Parameters:
	c_list: List of all currently loaded categories.
Purpose:
	Provided that any categories are loaded, this function will ask the user
	to select the desired category from the list, and then call the function
	remove_category to remove it, first asking the user if it should first
	be saved.
*/

void remove_choice(category_list *c_list)
{
bool finished;
struct category_node *catptr;
char ans[MAXWORD];


	clearscreen();
	finished = false;
	while (!finished)
		{
		if (c_list->head == NULL)
			{
			printf("\007No categories loaded\n\n");
			printf("Hit <enter> to continue");
			flush();
			getchar();
			finished = true;
			}
		else
			{
			catptr = ask_for_cat(*c_list,"remove",&finished);
			if (catptr != NULL)
				{
				if (catptr->saved == 0)
					{
					printf("\nSave %s before removing (Y/N)?  ",catptr->cat.name);
					gets(ans);
					if ((ans[0] == 'y') || (ans[0] == 'Y'))
						save_category(&(catptr->cat));
					}
				remove_category(c_list, catptr);
				}
			}
		}
	flush();
}

/*   tools_choice
Parameters:
	c_list: List of all currently loaded categories.
Purpose:
	Provided that any categories are loaded, this function will ask the user
	to select the desired category from the list, and then call the function
	category_tools to display the menu options.
*/

void tools_choice(category_list c_list, int maxloop)
{
bool finished;
struct category_node *catptr;


	clearscreen();
	finished = false;
	while (!finished)
		{
		if (c_list.head == NULL)
			{
			printf("\007No categories loaded\n\n");
			printf("Hit <enter> to continue");
			flush();
			getchar();
			finished = true;
			}
		else
			{
			catptr = ask_for_cat(c_list,"apply tools to",&finished);
			if (catptr != NULL)
				{
				category_tools(&(catptr->cat),maxloop);
				}
			}
		}
	flush();
}


/*   equality_test
Parameters:
	A:      pointer to a category
Purpose:
	To ask the user to input two paths, and to determine if the two
	paths are equal under the relations of the current category.
*/

void equality_test(category *A)
{
bool success;
int p1[MAXWORD], p2[MAXWORD], redp1[MAXWORD], redp2[MAXWORD];


	success = false;
	while (!success)
		{
		success = true;
		clearscreen();
		print_c(A,stdout);
		printf("\n\n");
		if (ask_for_paths(p1,p2,A))
			{
			if ((p1[0] == -2) || (p2[0] == -2))
				return;
			}
		else
			{
			if ((p1[0] == -2) || (p2[0] == -2))
				{
				success = false;
				printf("\nHit <enter> to continue.");
				flush();
				getchar();
				}
			}
		}
	if ((p1[0] == -2) || (p2[0] == -2))
		return;
	reduce(redp1, p1, *A);
	reduce(redp2, p2, *A);
	if ((equals(redp1, redp2)) && ((A->arr[p1[0]][1] == A->arr[p2[0]][1]) || (p1[0] == -1 || p2[0] == -1)))
		{
		printf("\n%s and ",path_to_string(p1,*A));
		printf("%s are equal paths.\n",path_to_string(p2,*A));
		}
	else
		{
		printf("\n%s and ",path_to_string(p1,*A));
		printf("%s are NOT equal paths.\n",path_to_string(p2,*A));
		}
	printf("\nHit <enter> to continue\n");
	flush();
	getchar();
}




int change_maxloop(int maxloop)
{
char str[MAXWORD];
int x;
bool done;

	done = false;
	while (!done)
		{
		clearscreen();
		printf("Change maximum order of enodmorphisms...\n\n");
		printf("Current value is %d.\n",maxloop);
		printf("New value >   ");
		flush();
		gets(str);
		x = atoi(str);
		if (x > 0)
			done = true;
		}
	return(x);
}
