/* READ.C */

/*        Categories database        */
/* M.Fleming, R.Gunther, R.Rosebrugh */
/* Mount Allison University 1994-95  */

#include        <stdio.h>
#include        <string.h>
#include        <ctype.h>
#include        "read.h"

typedef                                         /* used to pass error messages back to user */
		enum  {syntax,eof_e}
								e_type;
char    line[500];                      /* stores current line */
int     linenum = 0;      /* stores current line number */
char dummyA[MAXWORD];

/*
Displays the error message which is passed in as a parameter.
*/
void fatal_error(char *message)
{
	fprintf(outf,"\007 %s\n",message);
	exit(0);
}


/*
Parameters:
	t:              variable of type e_type indicating the type of error
	m1,m2:  strings to be used in the display of the error message

Purpose:
	To display an error message when a syntax or end-of-file error is
	encountered.
*/

void    read_error(e_type t,char *m1,char *m2)  /* error messaging routine */
{
	int i;
	rewind(f);                                      /* move back to the start */
	for (i=0;i<linenum;i++)
		fgets(line,500,f);      /* search for the line */
	fprintf(stdout,"\007Line : %d\n%s\n",linenum,line);
	switch(t)
		{
		case syntax :
					fprintf(stdout,"Syntax error: expected %s, found %s\n",m1,m2);
					break;
		case eof_e:
					fprintf(stdout,"End of file encountered before %s\n",m1);
					break;
		}
	exit(1);
	return;
}


/* compares first letter of c with ch if OK moves c to next letter */

char*   check(char *c,char ch)
{
char m[2];

	if (*c == ch)
		*c++;
	else
		{
		m[0] = ch;                      /* form ch into a string */
		m[1] = '\0';
		c[1] = '\0';            /* truncate c to one character */
		read_error(syntax,m,c);
		}
	return(c);
}


/* reads the string n from the string c and returns pointer to end of
		n in c */

char*   checkname(char *c,char *n)
{
	int i,l=strlen(n);
	for (i=0;i<l;i++)
		c = check(c,n[i]);      /* if error found check will flag it ! */
	return(c);
}


/* reads in next line from input buffer */

void    get_line(char *message)
{
	if (feof(f))
		{
		linenum--;
		read_error(eof_e,message,"");
		}
	fgets(line,500,f);
	linenum++;
	return;
}


/* converts string to upper case */

void    upper(char *strng)
{
	int     i;
	for(i=0; strng[i] != '\0'; i++)
		{
		strng[i] = (strng[i]>='a'&&strng[i]<='z')?strng[i]+'A'-'a':strng[i];
		}
}



/* clears leading and trailing spaces from global string line */

char    *clear()
{
	char *c = line;
	int     i = 0;
	while (isspace(line[i]))
		i++;
	c = &line[i];
	while (isalnum(line[i]))
		i++;
	line[i] = '\0';
	return(c);
}




/* reads the line with keyword from input buffer */

void    pass(char *keyword)
{       char *c;
	get_line(keyword);
	c=clear();
	while(!strcmp(c,""))
		{
		get_line("");
		c=clear();
		}
	strcpy(line,c);
	upper(line);
	upper(keyword);
	if (strcmp(line,keyword))
		read_error(syntax,keyword,c);
}



/* copies into temp the line with all spaces removed */

char    *skip(char temp[500])
{       char *c= temp;
	int i =0,l = strlen(line),k=0;
	while (i<l)
		{
		if (!isspace(line[i]))
			{
			temp[k] = line[i];
			k++;
			}
		i++;
		}
	temp[k] = '\0';
	return(c);
}


/*              reads in the name of the next definition.
		note: this is used by all routines and skips the label
		'category' passed in keyword */

void    readname(char *name,char *keyword)
{
	char    *c;
	static  bool    first;

	first = true;
	if (first)
		pass(keyword);          /*skip category label*/
	first = !first;
	get_line("");
	c=clear();
	while(!strcmp(c,""))
		{
		get_line("");
		c=clear();
		}
	strcpy(name,c);
}


/*
Parameter:
	obj:    a string to contain all objects in the category

Purpose:
	To read in all objects from a file and store in the category in memory.
	Returns an integer representing the number of objects.  Obj is
	changed so that it contains all objects read in.
*/

int     readobj(char *obj)
{
	
	int     num=0;                          /* number so far read in */
	char    *c,temp[500];
	strcpy(dummyA,"objects");
	pass(dummyA);                        /* skip objects keyword */
	get_line("");
	c = skip(temp);
	while (true)
		{
		while (*c == '\0')      /* >1 line input */
			{
			get_line("");     /* get next line */
			c = skip(temp);
			}
		obj[num] = *c;                  /* get next object */
		num++;
		*c++;
		if (*c == '.')          /* finished */
			return(num);
		*c++;                                           /* skip the comma */
		}
}


/*
Parameters:
	arr:    a string to contain all arrows in the category
	dom:    a string to contain the domains of all arrows
	cod:    a string to contain the codomains of all arrows
Purpose:
	To read in all arrows from a file and store in the category in memory.
	Returns an integer representing the number of arrows.  Arr, dom, and
	cod are changed so that they contain all arrows in the category,
	their domains and codomains.
*/

int     readarr(char *arr,char *dom,char *cod)
{
	
	int     num=0;                             /* number so far read in */
	char    *c,temp[500];
	strcpy(dummyA,"arrows");
	pass(dummyA);                         /* skip arrows keyword */
	get_line("");
	c = skip(temp);
	while (true)
		{
		while (*c == '\0')                      /* >1 line input */
			{
			get_line("");
			c = skip(temp);
			}
		if (*c == '.')                  /* finished */
			return(num);
		arr[num] = *c++;                        /* get next arrow */
		c = check(c,':');
		dom[num] = *c++;                        /* domain */
		c = check(c,'-');
		c = check(c,'>');
		cod[num++] = *c++;              /* co domain */
		if (*c == '.')                  /* finished */
			return(num);
		c = check(c,',');
		}
}


/*
Parameters:
	lhs, rhs:       two-dimensional arrays of characters representing the
				left and right sides of paths.
Purpose:
	To read in all relations from a file and store theie left and right sides
	in the arrays lhs and rhs, respectively.  Returns an integer representing
	the number of relations in the category.
*/

int     readrel(char lhs[MAXARR][MAXWORD],char rhs[MAXARR][MAXWORD])
{
	
	char *c,*p,temp[500];
	int     num = 0;                                /* number read in so far */
	bool    last=false;                     /* end of input flag */
	strcpy(dummyA,"relations");
	pass(dummyA);              /* skip keyword */
	get_line("");
	c = skip(temp);                 /* clear out spaces */
	p = c;
	while(true)
		{
		while (*p == '\0')
			{
			get_line("");                           /* read a new line */
			c = skip(temp);
			p = c;
			}
		if (*c == '.')                                  /* end of input */
			return(num);
		while (isalnum(*p))
			*p++;
		*p = '\0';
		strcpy(lhs[num],c);                     /* read and copy the first */
		*p++;
		c = p;
		while (isalnum(*p))                     /* read in second */
			*p++;
		last = (*p=='.');                               /* last one !*/
		if (!last)
			check(p,',');                           /* check separator */
		*p++ = '\0';
		strcpy(rhs[num],c);                     /* copy it */
		num++;
		c = p;                                                  /* move pointer to next entry */
		if (last)
			return(num);                            /* finished */
		}
}

int     readfobj(char *name,char *lhs,char *rhs)
{
	char *c,temp[500];
	int     num = 0;                                        /* number read in so far */
	strcpy(dummyA,"objects");
	pass(dummyA);                                /* skip keyword */
	get_line("");
	c = skip(temp);                         /* clear out spaces */
	while(true)
		{
		while (*c == '\0')
			{
			get_line("");                   /* read a new line */
			c = skip(temp);
			}
		c = checkname(c,name);
		c = check(c,'(');
		lhs[num] = *c++;                        /* copy the domain*/
		c = check(c,')');
		c = check(c,'=');
		rhs[num++] = *c++;              /* copy it */
		if (*c == '.')
			return(num);                    /* finished */
		c = check(c,',');
		}
}

int     readfarr(char *name,char *lhs,char rhs[MAXARR][MAXWORD])
{
	char *c,*p,temp[500];
	int     num = 0;                                        /* number read in so far */
	bool    last = false;                   /* end of input flag */
	strcpy(dummyA,"arrows");
	pass(dummyA);                         /* skip keyword */
	get_line("");
	c = skip(temp);                         /* clear out spaces */
	while(true)
		{
		if (*c == '\0')                 /* end of line */
			{
			get_line("");                   /* read a new line */
			c = skip(temp);
			}
		if (*c == '.')          /* end of input */
			return(num);
		c = checkname(c,name);
		c = check(c,'(');
		lhs[num] = *c++;                        /* copy the domain*/
		c = check(c,')');
		c = check(c,'=');
		p = c;                                          /* point at start of word */
		while (isalnum(*p))
			*p++;
		last = (*p=='.');                       /* end of input */
		if (!last)
			check(p,',');                   /* check separator */
		*p++='\0';
		strcpy(rhs[num],c);             /* copy word */
		num++;                                          /* increment entry counter */
		c = p;                                          /* move to next entry */
		if (last)                                       /* finished */
			return(num);
		}
}


int     readxobj(char *name,char *lhs,int *rhs)
{
	char *c,*p,temp[500];
	int     num = 0;                                        /* number read in so far */
	strcpy(dummyA,"objects");
	pass(dummyA);                                /* skip keyword */
	get_line("");
	c = skip(temp);                         /* clear out spaces */
	while(true)
		{
		if (*c == '\0')         /* empty line */
			{
			get_line("");                   /* read a new line */
			c = skip(temp);
			}
		if (*c == '.')                          /* finished return the number read in */
			return(num);
		c = checkname(c,name);
		c = check(c,'(');
		lhs[num] = *c++;                        /* copy the domain*/
		c = check(c,')');
		c = check(c,'=');
		c = check(c,'[');
		p = c;                                          /* point to start */
		while (isdigit(*p))
			*p++;
		check(p,']');
		*p++='\0';                                      /* terminate string */
		rhs[num]=atoi(c);                       /* copy it */
		num++;
		if (*p != '.')                          /* not the last one */
			c = check(p,',');
		else
			return(num);
		}
}

int     readxarr(char *name,char *lhs,int rhs[MAXARR][MAXWORD])
{
	char *c,*p,temp[500];
	int     num = 0;                                                        /* number read in so far */
	int     k,done;
	strcpy(dummyA,"arrows");
	pass(dummyA);                         /* skip keyword */
	get_line("");
	c = skip(temp);                         /* clear out spaces */
	while(true)
		{
		if (*c == '\0')
			{
			get_line("EOF");        /* read a new line */
			c = skip(temp);
			}
		if (*c == '.')
			return(num);
		c = checkname(c,name);
		c = check(c,'(');
		lhs[num] = *c++;                /* copy the domain*/
		c = check(c,')');
		c = check(c,'=');
		c = check(c,'(');
		k = 0;
		p = c;
		done = false;
		while (!done)
			{
			while (isdigit(*p))
				*p++;
			done=(*p==')');
			*p='\0';
			rhs[num][k] = atoi(c);
			k++;
			*p++;                                                   /* on the digit */
			c = p;
			}
		num++;
		if (*c == '.')
			return(num);
		*c++;
		}
} 
