/*RKAN.C*/

/*        Categories database        */
/* M.Fleming, R.Gunther, R.Rosebrugh */
/* Mount Allison University 1994-95  */

#include        "rkan.h"
#include        <string.h>


/* Global variables */
bool to_fil;
FILE *outfil;


/*   eliminate_betas
Parameters:
	path:   the current path of the tree
	domain: the domain of the current arrow of the path
	objects: an array which keeps track of how many times each object has
			been visited during the current path
	beta: a path from B to an F(A)
	B_tree: the tree that contains all of the betas
	product_tree: the tree that contains all of the betas which we will use
					in the product
Purpose:
	To eliminate all of the objects which can be reached from the current object
Design:
	Beta is an arrow from B-->F(A) and this recursive function finds every
	path alpha from A to A' and eliminates F(alpha)beta from the B_tree.
	If F(alpha)beta is in the product tree then it is eliminated from
	that tree as well.  A path is eliminated from a tree by setting the tree
	pointer of the last node of the path to NULL.
*/

void eliminate_betas(int path[MAXWORD], int domain, int objects[], struct functor_node *F, int beta[MAXWORD], struct nodetype *B_tree, struct nodetype *product_tree, int maxloop)
{

/*
	number: stores the number of arrows from an object
	current_arrow:  stores the number of the current arrow in the path
	codom:  stores the codomain of the current arrow
	arrow_array:    contains all of the arrows from an object
*/

int i,j, number, current_arrow, codom;
int arrow_array[MAXARR], reducedpath[MAXWORD], temp_path[MAXWORD];
struct nodetype  *pointer;
struct listnodetype *list, *temp_list;

list = (struct listnodetype*)malloc(sizeof(struct listnodetype));
temp_list = (struct listnodetype*)malloc(sizeof(struct listnodetype));

if(!(objects[domain] == maxloop))                       /*check to see current object has not*/
	{                                                   /*already been visited the maximum number of times*/
	number = all_arr(&(F->A),domain,arrow_array);   /*get all arrows from current object*/
	for(i=0;i<number;i++)   /*find path for each arrow out of object*/
		{
		objects[domain]++;              /*increment number of times object has been visited*/
		current_arrow = arrow_array[i];
		for((j=path_len(path)+1);j>0;j--)       /*get current arrow and add to path*/
			{
			path[j] = path[j-1];
			}
		path[0]=current_arrow;
		if(objects[codom = (F->A.arr[current_arrow][1])] != maxloop)
			{       /*set codom equal to codomain of current arrow and check that
				the object of the codomain has not already been visited the
				maximum number of times*/
			functor_on_path(*F,path,temp_path);
			append_path(temp_path,beta);
			reduce(reducedpath,temp_path,F->A);             /*make F(alpha)beta, store*/
			make_list(&list,reducedpath);           /*in list and check to see if*/
			if(search_tree(B_tree,list,&pointer))      /*in B_tree or product tree*/
				{
				temp_path[0] = codom;         /*find what A is for F(alpha)beta*/
				temp_path[1] = -1;
				make_list(&temp_list,temp_path);
				search_tree(pointer->tree,temp_list,&pointer);
				pointer->info = 1;
				}                                   /*if in a tree set the tree pointer*/
			if(search_tree(product_tree,list,&pointer)) /*of the last node of the path to NULL*/
				{
				if(search_tree(pointer->tree,temp_list,&pointer))
					{
					pointer->tree = NULL;
					}
				}
			free_list(list);
			}

			/*keep going along path with codomain of current arrow as new domain*/

			eliminate_betas(path,codom,objects,F,beta,B_tree,product_tree,maxloop);
			for(j=1;j<=path_len(path);j++)
				{
				path[j-1] = path[j];    /*arrow no longer in path, remove it*/
				}
			objects[domain]--;      /*object no longer part of path, decrement number*/
		}                                               /*of times it has been visited*/
	}
}


/*   get_all_betas
Parameters:
	tree:   a pointer to the current node of the beta_tree
	beta: the current beta of the tree
	F: a pointer to a functor_node which contains the current functor
	product_tree: the tree which contains all of the betas that will make
					up the product
	beta_tree: a tree which contains all of the betas
Purpose:
	To get all of the betas which either have no path into them or are a representative
	of a loop with no path into it.
Design:
	This is a recursive function which goes through a tree of betas and gets
	each beta and eliminates all of the paths which it goes to.  It then adds beta
	to the product_tree.
*/

void get_all_betas(struct nodetype *tree, int beta[MAXWORD],struct functor_node *F, struct nodetype *product_tree, struct nodetype *betatree, int maxloop)
{
struct nodetype* temp;
int i,j;
int objects[MAXWORD], path[MAXWORD], reduced_beta[MAXWORD], temp_path[MAXWORD];
listtype list, Alist;
struct nodetype *pointer, *tp, *temp_pointer;
bool check, done;


	temp = tree;
	if(temp->data != -1)    /*make sure current node is not the top of the tree*/
		{
		for(i=path_len(beta)+1;i>0;i--)
			{
			beta[i] = beta[i-1];    /*add the arrow of the current node to the beta path*/
			}
		beta[0] = temp->data;
		}
	if (beta[0] == -2)
		{
		beta[0] = -1;
		}
	if (temp->tree != NULL)                 /*if the tree pointer of the current node does not*/
		{                                       /*equal NULL then we have a beta*/
		temp_pointer = temp->tree->down;    /*set temp_pointer to point to A-tree of beta*/
		done = false;
		while(!done)
			{
			if(temp_pointer->info == 0)  /*if tree is not NULL then this (beta,A) has not been used yet*/
				{
				path[0] = -1;
				for(i=0;i<F->A.numobj;i++)
					{
					objects[i] = 0;
					}

				/*get rid of all F(alpha)beta for alpha:A-->A'*/

				if (beta[0] == -2)
					beta[0] = -1;
				eliminate_betas(path,temp_pointer->data,objects,F,beta,betatree,product_tree,maxloop);
				if (beta[0] == -1)
					{
					beta[0] = -2; beta[1] = -1;
					}
				reduce(reduced_beta,beta,F->B);
				make_list(&list,reduced_beta);
				search_tree(betatree,list,&pointer);
				search_and_add(product_tree,list,&tp);
				if(tp->tree == NULL)
					{
					tp->tree = (struct nodetype*)malloc(sizeof(struct nodetype));
					tp->tree = create_tree();
					}
				free_list(list);
				path[0] = temp_pointer->data;
				path[1] = -1;
				make_list(&list,path);
				search_tree(pointer->tree,list,&pointer);
				pointer->info = 1;                   /*eliminate from beta tree and*/
				search_and_add(tp->tree,list,&tp);      /*add to product tree*/
				if(temp_pointer->tree != temp_pointer)
					{
					tp->tree = temp_pointer->tree;
					}
				else
					{
					tp->tree = tp;
					}
				free_list(list);
				}
			if(temp_pointer->flag == 0)     /*if flag is 0 there are no more A's with this beta*/
				{
				done = true;
				}
			else                                            /*otherwise, there is another beta,A'*/
				{
				temp_pointer = temp_pointer->across;
				}
			}
		}
	if (temp->down != NULL)
		{
		get_all_betas(temp->down, beta, F, product_tree, betatree, maxloop);  /*keep searching tree*/
		}
	for(i=0;i<=path_len(beta)-1;i++)
		{
		beta[i] = beta[i+1];
		}
	if (temp->flag)         /*if can go across to another child then go across*/
		{
		get_all_betas(temp->across, beta, F, product_tree, betatree, maxloop);
		}
}


/*   make_beta_objects
Parameters:
	path:   keeps track of the current path
	domain: the domain of the current object
	objects: an array which keeps track of how many times an object has
		been visited during the current path
	tree: the tree which all of the betas will be stored in
	F: a pointer to a functor
Purpose:
	To get all of the paths in B which have an F(A) as a codomain.
Design:
	This is a recursive function which goes through a category and gets
	all of the paths which have an F(A) as a codomain.
*/

void    make_beta_objects(int path[MAXWORD], int domain, int objects[MAXWORD], struct nodetype* tree, struct functor_node *F, int maxloop)
{

/*
	number: stores the number of arrows from an object
	current_arrow:  stores the number of the current arrow in the path
	codom:  stores the codomain of the current arrow
	arrow_array:    contains all of the arrows from an object
*/

int i,j, number, current_arrow, codom;
int arrow_array[MAXARR], reduced_path[MAXWORD], temp_path[MAXWORD];
struct nodetype *pointer;
struct listnodetype *list;

list = (struct listnodetype*)malloc(sizeof(struct listnodetype));

if(!(objects[domain] == maxloop))             /*make sure domain has not already been*/
	{                                   /*visited the maximum number of times*/
	number = all_arr(&(F->B),domain,arrow_array);   /*get all arrows from current object*/
	for(i=0;i<number;i++)   /*find path for each arrow out of object*/
		{
		objects[domain]++;              /*increment number of times object has been visited*/
		current_arrow = arrow_array[i];
		for((j=path_len(path)+1);j>0;j--)       /*get current arrow and add to path*/
			{
			path[j] = path[j-1];
			}
		path[0] = current_arrow;
		if(objects[codom = F->B.arr[current_arrow][1]] != maxloop)
			{
			for(j=0;j<F->A.numobj;j++)      /*check that codomain of current path*/
				{                       /*has not already been visited the maximum*/
				if(F->f.obj[j] == codom)/*number of times and then see if it is an F(A)*/
					{
					reduce(reduced_path, path, F->B);
					if (reduced_path[0] != -1)
						{
						make_list(&list,reduced_path);
						search_and_add(tree,list,&pointer);     /*put it in beta tree and set the*/
						free_list(list);
						temp_path[0] = j;     /*make path with object of A in it*/
						temp_path[1] = -1;
						make_list(&list,temp_path);
						if(pointer->tree == NULL)
							{
							pointer->tree = (struct nodetype*)malloc(sizeof(struct nodetype));
							pointer->tree = create_tree();
							}
						search_and_add(pointer->tree,list,&pointer);    /*put in sub_tree of the last node of path*/
						pointer->tree = pointer;            /*tree pointer of the node of object in A to point*/
						free_list(list);                    /*to the same node*/
						}
					}
				}
			}

		/*keep going along path with codomain of current arrow as new domain*/

		make_beta_objects(path,codom,objects,tree,F,maxloop);
		for(j=1;j<=path_len(path);j++)
			{
			path[j-1] = path[j];    /*arrow no longer in path, remove it*/
			}
		objects[domain]--;      /*object no longer part of path, decrement number*/
		}                       /*of times it has been visited*/
	}
}


/*  make_id_beta_object  
Parameters:
	B:  the object B
	tree:  the B/F tree for that object
	F:  the functor F from the Right Kan extension
Purpose:
	This function makes the identity a beta object if B = F(A),
	i.e. if object B is the image of some object under the functor 
*/

void make_id_beta_object(int B, struct nodetype* tree, struct functor_node *F)
{
int i,j, beta[MAXWORD], path[MAXWORD],newpath[MAXWORD],temp_path[MAXWORD], objects[MAXWORD];
struct listnodetype *list;
struct nodetype *pointer, *tp;

	list = (struct listnodetype*)malloc(sizeof(struct listnodetype));

	path[0] = -2;   /* store identity as -2 */
	path[1] = -1;
	for (j=0;j<F->A.numobj;j++)
		{
		if (F->f.obj[j] == B)
			{
			make_list(&list,path);
			search_and_add(tree,list,&pointer);
			free_list(list);
			temp_path[0] = j;     /*make path with object of A in it*/
			temp_path[1] = -1;
			make_list(&list,temp_path);
			if(pointer->tree == NULL)
				{
				pointer->tree = (struct nodetype*)malloc(sizeof(struct nodetype));
				pointer->tree = create_tree();
				}
			search_and_add(pointer->tree,list,&tp);     /*put in sub_tree of the last node of path*/
			tp->tree = tp;                                                  /*tree pointer of the node of object in A to point*/
			free_list(list);                                /*to the same node*/
			}
		}
}


/* go_through_betas 
Purpose:
	This function takes an arrow in A and an object from the 
	B/F tree and checks to see if the arrow is an arrow in B/F. 
	If it is, it adds it to the B/F tree.
*/

void go_through_betas(struct nodetype *tree, int beta[MAXWORD], int codomain, int firstbeta[MAXWORD], int alpha[MAXWORD], struct functor_node *F, struct nodetype *pointer)
{
struct nodetype* temp;
int i;
int new_path[MAXWORD], reduced_newpath[MAXWORD], red_beta[MAXWORD];
struct listnodetype *list;
struct nodetype *temp_pointer, *tp;
bool done;

	list = (struct listnodetype*)malloc(sizeof(struct listnodetype));

	temp = tree;
	if(temp->data != -1)    /*make sure current node is not the top of the tree*/
		{
		for(i=path_len(beta)+1;i>0;i--)
			{
			beta[i] = beta[i-1];    /*add the arrow of the current node to the beta path*/
			}
		beta[0] = temp->data;
		}
		if (beta[0] == -2)
			{
			beta[0] = -1;  /* stored -2 represents identity */
			}
		if (temp->tree != NULL)                 /*if the tree pointer of the current node does not*/
		{                                       /*equal NULL then we have a beta*/
		temp_pointer = temp->tree->down;        /*set temp_pointer to point to A-tree of beta*/
		done = false;
		while(!done)
			{
			if(temp_pointer->data == codomain)
				{
				functor_on_path(*F,alpha,new_path);
				append_path(new_path,firstbeta);        /*check to see if F(alpha)firstbeta = beta*/
				reduce(reduced_newpath,new_path,F->B);  /*if it does, we have an arrow*/
				reduce(red_beta,beta,F->B);
				if(equals(reduced_newpath,red_beta))
					{
					if(pointer->tree == pointer)
						{
						pointer->tree = (struct nodetype*)malloc(sizeof(struct nodetype));
						pointer->tree = create_tree();  /*store arrow in tree*/
						}
					make_list(&list,alpha);
					search_and_add(pointer->tree,list,&tp);
					tp->tree = temp_pointer;
					free_list(list);
					}
				}
			if(temp_pointer->flag == 0)     /*if flag is 0 there are no more A's with this beta*/
				{
				done = true;
				}
			else                                            /*otherwise, there is another beta,A'*/
				{
				temp_pointer = temp_pointer->across;
				}
			}
		}
	if (temp->down != NULL)
		{
		go_through_betas(temp->down,beta,codomain,firstbeta,alpha,F,pointer);
		}
	for(i=0;i<=path_len(beta)-1;i++)
		{
		beta[i] = beta[i+1];
		}
	if (temp->flag)         /*if can go across to another child then go across*/
		{
		go_through_betas(temp->across,beta,codomain,firstbeta,alpha,F,pointer);
		}
}


/* get_all_arrows 
Purpose:
	This function gets all of the arrows out of an object 
	in the B/F tree and checks to see if they are arrows in B/F.
*/

void get_all_arrows(int path[MAXWORD], int domain, int objects[], struct functor_node *F, int firstbeta[MAXWORD], struct nodetype *pointer, struct nodetype *betatree, int maxloop)
{
int i,j, number, current_arrow, codom;
int arrow_array[MAXARR], temp_path[MAXWORD], beta[MAXWORD];

if(!(objects[domain] == maxloop))             /*check to see current object has not*/
	{                                           /*already been visited the maximum number of times*/
	number = all_arr(&(F->A),domain,arrow_array);   /*get all arrows from current object*/
	for(i=0;i<number;i++)   /*find path for each arrow out of object*/
		{
		objects[domain]++;              /*increment number of times object has been visited*/
		current_arrow = arrow_array[i];
		for((j=path_len(path)+1);j>0;j--)       /*get current arrow and add to path*/
			{
			path[j] = path[j-1];
			}
		path[0]=current_arrow;
		if(objects[codom = (F->A.arr[current_arrow][1])] != maxloop)
			{       /*set codom equal to codomain of current arrow and check that
				the object of the codomain has not already been visited the
				maximum number of times*/
			beta[0] = -1;
			go_through_betas(betatree,beta,codom,firstbeta,path,F,pointer);
			}

			/*keep going along path with codomain of current arrow as new domain*/

			get_all_arrows(path,codom,objects,F,firstbeta,pointer,betatree,maxloop);
			for(j=1;j<=path_len(path);j++)
				{
				path[j-1] = path[j];    /*arrow no longer in path, remove it*/
				}
			objects[domain]--;      /*object no longer part of path, decrement number*/
		}                                               /*of times it has been visited*/
	}
}


/* make_beta_arrows
Purpose:
	For each object in the B/F tree this function will get 
	all of the arrows out of that object .
*/

void make_beta_arrows(struct nodetype *tree, struct functor_node *F, struct nodetype *betatree, int beta[MAXWORD], int maxloop)
{
struct nodetype* temp;
int i;
int objects[MAXWORD], path[MAXWORD];
struct nodetype *temp_pointer;
bool done;

	temp = tree;
	if(temp->data != -1)    /*make sure current node is not the top of the tree*/
		{
		for(i=path_len(beta)+1;i>0;i--)
			{
			beta[i] = beta[i-1];    /*add the arrow of the current node to the beta path*/
			}
		beta[0] = temp->data;
		}
		if (beta[0] == -2)
			{
			beta[0] = -1;   /* stored -2 represents identity */
			}
	if (temp->tree != NULL)                 /*if the tree pointer of the current node does not*/
		{                           /*equal NULL then we have a beta*/
		temp_pointer = temp->tree->down;        /*set temp_pointer to point to A-tree of beta*/
		done = false;
		while(!done)
			{
			path[0] = -1;
			for(i=0;i<F->A.numobj;i++)
				{
				objects[i] = 0;
				}
			get_all_arrows(path,temp_pointer->data,objects,F,beta,temp_pointer,betatree,maxloop);
			if(temp_pointer->flag == 0)     /*if flag is 0 there are no more A's with this beta*/
				{
				done = true;
				}
			else                                            /*otherwise, there is another beta,A'*/
				{
				temp_pointer = temp_pointer->across;
				}
			}
		}
	if (temp->down != NULL)
		{
		make_beta_arrows(temp->down,F,betatree,beta,maxloop);
		}
	if (beta[0] != -1)
		{
		for(i=0;i<=path_len(beta)-1;i++)
			{
			beta[i] = beta[i+1];
			}
		}
	if (temp->flag)         /*if can go across to another child then go across*/
		{
		make_beta_arrows(temp->across,F,betatree,beta,maxloop);
		}
}


/* display_Btree 
Purpose:
	This function displays the betas contained in the 
	B/F tree passed in .
*/

int display_Btree(struct nodetype *tree, int beta[MAXWORD],struct functor_node* F,
	int count, char let)
{
struct nodetype* temp;
int i,j,k;
int objects[MAXWORD], path[MAXWORD];
listtype list;
struct nodetype *pointer;
bool done;
int arr[MAXARR];
int num_arr;

	temp = tree;
	if(temp->data != -1)  /* make sure current node is not top of the tree */
		{
		for(i=path_len(beta)+1;i>0;i--)
			{
			beta[i] = beta[i-1];
			}
		beta[0] = temp->data;    /* add the arrow of the current node to the beta path */
		}
		if (beta[0] == -2)
			{
			beta[0] = -1;   /* stored -2 represents identity */
			}
	if (temp->tree != NULL)
		{
		pointer = temp->tree->down;   /* set pointer to A-tree of beta */
		done = false;
		while(!done)
			{
			/* print all betas */
			if(pointer->tree != NULL)
				{
				printf("%c%d = (%s,%c)\t",let,count,path_to_string(beta,F->B),F->A.obj[pointer->data]);
				if (to_fil) fprintf(outfil,"%c%d = (%s,%c)\t",let,count,path_to_string(beta,F->B),F->A.obj[pointer->data]);
				count++;
				}
			/* go across to any other objects in A-tree */
			if(pointer->flag != 0)
				{
				pointer = pointer->across;
				}
			else
				{
				done = true;
				}
			}
		}

	if (temp->down != NULL)
		{
		count = display_Btree(temp->down, beta, F, count, let);
		}
	for(i=0;i<=path_len(beta)-1;i++)
		{
		beta[i] = beta[i+1];    /* remove last arrow from beta path */
		}
	if (temp->flag)
		{
		count = display_Btree(temp->across, beta, F, count, let);
		}

	return(count);
}


/* display_t */
/* displays the paths in a tree from a beta in the B/F tree */
/* used in above function to display paths out of an object */
/* not currently used anywhere */

void display_t(struct nodetype *tree, int beta[MAXWORD],struct functor_node* F)
{
struct nodetype* temp;
int i,j;
int objects[MAXWORD], path[MAXWORD];
bool done;

	temp = tree;
	if(temp->data != -1)
		{
		for(i=path_len(beta)+1;i>0;i--)
			{
			beta[i] = beta[i-1];
			}
		beta[0] = temp->data;
		}
	if (beta[0] == -2)
		{
		beta[0] = -1;
		}
	if(temp->tree != NULL)
		{
		printf("(%s)\n",path_to_string(beta,F->A));
		}
	if (temp->down != NULL)
		{
		display_t(temp->down, beta, F);
		}
	for(i=0;i<=path_len(beta)-1;i++)
		{
		beta[i] = beta[i+1];
		}
	if (temp->flag)
		{
		display_t(temp->across, beta, F);
		}
}


/* add_to_rkan_list 
Purpose:
	This function adds an rkan_node with an a_beta value 
	of the number passed in to the rkan_list. 
*/

void    add_to_rkan_list(struct rkan_list *r_list, int number)
{
struct rkan_node *temppointer;

	/*create new rkan_node and allocate memory for it*/

	temppointer = (struct rkan_node*)malloc(sizeof(struct rkan_node));
	temppointer->a_beta = number;
	temppointer->next = NULL;
	temppointer->valid = 1;
	temppointer->list.head = NULL;
	temppointer->list.tail = NULL;

	if(r_list->head == NULL)
		{
		r_list->head = temppointer;     /*if list is empty sets head and tail to point*/
		r_list->tail = r_list->head;  /*to new node*/
		}
	else
		{
		r_list->tail->next = temppointer;       /*otherwise sets next pointer of last node to point to*/
		r_list->tail = temppointer;             /*new node and sets tail to point to new node*/
		}
}



/* add_to_big_list
Purpose:
	This function adds an rkan_list to the big_list passed in.  
	Pointer points to a tree node which contains all of the 
	information needed to make the rkan_list.  
	i.e. what A is so that we can find X(A).  
	The functor passed in is a SET functor 
*/

void    add_to_big_list(struct big_list *beta_list, struct nodetype *pointer, struct functor_node *F)
{
struct rkan_list *temppointer;
int i;

	/*create new rkan_list and allocate memory for it*/

	temppointer = (struct rkan_list*)malloc(sizeof(struct rkan_list));
	temppointer->head = NULL;
	temppointer->tail = NULL;
	temppointer->next = NULL;
	if(pointer->tree == pointer)
		{
		temppointer->paths = NULL;      /*if the beta the rkan_list is for has no*/
		}                           /*paths out of it then set paths pointer to NULL*/
	else
		{
		temppointer->paths = pointer->tree;     /*otherwise, set paths pointer to point to*/
		}                                   /*tree which contains all paths out of beta*/
	for(i=1;i<=F->f.obj[pointer->data];i++)
		{
		add_to_rkan_list(temppointer,i);        /*add all of the rkan_nodes to the rkan_list*/
		}                                   /*i.e. if X(A) = 3, then add nodes with 1,2, and 3 in them*/
	if(beta_list->head == NULL)
		{
		beta_list->head = temppointer;     /*if list is empty, sets head and tail to point*/
		beta_list->tail = beta_list->head;  /*to new node*/
		}
	else
		{
		beta_list->tail->next = temppointer;    /*otherwise sets next pointer of last node to point to*/
		beta_list->tail = temppointer;          /*new node and sets tail to point to new node*/
		}
}


/*make_big_list
Purpose:
	This function makes a big_list structure from the betas 
	in the product tree 
*/

void    make_big_list(int beta[MAXWORD], struct nodetype *product_tree, struct big_list *beta_list, struct functor_node *F)
{
struct nodetype* temp;
int i;
struct nodetype *pointer;
bool done;

	/*go through product tree and get all of the betas*/

	temp = product_tree;
	if(temp->data != -1)
		{
		for(i=path_len(beta)+1;i>0;i--)
			{
			beta[i] = beta[i-1];
			}
		beta[0] = temp->data;
		}
	if (beta[0] == -2)
		{
		beta[0] = -1;
		}
	if (temp->tree != NULL)
		{
		pointer = temp->tree->down;
		done = false;
		while(!done)
			{
			if(pointer->tree != NULL)       /*if we have a beta, add it to the beta_list*/
				{
				add_to_big_list(beta_list,pointer,F);
				}
			if(pointer->flag != 0)
				{
				pointer = pointer->across;
				}
			else
				{
				done = true;
				}
			}
		}
	if (temp->down != NULL)
		{
		make_big_list(beta,temp->down,beta_list,F);
		}
	for(i=0;i<=path_len(beta)-1;i++)
		{
		beta[i] = beta[i+1];
		}
	if (temp->flag)
		{
		make_big_list(beta,temp->across,beta_list,F);
		}
}


/* assign_pointers
Purpose:
	This function makes and stores all of the pointers 
	that are needed in the beta_list. 
*/

void    assign_pointers(struct big_list *beta_list)
{
struct rkan_list *temp1, *temp2;
struct rkan_node *temp3, *temp4;
struct b_node *t_node;
struct pointer_node *p_node;

	temp1 = beta_list->head;
	while (temp1 != NULL)
		{
		temp2 = temp1->next;
		while (temp2 != NULL)
			{                           /*go through beta_list and make pointers*/
			temp3 = temp1->head;        /*for each rkan_node*/
			while(temp3 != NULL)
				{
				temp4 = temp2->head;
				if(temp4 != NULL)
					{
					/*create and allocate memory for a new t_node*/

					t_node = (struct b_node*)malloc(sizeof(struct b_node));
					t_node->p_list.head = t_node->p_list.tail = NULL;
					t_node->next = NULL;

					while(temp4 != NULL)
						{
						/*create and allocate memory for a new p_node*/
						/*assign the pointer in the p_node to point to the correct rkan_node*/

						p_node = (struct pointer_node*)malloc(sizeof(struct pointer_node));
						p_node->p = temp4;
						p_node->next = NULL;

						/*add newly created p_node to t_node->p_list*/

						if(t_node->p_list.head == NULL)
							{
							t_node->p_list.head = t_node->p_list.tail = p_node;
							}
						else
							{
							t_node->p_list.tail->next = p_node;
							t_node->p_list.tail = p_node;
							}
						temp4 = temp4->next;
						}

						/*add t_node to rkan_node list*/

						if(temp3->list.head == NULL)
							{
							temp3->list.head = temp3->list.tail = t_node;
							}
						else
							{
							temp3->list.tail->next = t_node;
							temp3->list.tail = t_node;
							}
					}
					temp3 = temp3->next;
				}
			temp2 = temp2->next;
			}
		temp1 = temp1->next;
		}
}


/* find_p1
Purpose:
	This function finds all the paths out of the first beta 
	and calls find_p2 to compare this path to all paths out
	of the second beta.
*/

void    find_p1(int beta[MAXWORD], struct nodetype *pointer1, struct nodetype *pointer2, struct functor_node *F, struct rkan_list *temp1, struct rkan_list *temp2)
{
struct nodetype *temp, *gammapointer;
int i, newpath[MAXWORD];
bool done;

	/*go through tree of paths and get all paths*/

	temp = pointer1;
	if(temp->data != -1)
		{
		for(i=path_len(beta)+1;i>0;i--)
			{
			beta[i] = beta[i-1];
			}
		beta[0] = temp->data;
		}
	if (beta[0] == -2)
		{
		beta[0] = -1;
		}
	/*if we have a path then find all of the paths from the second beta and
	test each one with the path just found*/

	if (temp->tree != NULL)
		{
		gammapointer = temp->tree;
		newpath[0] = -1;
		find_p2(newpath,pointer2,F,gammapointer,beta,temp1,temp2);
		}
	if (temp->down != NULL)
		{
		find_p1(beta,temp->down,pointer2,F,temp1,temp2);
		}
	for(i=0;i<=path_len(beta)-1;i++)
		{
		beta[i] = beta[i+1];
		}
	if (temp->flag)
		{
		find_p1(beta,temp->across,pointer2,F,temp1,temp2);
		}
}

/* remove_pointer
Purpose:    
	This function removes the pointer_node which contains 
	a pointer equal to temp4 from a pointer_list
	of the rkan_node pointed to by temp3 .
*/

void    remove_pointer(struct rkan_node *temp3, struct rkan_node *temp4)
{
struct b_node *temp;
struct pointer_node *temp1, *temp2;

	temp = temp3->list.head;
	while(temp != NULL)
		{
		if(temp->p_list.head != NULL)
			{
			temp1 = temp->p_list.head;
			temp2 = temp->p_list.head;
			if(temp1->p == temp4)   /*if first pointer in list is pointer to be removed*/
				{
				if(temp->p_list.tail == temp->p_list.head)      /*if pointer to be removed is only pointer in list*/
					{
					temp->p_list.head = temp->p_list.tail = NULL;
					}
				else
					{
					temp->p_list.head = temp->p_list.head->next;
					}
				free(temp1);
				return;
				}
			else
				{
				while((temp1->p != temp4) && (temp1->next != NULL))
					{
					temp1 = temp1->next;
					if(temp1->p == temp4)
						{
						temp2->next = temp1->next;
						if(temp1 == temp->p_list.tail)
							{
							temp->p_list.tail = temp2;
							}
						free(temp1);
						return;
						}
					temp2 = temp2->next;
					}
				}
			}
		temp = temp->next;
		}
}


/* find_p2
Purpose:
	This function finds all of the paths out of the second beta 
	and tests each path with the path pointed to by the gammapointer. 
	If the equations for the paths don't equal each
	other then it removes the relevent pointer from the beta_list. 
*/

void    find_p2(int newpath[MAXWORD], struct nodetype *pointer, struct functor_node *F, struct nodetype *gammapointer, int beta[MAXWORD], struct rkan_list *temp1, struct rkan_list *temp2)
{
struct nodetype *temp, *gammapointer2;
struct rkan_node *temp3, *temp4;
int i;
bool done;

	temp = pointer;
	if(temp->data != -1)
		{
		for(i=path_len(newpath)+1;i>0;i--)
			{
			newpath[i] = newpath[i-1];
			}
		newpath[0] = temp->data;
		}
	if (newpath[0] == -2)
		{
		newpath[0] = -1;
		}
	if (temp->tree != NULL)
		{
		gammapointer2 = temp->tree;

		/*if the paths have the same object as their codomain then test them*/

		if (gammapointer2->data == gammapointer->data)
			{
			temp3 = temp1->head;
			while (temp3 != NULL)
				{
				temp4 = temp2->head;
				while (temp4 != NULL)
					{

					/*if the equations don't equal each other then remove the relevent pointer from
					the beta_list*/

					if (!((function_value(beta,F->f,temp3->a_beta) == function_value(newpath,F->f,temp4->a_beta))))
						{
						remove_pointer(temp3,temp4);
						}
					temp4 = temp4->next;
					}
				temp3 = temp3->next;
				}
			}
		}
	if (temp->down != NULL)
		{
		find_p2(newpath,temp->down,F,gammapointer,beta,temp1,temp2);
		}
	for(i=0;i<=path_len(newpath)-1;i++)
		{
		newpath[i] = newpath[i+1];
		}
	if (temp->flag)
		{
		find_p2(newpath,temp->across,F,gammapointer,beta,temp1,temp2);
		}
}


/* check_pointers 
Purpose:
	This function goes through the beta_list and removes all 
	of the pointers for which the equations don't hold for 
	their rkan_nodes. 
*/

void check_pointers(struct big_list *beta_list, struct functor_node *F)
{
struct rkan_list *temp1, *temp2;
int beta[MAXWORD];

	temp1 = beta_list->head;
	while (temp1 != NULL)
		{
		beta[0] = -1;
		if (temp1->paths != NULL)
			{
			test_self(beta,temp1->paths,F,temp1,beta_list,temp1->paths);
			}
		temp2 = temp1->next;
		while (temp2 != NULL)
			{
			beta[0] = -1;
			if (temp1->paths != NULL)
				{
				if(temp2->paths != NULL)
					{
					find_p1(beta,temp1->paths,temp2->paths,F,temp1,temp2);
					}
				}
			temp2 = temp2->next;
			}
		temp1 = temp1->next;
		}
}


/* num_betas
Purpose:
	To return the number of betas in the beta_list. 
*/

int     num_betas(struct big_list *beta_list)
{
int num;
struct rkan_list *temp_pointer;

	num = 0;
	temp_pointer = beta_list->head;
	while(temp_pointer != NULL)
		{
		num++;
		temp_pointer = temp_pointer->next;
		}
	return(num);
}


/* find_RofB
Purpose:
	This function goes through the beta_tree and examines the 
	pointers to see which tuples belong in the R(B) tree.
	It also prints the tuples in R(B), and calls the function 
	display_images to print the values of R(f) for each arrow f 
	that has domain B. 
*/

int find_RofB(struct rkan_node *node_pointer, int path[MAXWORD], int counter,
	int num_betas, struct nodetype *tree, struct big_list *beta_list,
	int num_printed, struct nodetype* product_tree[MAXWORD],
	int maxes[MAXWORD], int dom, struct functor_node *F, struct functor_node *X,
	int maxloop)
{
struct pointer_node *temp1, *temp4;
struct listnodetype *list;
struct nodetype *p;
struct rkan_node *n_pointer, *temp2;
struct rkan_list *temp;
struct b_node *temp3, *temp5;
int i, j, k;
bool done, found;

	list = (struct listnodetype*)malloc(sizeof(struct listnodetype));

	done = false;
	while(!done)
		{
		done = true;
		path[counter+1] = -1;
		path[counter] = node_pointer->a_beta;
		temp = beta_list->head;
		found = true;
		for(j=0;j<counter - 1;j++)
			{
			temp2 = temp->head;
			while(temp2->a_beta != path[j])
				{
				temp2 = temp2->next;
				}
			temp3 = temp2->list.head;
			for(k=1;k<counter-j;k++)
				{
				temp3 = temp3->next;
				}
			temp4 = temp3->p_list.head;
			found = false;
			while(temp4 != NULL)
				{
				if(temp4->p == node_pointer)
					{
					found = true;
					}
				temp4 = temp4->next;
				}
			if(!found)
				{
				break;
				}
			temp = temp->next;
			}
		if(!found)
			{
			path[counter] = -1;
			return(num_printed);
			}
		temp5 = node_pointer->list.head;
		if (temp5 != NULL)
			{
			temp1 = temp5->p_list.head;
			while(temp1 != NULL)
				{
				n_pointer = temp1->p;
				counter++;
				num_printed = find_RofB(n_pointer,path,counter,num_betas,tree,beta_list,num_printed,product_tree,maxes,dom,F,X,maxloop);
				counter--;
				temp1 = temp1->next;
				}
			}

		/*if at bottom level of tree and all the pointers are there then
		add the tuple to the R(B) tree*/

		if((counter == num_betas) && (node_pointer->valid == 1))
			{
			printf("\n");
			if (to_fil) fprintf(outfil,"\n");
			for(i=0;i<path_len(path);i++)
				{
				printf("%3d",path[i]);
				if (to_fil) fprintf(outfil,"%3d",path[i]);
				}
			printf("  ");
			if (to_fil) fprintf(outfil,"  ");

			display_images(product_tree,dom,F,path,maxes,X,maxloop);
			num_printed++;
			if (mod10(num_printed) == 0)
				{
				pause_print();
				}
			make_list(&list,path);
			search_and_add(tree,list,&p);
			}

		if(counter == 0)
			{
			if((node_pointer = node_pointer->next) != NULL)
				{
				done = false;
				}
			}
	}
	return(num_printed);
}





/* test_self */
/*
Purpose:
	Given a beta, goes through all paths out of that beta.  Calls 
	test_self2 to compare each such path with every other one 
	with the same codomain.
	Note: test_self and test_self2 are basically identical to 
	find_p1 and find_p2, except that they test a beta against 
	itself, rather than comparing two betas.
*/

void test_self(int beta[MAXWORD], struct nodetype *pointer1, struct functor_node *F, struct rkan_list *temp1, struct big_list *beta_list, struct nodetype *paths)
{
struct nodetype *temp, *gammapointer;
int i, newpath[MAXWORD];
bool done;

	/*go through tree of paths and get all paths*/

	temp = pointer1;
	if(temp->data != -1)
		{
		for(i=path_len(beta)+1;i>0;i--)
			{
			beta[i] = beta[i-1];
			}
		beta[0] = temp->data;
		}
	if (beta[0] == -2)
		{
		beta[0] = -1;
		}
	/*if we have a path then find all other paths from the beta and
	test each one with the path just found*/

	if ((temp->tree != NULL) /*&& (temp->info == 1)*/)
		{
		gammapointer = temp->tree;
		newpath[0] = -1;
		test_self2(newpath,paths,F,gammapointer,beta,temp1,beta_list);
		}
	if (temp->down != NULL)
		{
		test_self(beta,temp->down,F,temp1,beta_list,paths);
		}
	for(i=0;i<=path_len(beta)-1;i++)
		{
		beta[i] = beta[i+1];
		}
	if (temp->flag)
		{
		test_self(beta,temp->across,F,temp1,beta_list,paths);
		}
}


/* test_self2 */
/*
Purpose:
	Finds a second path out of the beta with the same codomain as 
	the first one (found in test_self).  If any element of SET 
	has different images under any two such paths, tuples 
	containing that element will not be included in the B/F tree.
	Note: test_self and test_self2 are basically identical to 
	find_p1 and find_p2, except that they test a beta against 
	itself, rather than comparing two betas.
*/

void test_self2(int newpath[MAXWORD], struct nodetype *pointer, struct functor_node *F,
					struct nodetype *gammapointer, int beta[MAXWORD], struct rkan_list *temp1,struct big_list *beta_list)
{
struct nodetype *temp, *gammapointer2;
struct rkan_node *temp3, *temp4;
int i;
bool done;
int num;

	temp = pointer;
	if(temp->data != -1)
		{
		for(i=path_len(newpath)+1;i>0;i--)
			{
			newpath[i] = newpath[i-1];
			}
		newpath[0] = temp->data;
		}
	if (newpath[0] == -2)
		{
		newpath[0] = -1;
		}
	if ((temp->tree != NULL) /*&& (temp->info == 1)*/)
		{

		gammapointer2 = temp->tree;

		/*if the paths have the same object as their codomain then test them*/

		if (gammapointer2->data == gammapointer->data)
			{
			temp3 = temp1->head;
			while (temp3 != NULL)
				{

					/*if the equations don't equal each other then remove the relevant pointer from
					the beta_list*/

					if (!((function_value(beta,F->f,temp3->a_beta) == function_value(newpath,F->f,temp3->a_beta))))
						{
						num = num_betas(beta_list);
						if (temp1->next == NULL)
							temp3->valid = 0;
						if ((num == 1))
							{
							remove_entire_node(temp3,temp1);
							}
						else
							{
							remove_all_pointers(temp3);
							}
						}
				temp3 = temp3->next;
				}
			}
		}
	if (temp->down != NULL)
		{
		test_self2(newpath,temp->down,F,gammapointer,beta,temp1,beta_list);
		}
	for(i=0;i<=path_len(newpath)-1;i++)
		{
		newpath[i] = newpath[i+1];
		}
	if (temp->flag)
		{
		test_self2(newpath,temp->across,F,gammapointer,beta,temp1,beta_list);
		}
}



/* remove_all_pointers
Purpose:
  To remove all pointers from a given pointer list.
*/

void remove_all_pointers(struct rkan_node *temp3)
{
struct b_node *temp;

	temp = temp3->list.head;
	while (temp != NULL)
		{
		temp->p_list.head = temp->p_list.tail = NULL;
		temp->next = NULL;
		temp = temp->next;
		}
	free(temp);
}


/* remove_entire_node
Purpose:
	To remove a node entirely from the big_list structure.
*/

void remove_entire_node(struct rkan_node *deleted, struct rkan_list *temp1)
{
struct rkan_node *temp;

	temp = temp1->head;
	if (temp == deleted)
		{
		if (temp1->tail == deleted)
			{
			temp1->head = temp1->tail = NULL;
			}
		else
			{
			temp1->head = temp1->head->next;
			}
		}
	else
		{
		while(temp != NULL)
			{
			if (temp->next == deleted)
				{
				if (temp1->tail == deleted)
					{
					temp1->tail = temp;
					}
				else
					{
					temp->next = deleted->next;
					}
				}
			temp = temp->next;
			}
		}
}


/* pause in the printing of a long list of items, and prompt the user to
	continue */
void pause_print()
{
	printf("\n\nHit <enter> for more...");
	flush();
	getchar();
	printf("\n");
}



/*  display_heading
Purpose:
	To display a heading for the Right Kan output.
	This function displays the heading for the portion of the Right Kan
	output which shows R(B) for an object B, and rho(A) for each 
	object A with B=FA.
*/

void display_heading(int count, char idarray[MAXWORD], bool zero, struct functor_node *X, char obj)
{
int k,i;
FILE *f;

	for (i=0; i<2; i++)
		{
		if (i==0) f = stdout;
		if (i==1)
			{
			if (to_fil)
				{f = outfil;}
			else
				{return;}
			}

		fprintf(f,"\n\n");

		fprintf(f," R(%c)",obj);
		for (k=2; k<count; k++)
			fprintf(f,"   ");
		fprintf(f,"    ");
		for (k=1; k<=idarray[0]; k++)
			fprintf(f," rho(%c)",idarray[k]);

		fprintf(f,"\n ~~~~");
		for (k=2; k<count; k++)
			fprintf(f,"   ");
		fprintf(f,"    ");
		for (k=1; k<=idarray[0]; k++)
			fprintf(f," ~~~~~~");
		fprintf(f,"\n");

		if (zero) count = 0;

		if (zero) fprintf(f,"   ");

		for (k=1; k<count; k++)
			fprintf(f," a%d",k);
		fprintf(f,"    ");

		fprintf(f,"\n");


		for (k=1; k<count; k++)
			fprintf(f," ~~");
		fprintf(f,"    ");

		fprintf(f,"\n");

		if (zero)
			{
			fprintf(f,"  0     ");
			for (k=1; k<=idarray[0]; k++)
				{
				if (X->f.obj[object_number(idarray[k],X->A)] == 0)
					{fprintf(f,"     1 ");}
				else
					{fprintf(f,"     ! ");}
				}
			fprintf(f,"\n");
			}
		}
}



/*  display_heading2
Purpose:
	To display a heading for the Right Kan output.
	This function displays the portion of the Right Kan output 
	which shows R(B) for an object B, and R(f) for all generating 
	arrows out of B. 
*/

void display_heading2(int count[MAXWORD], int num_arr, int arrs[MAXARR], category B, int obj, bool zero, bool newzero[MAXWORD])
{
int i,j,k,m;
char let;
FILE *f;

	if (zero) count[0] = 1;
	for (k=1; k<=num_arr; k++)
		if (newzero[k-1]) count[k] = 1;

	for (m=0; m<2; m++)
		{
		if (m==0) f = stdout;
		if (m==1)
			{
			if (to_fil)
				{f = outfil;}
			else
				{return;}
			}

		fprintf(f,"\n\n");

		fprintf(f,"     ");
		for (k=2; k<count[0]; k++)
			fprintf(f,"   ");
		fprintf(f,"    ");
		for (i=1; i<=num_arr; i++)
			{
			fprintf(f," R(%c)",B.arrow[arrs[i-1]]);
			if (!zero)
				{
				for (j=1; j<(count[i]-1); j++)
					fprintf(f,"   ");
				}
			fprintf(f,"  ");
			}
		fprintf(f,"\n");


		fprintf(f," R(%c)",B.obj[obj]);
		for (k=2; k<count[0]; k++)
			fprintf(f,"   ");
		fprintf(f,"    ");

		for (i=1; i<=num_arr; i++)
			{
			fprintf(f," R(%c)",B.obj[B.arr[arrs[i-1]][1]]);
			if (!zero)
				{
				for (j=1; j<(count[i]-1); j++)
					fprintf(f,"   ");
				}
			fprintf(f,"  ");
			}
		fprintf(f,"\n");
		fprintf(f," ~~~~");
		for (k=2; k<count[0] ;k++)
			fprintf(f,"   ");
		fprintf(f,"    ");

		for (i=1; i<=num_arr; i++)
			{
			fprintf(f," ~~~~");
			if (!zero)
				{
				for (j=1; j<(count[i]-1); j++)
					fprintf(f,"   ");
				}
			fprintf(f,"  ");
			}
		fprintf(f,"\n");


		if (zero) fprintf(f,"   ");
		for (k=1; k<count[0]; k++)
			fprintf(f," a%d",k);
		fprintf(f,"      ");

		let = 'a';
		for (i=1; i<=num_arr; i++)
			{
			let = let + 'b' - 'a';
			if ((count[i] == 1) || (zero))
				{
				fprintf(f,"   ");
				}
			else
				{
				for (j=1; j<count[i]; j++)
					fprintf(f," %c%d",let,j);
				}
			fprintf(f,"    ");
			}
		fprintf(f,"\n");


		for (k=1; k<count[0]; k++)
			fprintf(f," ~~");
		fprintf(f,"      ");

		for (i=1; i<=num_arr; i++)
			{
			if ((!zero) && (count[i] != 1))
				{
				fprintf(f," ~~");
				for (j=2; j<count[i]; j++)
					fprintf(f," ~~");
				}
			else
				{
				fprintf(f,"   ");
			}
			fprintf(f,"    ");
			}
		fprintf(f,"\n");

		if (zero)
			{
			fprintf(f,"  0    ");
			for (i=1; i<=num_arr; i++)
				{
				if (newzero[i-1])
					{fprintf(f,"    1  ");}
				else
					{fprintf(f,"    !  ");}
				}
			fprintf(f,"\n");
			}
		}
}




/* right_kan */
/*
Purpose:
	This function gets the necessary input from the user and then calls
	all other functions to do the computations necessary to determine
	a Right Kan extension.
*/

void right_kan(category_list c_list, functor_list f_list, int maxloop)
{
bool finished;
char name[MAXWORD];
int num, i;
struct category_node *catptrA, *catptrB;
struct functor_node *functorptrAB, *functorptrASET;
int obj,levels;
int path[MAXWORD];
struct nodetype* B_tree[MAXWORD];
struct nodetype* product_tree[MAXWORD];
struct nodetype *RofB_tree, *RB_rho_tree;
int count[MAXWORD];
int newobj;
bool success;
int ids, dummy;
char idarray[MAXWORD], dummy1[MAXWORD];
int rhoarr[MAXWORD], dummy2[MAXWORD];
int objects[MAXWORD];
struct big_list beta_list;
int arrs[MAXARR];
int num_arr;
int q;
char let;
int num_printed;
bool zero;
bool newzero[MAXWORD];
char fname[MAXWORD];

	catptrA = (struct category_node*)malloc(sizeof(struct category_node));
	catptrB = (struct category_node*)malloc(sizeof(struct category_node));
	functorptrAB = (struct functor_node*)malloc(sizeof(struct functor_node));
	functorptrASET = (struct functor_node*)malloc(sizeof(struct functor_node));


	clearscreen();
	finished = false;

	/*get user to choose categories A and B*/
	catptrA = get_cat('A',c_list);
	if (catptrA == NULL) return;
	catptrB = get_cat('B',c_list);
	if (catptrB == NULL) return;

	/*get user to choose functor from A to B and make sure that it really is
	  from A to B*/
	functorptrAB = get_fun(f_list,catptrA,catptrB);
	if (functorptrAB == NULL) return;

	/*get user to enter functor from A to SET and make sure it really
	is from A to SET*/
	functorptrASET = get_sf(f_list,catptrA);
	if (functorptrASET == NULL) return;

	printf("\n\n(Hit <enter> if output not to be saved)\n");
	printf("Name of output file: ");
	flush();
	gets(fname);
	to_fil = (fname[0] != '\0');
	if (to_fil)
		outfil = fopen(fname,"w");

	clearscreen();
	printf("Making the sets of relations confluent.....\n\n\n");
	printf("Category A...");
	make_confluent(&(catptrA->cat));
	printf("\n\nCategory B...");
	make_confluent(&(catptrB->cat));
	printf("\n\n");

	printf("\nHit <enter> to continue...");
	flush();
	getchar();
	clearscreen();

	/*loop through objects of category B, form B/F, and compute
	  the Right Kan extension for that object*/

	for (obj=0;obj<catptrB->cat.numobj;obj++)
		{
		B_tree[obj] = (struct nodetype*)malloc(sizeof(struct nodetype));
		B_tree[obj] = create_tree();
	product_tree[obj] = (struct nodetype*)malloc(sizeof(struct nodetype));
		product_tree[obj] = create_tree();

		/*put objects in B/F tree*/
		path[0] = -1;
		for(i=0;i<functorptrAB->B.numobj;i++)
			{
			objects[i] = 0;
			}
		make_beta_objects(path,obj,objects,B_tree[obj],functorptrAB,maxloop);

		/*put identity in B/F tree*/
		path[0] = -1;
		make_id_beta_object(obj,B_tree[obj],functorptrAB);

		/*put arrows in B/F tree*/
		path[0] = -1;
		make_beta_arrows(B_tree[obj],functorptrAB,B_tree[obj],path,maxloop);

		/*make product tree for B/F category*/
		path[0] = -1;
		get_all_betas(B_tree[obj],path,functorptrAB,product_tree[obj],B_tree[obj],maxloop);
		}

	for (obj=0; obj<catptrB->cat.numobj; obj++)
		{
		/*make list for R(B) tree from product tree*/
		beta_list.head = NULL;
		beta_list.tail = NULL;
		make_big_list(path,product_tree[obj],&beta_list,functorptrASET);

		/*make all pointers and then eliminate ones for which equations
		don't hold*/
		assign_pointers(&beta_list);
		check_pointers(&beta_list,functorptrASET);

		clearscreen();
		path[0] = -1;
		zero = check_zero(product_tree[obj],path,functorptrASET);
		printf("Right Kan Extension ... object %c\n\n",catptrB->cat.obj[obj]);
		if (to_fil) fprintf(outfil,"Right Kan Extension ... object %c\n\n",catptrB->cat.obj[obj]);
		ids = 0;
		idarray[0] = rhoarr[0] = 0;
		path[0] = -1;
		make_rhoarr(product_tree[obj],path,functorptrAB,1,&ids,idarray,rhoarr);
		count[0] = 1;
		if (!zero)
			{
			path[0] = -1;
			count[0] = display_Btree(product_tree[obj],path,functorptrAB,1,'a');
			if (count[0] == 1)
				{
				printf("No betas for object %c",catptrB->cat.obj[obj]);
				if (to_fil) fprintf(outfil,"No betas for object %c",catptrB->cat.obj[obj]);
				}
			printf("\n");
			if (to_fil) fprintf(outfil,"\n");
			}
		display_heading(count[0],idarray,zero,functorptrASET,catptrB->cat.obj[obj]);
		if (!zero)
			{
			RB_rho_tree = (struct nodetype*)malloc(sizeof(struct nodetype));
			RB_rho_tree = create_tree();
			path[0] = -1;
			levels = num_betas(&beta_list) - 1;
			if (levels == -1)
				{
				printf("  1\n");
				if (to_fil) fprintf(outfil,"  1\n");
				}
			else
				{
				num_printed = display_RB_rho(beta_list.head->head,path,0,levels,RB_rho_tree,&beta_list,0,rhoarr);
				if (num_printed == 0)
					{
					printf("     0");
					if (to_fil) fprintf(outfil,"     0");
					}
				}
			}
		flush();
		getchar();
		clearscreen();

		printf("Right Kan Extension ... object %c\n\n",catptrB->cat.obj[obj]);
		printf("R(%c):  ",catptrB->cat.obj[obj]);
		if (zero) printf("empty\n");
		if (to_fil)
			{
			fprintf(outfil,"\n\n\n------------------------------------------\n\n");
			fprintf(outfil,"Right Kan Extension ... object %c\n\n",catptrB->cat.obj[obj]);
			fprintf(outfil,"R(%c):  ",catptrB->cat.obj[obj]);
			if (zero) printf("empty\n");
			}
		count[0] = 1;
		if (!zero)
			{
			path[0] = -1;
			count[0] = display_Btree(product_tree[obj],path,functorptrAB,1,'a');
			if (count[0] == 1)
				{
				printf("No betas for object %c",catptrB->cat.obj[obj]);
				if (to_fil) fprintf(outfil,"No betas for object %c",catptrB->cat.obj[obj]);
				}
			printf("\n");
			if (to_fil) fprintf(outfil,"\n");
			}
		num_arr = all_arr(&(catptrB->cat),obj,arrs);
		let = 'a';
		for (q=0; q<num_arr; q++)
			{
			let = let + 'b' - 'a';
			newobj = catptrB->cat.arr[arrs[q]][1];
			printf("R(%c):  ",catptrB->cat.obj[newobj]);
			if (to_fil) fprintf(outfil,"R(%c):  ",catptrB->cat.obj[newobj]);
			path[0] = -1;
			newzero[q] = check_zero(product_tree[newobj],path,functorptrASET);
			if (newzero[q])
				{
				printf("empty\n");
				if (to_fil) fprintf(outfil,"empty\n");
				}
			count[q+1] = 1;
			if (!newzero[q])
				{
				path[0] = -1;
				count[q+1] = display_Btree(product_tree[newobj],path,functorptrAB,1,let);
				if (count[q+1] == 1)
					{
					printf("No betas for object %c",catptrB->cat.obj[newobj]);
					if (to_fil) fprintf(outfil,"No betas for object %c",catptrB->cat.obj[newobj]);
					}
				printf("\n");
				if (to_fil) fprintf(outfil,"\n");
				}
			}
		display_heading2(count,num_arr,arrs,catptrB->cat,obj,zero,newzero);

		if (!zero)
			{
			/*make R(B) tree from the list*/
			RofB_tree = (struct nodetype*)malloc(sizeof(struct nodetype));
			RofB_tree = create_tree();
			path[0] = -1;
			levels = num_betas(&beta_list)-1;
			if (levels == -1)
				{
				printf("\n  1      ");
				if (to_fil) fprintf(outfil,"\n  1      ");
				for (q=0; q<num_arr; q++)
					{
					printf("  1    ");
					if (to_fil) fprintf(outfil,"  1    ");
					}
				}
			else
				{
				num_printed = find_RofB(beta_list.head->head,path,0,levels,RofB_tree,&beta_list,0,product_tree,count,obj,functorptrAB,functorptrASET,maxloop);
				if (num_printed == 0)
					{
					printf("     0");
					if (to_fil) fprintf(outfil,"     0");
					for (q=0; q<num_arr;q++)
						{
						printf("  !    ");
						if (to_fil) fprintf(outfil,"  !    ");
						}
					}
				}
			}

		if (to_fil) fprintf(outfil,"\n\n\n------------------------------------------\n\n");

/*  free_big_list is a function to deallocate memory used
	by the big_list.  It is presently not working. */

		free_big_list(&beta_list);
		free(beta_list.head);
		free(beta_list.tail);
		beta_list.head = beta_list.tail = NULL;
		flush();
		getchar();



/*              free_tree(RofB_tree);
		free_tree(RB_rho_tree);*/
		}

/*              for (obj=0; obj<catptrB->cat.numobj; obj++)
			{
			free_tree(B_tree[obj]);
			free_tree(product_tree[obj]);
			}                    */

	if (to_fil)
		fclose(outfil);
}


/* get_cat
Purpose:
	This function gets user input for a category
	to be used in a Right Kan extension. 
*/

struct category_node *get_cat(char ch,category_list c_list)
{
bool finished;
char name[MAXWORD];
int num;
struct category_node *catptr;

	catptr = (struct category_node*)malloc(sizeof(struct category_node));

	finished = false;
	while (!finished)
		{
		display_cat_list(c_list);
		printf("(-1 to cancel)\n");
		printf("Category %c> ",ch);
		flush();
		gets(name);
		if (strcmp(name,"-1") == 0)
			return(NULL);
		num = atoi(name);
		if ((num >= 1) && (num <= num_cat(c_list)))
			{
			finished = true;
			catptr = find_category_number(c_list,num-1);
			}
		}
	return (catptr);
}


/* get_fun
Purpose:
	This function gets user input for a functor
	to be used in a Right Kan extension. 
*/

struct functor_node *get_fun(functor_list f_list,struct category_node *catptrA,struct category_node *catptrB)
{
bool finished;
struct functor_node *functorptrAB;
char name[MAXWORD];
int num;

	functorptrAB = (struct functor_node*)malloc(sizeof(struct functor_node));

	finished = false;
	while (!finished)
		{
		display_functor_list(f_list);
		printf("(-1 to cancel)\n");
		printf("Functor from %s to ",catptrA->cat.name);
		printf("%s > ",catptrB->cat.name);
		flush();
		gets(name);
		if (strcmp(name,"-1") == 0)
			return(NULL);
		num = atoi(name);
		if ((num >= 1) && (num <= num_functors(f_list)))
			{
			functorptrAB = find_functor_number(f_list,num-1);
			if((strcmp(functorptrAB->A.name,catptrA->cat.name) == 0) && (strcmp(functorptrAB->B.name,catptrB->cat.name) == 0))
				{
				finished = true;
				}
			else
				{
				printf("\n\007%s is not a functor from",functorptrAB->f.name);
				printf(" %s to",catptrA->cat.name);
				printf(" %s.\n",catptrB->cat.name);
				printf("Hit <enter> to re-enter...");
				flush();
				getchar();
				}
			}
		}
	return(functorptrAB);
}



/*  get_sf
Purpose: 
	This function gets user input for a set functor
	to be used in a Right Kan extension. 
*/

struct functor_node *get_sf(functor_list f_list,struct category_node *catptrA)
{
bool finished;
struct functor_node *functorptrASET;
char name[MAXWORD];
int num;

	functorptrASET = (struct functor_node*)malloc(sizeof(struct functor_node));

	finished = false;
	while (!finished)
		{
		display_functor_list(f_list);
		printf("(-1 to cancel)\n");
		printf("Functor from %s to set> ",catptrA->cat.name);
		flush();
		gets(name);
		if (strcmp(name,"-1") == 0)
			return(NULL);
		num = atoi(name);
		if ((num >= 1) && (num <= num_functors(f_list)))
			{
			functorptrASET = find_functor_number(f_list,num-1);
			if((strcmp(functorptrASET->A.name,catptrA->cat.name) == 0) && (functorptrASET->functor_type == 1))
				{
				finished = true;
				}
			else
				{
				printf("\n\007%s is not a functor from",functorptrASET->f.name);
				printf(" %s to SET.\n",catptrA->cat.name);
				printf("Hit <enter> to re-enter...");
				flush();
				getchar();
				}
			}
		}
	return(functorptrASET);
}







/*  display_images
Purpose:
	This function is called by find_RofB.
	It is used to display the action of R on all arrows out of the
	object B.
*/

void display_images(struct nodetype* product_tree[MAXWORD], int dom, struct functor_node *F, int tuple[MAXWORD], int num_betas[MAXWORD], struct functor_node *X, int maxloop)
{
int num_arr, i, cod;
int arr[MAXARR];
int beta[MAXWORD];

	num_arr = all_arr(&(F->B),dom,arr);
	for (i=0; i<num_arr; i++)
		{
		printf("    ");
		if (to_fil) fprintf(outfil,"    ");
		cod = F->B.arr[arr[i]][1];
		beta[0] = -1;
		if ((num_betas[i+1] == 1) || (num_betas[0] == 1))
			{
			printf("  1");
			if (to_fil) fprintf(outfil,"  1");
			}
		else
			{
			traverse_cod_tree(product_tree[cod],beta,F,product_tree[dom],arr[i],tuple,X,maxloop);
			}
		}
}




/*  traverse_cod_tree
Purpose:
	This function goes through "tree" and looks at each beta object
	in this tree.  (These are the betas for the object which is
	the codomain of the arrow f for which we are caluclating R(f)).
	It then calls traverse_dom_tree to go through the betas in
	the tree for the domain.
*/

void traverse_cod_tree(struct nodetype *tree, int beta[MAXWORD],struct functor_node* F,
	struct nodetype *dom_tree, int arr, int tuple[MAXWORD], struct functor_node *X,
	int maxloop)
{
struct nodetype* temp;
int i;
int path[MAXWORD], newbeta[MAXWORD];
struct nodetype *pointer;
bool done;
bool found;
int success;
int dest;
int num;

	temp = tree;
	if(temp->data != -1)
		{
		for(i=path_len(beta)+1;i>0;i--)
			{
			beta[i] = beta[i-1];
			}
		beta[0] = temp->data;
		}
		if (beta[0] == -2)
			{
			beta[0] = -1;
			}
	if (temp->tree != NULL)
		{
		pointer = temp->tree->down;
		done = false;
		while(!done)
			{
			if(pointer->tree != NULL)
				{
				path[0] = -1;
				copy_path(newbeta,beta);
				found = false;
				dest = pointer->data;
				num = traverse_dom_tree(dom_tree,path,F,arr,newbeta,tuple,0,&found,dest);
				if (!found)
					{
					num = alt_dom_tree(dom_tree,path,F,X,arr,newbeta,pointer->data,tuple,0,&found,maxloop);
					if (!found)
						{
						printf("  ?");
						if (to_fil) fprintf(outfil,"  ?");
						}
					}
				}
			if(pointer->flag != 0)
				{
				pointer = pointer->across;
				}
			else
				{
				done = true;
				}
			}
		}
	if (temp->down != NULL)
		{
		traverse_cod_tree(temp->down,beta,F,dom_tree,arr,tuple,X,maxloop);
		}
	for(i=0;i<=path_len(beta)-1;i++)
		{
		beta[i] = beta[i+1];
		}
	if (temp->flag)
		{
		traverse_cod_tree(temp->across,beta,F,dom_tree,arr,tuple,X,maxloop);
		}
}




/*  traverse_dom_tree
Purpose:
	This function goes through the betas in the tree for the domain
	of an arrow f.  The parameter cod_path is a beta' out of the
	codomain.  If beta'f is found to be equal to any beta in the
	domain tree, then element beta' of the image tuple under the
	arrow is defined to be element beta of the original tuple.
*/

int traverse_dom_tree(struct nodetype *tree, int beta[MAXWORD],struct functor_node* F,
	int arr, int cod_path[MAXWORD], int tuple[MAXWORD], int count,
	bool *found, int dest)
{
struct nodetype* temp;
int i;
int newpath[MAXWORD], new_cod_path[MAXWORD];
int reduced_cod[MAXWORD], reduced_beta[MAXWORD], reduced_new_cod[MAXWORD];
struct nodetype *pointer;
bool done;
int source;

	*found = false;
	temp = tree;
	if(temp->data != -1)
		{
		for(i=path_len(beta)+1;i>0;i--)
			{
			beta[i] = beta[i-1];
			}
		beta[0] = temp->data;
		}
		if (beta[0] == -2)
			{
			beta[0] = -1;
			}
	if (temp->tree != NULL)
		{
		pointer = temp->tree->down;
		done = false;
		while(!done)
			{
			if(pointer->tree != NULL)
				{
				newpath[0] = arr;
				newpath[1] = -1;
				copy_path(new_cod_path,cod_path);
				append_path(new_cod_path,newpath);
				reduce(reduced_cod,cod_path,F->B);
				reduce(reduced_new_cod,new_cod_path,F->B);
				reduce(reduced_beta,beta,F->B);
				source = pointer->data;

				if((equals(reduced_new_cod,reduced_beta)) && (source == dest))
					{
					printf("%3d",tuple[count]);
					if (to_fil) fprintf(outfil,"%3d",tuple[count]);
					*found = true;
					return(count);
					}
				count++;
				}
			if(pointer->flag != 0)
				{
				pointer = pointer->across;
				}
			else
				{
				done = true;
				}
			}
		}
	if (temp->down != NULL)
		{
		if (!(*found))
			{
			count = traverse_dom_tree(temp->down,beta,F,arr,cod_path,tuple,count,found,dest);
			if (*found) return(count);
			}
		}
	for(i=0;i<=path_len(beta)-1;i++)
		{
		beta[i] = beta[i+1];
		}
	if (temp->flag)
		{
		if (!(*found))
			{
			count = traverse_dom_tree(temp->across,beta,F,arr,cod_path,tuple,count,found,dest);
			if (*found) return (count);
			}
		}
	return(count);
}



/*  alt_dom_tree
Purpose:
	This function, given a (beta,A) and a (beta',A') (cod_path) tests to 
	see if there is an arrow alpha from A to A' such that F(alpha)beta
	is equal to beta'f.  If so, then the beta'th element of the 
	image tuple under the arrow f is defined as X(alpha)(b), 
	where b is element beta of the original tuple.
*/

int alt_dom_tree(struct nodetype *tree,int beta[MAXWORD], struct functor_node *F,
	struct functor_node *X, int arr, int cod_path[MAXWORD], int dest,
	int tuple[MAXWORD], int count, bool *found, int maxloop)
{
struct nodetype* temp;
int i;
int newpath[MAXWORD], new_cod_path[MAXWORD];
int reduced_cod[MAXWORD], reduced_beta[MAXWORD], reduced_new_cod[MAXWORD];
struct nodetype *pointer;
bool done;
int source;
int objects[MAXWORD];
int dummy_path[MAXWORD];
bool finished;
int the_path[MAXWORD];
int val;


	*found = false;
	temp = tree;
	if(temp->data != -1)
		{
		for(i=path_len(beta)+1;i>0;i--)
			{
			beta[i] = beta[i-1];
			}
		beta[0] = temp->data;
		}
		if (beta[0] == -2)
			{
			beta[0] = -1;
			}
	if (temp->tree != NULL)
		{
		pointer = temp->tree->down;
		done = false;
		while(!done)
			{
			if(pointer->tree != NULL)
				{
				newpath[0] = arr;
				newpath[1] = -1;
				copy_path(new_cod_path,cod_path);
				append_path(new_cod_path,newpath);
				reduce(reduced_cod,cod_path,F->B);
				reduce(reduced_new_cod,new_cod_path,F->B);
				reduce(reduced_beta,beta,F->B);

				source = pointer->data;

					if ((equals(reduced_new_cod,reduced_beta)) && (source == dest))
						{
						printf("%3d",tuple[count]);
						if (to_fil) fprintf(outfil,"%3d",tuple[count]);
						*found = true;
						return(count);
						}
					else
						{
						if (!(*found))
							{
							dummy_path[0] = -1;
							objects[0] = -1;
							finished = false;
							find_path_from(source,dest,reduced_new_cod,beta,F,dummy_path,source,objects,&finished,maxloop,the_path);
							if (finished)
								{
								val = function_value(the_path,X->f,tuple[count]);
								printf("%3d",val);
								if (to_fil) fprintf(outfil,"%3d",val);
								*found = true;
								return(count);
								}
							}
						}

				count++;
				}
			if(pointer->flag != 0)
				{
				pointer = pointer->across;
				}
			else
				{
				done = true;
				}
			}
		}
	if (temp->down != NULL)
		{
		if (!(*found))
			{
			count = alt_dom_tree(temp->down,beta,F,X,arr,cod_path,dest,tuple,count,found,maxloop);
			if (*found) return(count);
			}
		}
	for(i=0;i<=path_len(beta)-1;i++)
		{
		beta[i] = beta[i+1];
		}
	if (temp->flag)
		{
		if (!(*found))
			{
			count = alt_dom_tree(temp->across,beta,F,X,arr,cod_path,dest,tuple,count,found,maxloop);
			if (*found) return(count);
			}
		}
	return(count);

}





/*  find_path_from
Purpose:
	This function is called by alt_dom_tree, and is used to find
	the desired path alpha described in the purpose of that function.
	This path is returned through the parameter "output".
*/

void find_path_from(int source, int dest, int reduced_cod[MAXWORD], int beta[MAXWORD],
	struct functor_node *F, int path[], int domain, int objects[], bool *found,
	int maxloop, int output[MAXWORD])
{
int i,j, number,current_arrow, codomain;
bool done;
int arrow_array[MAXARR];
int newpath[MAXWORD], pathcopy[MAXWORD], reduced_newpath[MAXWORD];
int alpha[MAXWORD];
int loopcount;

loopcount = 0;
for(i=0;i<path_len(objects);i++)
	{
	if(objects[i] == domain)
		{
		loopcount++;
		objects[path_len(objects)+1] = -1;
		objects[path_len(objects)] = domain;
		if (loopcount == maxloop)
			{
			i = path_len(objects);
			path[0] = -2; path[1] = -1;
			copy_path(output,path);
			return;
			}
		}
	}
	number = all_arr(&(F->A),domain,arrow_array);
	objects[path_len(objects)+1] = -1;
	objects[path_len(objects)] = domain;
	for(i=0;i<number;i++)
		{
		current_arrow = arrow_array[i];
		for((j=path_len(path)+1);j>0;j--)
			{
			path[j] = path[j-1];
			}
		path[0] = current_arrow;
		codomain = F->A.arr[current_arrow][1];
		if (codomain == dest)
			{
			copy_path(pathcopy,path);
			functor_on_path(*F,pathcopy,newpath);
			append_path(newpath,beta);
			reduce(reduced_newpath,newpath,F->B);

			if (equals(reduced_newpath,reduced_cod))
				{
				*found = true;
				copy_path(output,path);
				return;
				}
			}

		if (!(*found))
			{
			copy_path(alpha,path);
			find_path_from(source,dest,reduced_cod,beta,F,alpha,codomain,objects,found,maxloop,path);
			}
		if (*found)
			{
			copy_path(output,path);
			return;
	    }
		if (!(*found) || (path[0] == -2))
			{
			for(j=1;j<=path_len(path);j++)
				{
				path[j-1] = path[j];
				}
			objects[path_len(objects)-1] = -1;
			}
		}

copy_path(output,path);
return;
}









/*  display_RB_rho
Purpose:
	This function is very similar to find_RofB.  It goes through
	the beta_list and examines the pointers to determine which 
	tuples are part of R(B) and should be displayed.  It also
	displays information for the natural transformation rho
	for each object A where B=F(A).
*/

int display_RB_rho(struct rkan_node *node_pointer, int path[MAXWORD], int counter, int num_betas, struct nodetype *tree, struct big_list *beta_list, int num_printed, int rhoarr[MAXWORD])
{
struct pointer_node *temp1, *temp4;
struct listnodetype *list;
struct nodetype *p;
struct rkan_node *n_pointer, *temp2;
struct rkan_list *temp;
struct b_node *temp3, *temp5;
int i, j, k;
bool done, found;

	list = (struct listnodetype*)malloc(sizeof(struct listnodetype));

	done = false;
	while(!done)
		{
		done = true;
		path[counter+1] = -1;
		path[counter] = node_pointer->a_beta;
		temp = beta_list->head;
		found = true;
		for(j=0;j<counter - 1;j++)
			{
			temp2 = temp->head;
			while(temp2->a_beta != path[j])
				{
				temp2 = temp2->next;
				}
			temp3 = temp2->list.head;
			for(k=1;k<counter-j;k++)
				{
				temp3 = temp3->next;
				}
			temp4 = temp3->p_list.head;
			found = false;
			while(temp4 != NULL)
				{
					if(temp4->p == node_pointer)
					{
					found = true;
					}
				temp4 = temp4->next;
				}
			if(!found)
				{
				break;
				}
			temp = temp->next;
			}
		if(!found)
			{
			path[counter] = -1;
			return(num_printed);
			}
		temp5 = node_pointer->list.head;
		if (temp5 != NULL)
			{
			temp1 = temp5->p_list.head;
			while(temp1 != NULL)
				{
				n_pointer = temp1->p;
				counter++;
				num_printed = display_RB_rho(n_pointer,path,counter,num_betas,tree,beta_list,num_printed,rhoarr);
				counter--;
				temp1 = temp1->next;
				}
			}

		/*if at bottom level of tree and all the pointers are there then
		add the tuple to the R(B) tree*/


		if((counter == num_betas) && (node_pointer->valid == 1))
			{
			printf("\n");
			if (to_fil) fprintf(outfil,"\n");
			for(i=0;i<path_len(path);i++)
				{
				printf("%3d",path[i]);
				if (to_fil) fprintf(outfil,"%3d",path[i]);
				}
			printf("    ");
			if (to_fil) fprintf(outfil,"    ");
			for(i=1; i<=rhoarr[0]; i++)
				{
				printf("%7d",path[rhoarr[i] - 1]);
				if (to_fil) fprintf(outfil,"%7d",path[rhoarr[i] -1]);
				}
			num_printed++;
			if (mod10(num_printed) == 0)
				{
				pause_print();
				}
			make_list(&list,path);
			search_and_add(tree,list,&p);
			}

		if(counter == 0)
			{
			if((node_pointer = node_pointer->next) != NULL)
				{
				done = false;
				}
			}
		}
	return(num_printed);
}


/* mod_10
Purpose:
	Returns the remainder when "number" is divided by 10.
*/

int mod10(int number)
{
int temp;

	temp = number;
	while (temp >= 10)
		{
		temp = temp - 10;
		}
	return(temp);
}




/*  check_zero
Purpose:
	This function checks to see if, for any of the (beta,A) in the
	B/F tree, X(A) = 0.  If so, it returns true; otherwise false.
*/

bool check_zero(struct nodetype *tree, int beta[MAXWORD], struct functor_node *X)
{
struct nodetype *temp, *pointer;
int i;
bool done, zero;

	zero = false;
	temp = tree;
	if(temp->data != -1)
		{
		for(i=path_len(beta)+1;i>0;i--)
			{
			beta[i] = beta[i-1];
			}
		beta[0] = temp->data;
		}
		if (beta[0] == -2)
			{
			beta[0] = -1;
			}
	if (temp->tree != NULL)
		{
		pointer = temp->tree->down;
		done = false;
		while(!done)
			{
			if(pointer->tree != NULL)
				{
				if (X->f.obj[pointer->data] == 0)
					return(true);
				}
			if(pointer->flag != 0)
				{
				pointer = pointer->across;
				}
			else
				{
				done = true;
				}
			}
		}
	if (temp->down != NULL)
		{
		zero = check_zero(temp->down,beta,X);
		if (zero) return(true);
		}
	for(i=0;i<=path_len(beta)-1;i++)
		{
		beta[i] = beta[i+1];
		}
	if (temp->flag)
		{
		zero = check_zero(temp->across,beta,X);
		if (zero) return(true);
		}

	return(zero);
}




/* make_rhoarr 
Purpose:
	Given the B/F tree for an object B, this function forms two arrays:
	idarray contains a character representing each object A for which
	B=F(A).
	rhoarr stores, for each object A, which beta in the tree is the
	identity (1,A).
	idarray[0] and rhoarr[0] store the number of objects A.
*/

void make_rhoarr(struct nodetype *tree, int beta[MAXWORD],struct functor_node* F,
	int count, int *ids, char idarray[MAXWORD], int rhoarr[MAXWORD])
{
struct nodetype* temp;
int i,j,k;
int objects[MAXWORD], path[MAXWORD];
listtype list;
struct nodetype *pointer;
bool done;
int arr[MAXARR];
int num_arr;

	temp = tree;
	if(temp->data != -1)
		{
		for(i=path_len(beta)+1;i>0;i--)
			{
			beta[i] = beta[i-1];
			}
		beta[0] = temp->data;
		}
		if (beta[0] == -2)
			{
			beta[0] = -1;
			}
	if (temp->tree != NULL)
		{
		pointer = temp->tree->down;
		done = false;
		while(!done)
			{
			if(pointer->tree != NULL)
				{
				if (beta[0] == -1)
					{
					(*ids)++;
					idarray[*ids] = F->A.obj[pointer->data];
					rhoarr[*ids] = count;
					idarray[0] = rhoarr[0] = *ids;
					}
				count++;
				}
			if(pointer->flag != 0)
				{
				pointer = pointer->across;
				}
			else
				{
				done = true;
				}
			}
		}
	if (temp->down != NULL)
		{
		make_rhoarr(temp->down, beta, F, count, ids, idarray, rhoarr);
		}
	for(i=0;i<=path_len(beta)-1;i++)
		{
		beta[i] = beta[i+1];
		}
	if (temp->flag)
		{
		make_rhoarr(temp->across, beta, F, count, ids, idarray, rhoarr);
		}
}



/******************************************************************/
/*THIS STUFF DOESN'T WORK YET IT'S SUPPOSED TO DEALLOCATE MEMORY USED
  BY BETA_LIST*/
/******************************************************************/

void free_big_list(struct big_list *list)
{
struct rkan_list *temp, *temp2;

	temp = list->head;
	while (temp != NULL)
		{
		free_rkan_list(temp);
		temp2 = temp;
		temp = temp->next;
		list->head = temp;
		free(temp2);
		}
	list->tail = temp;
	free(temp);
}

void free_rkan_list(struct rkan_list *t)
{
struct rkan_node *temp, *temp2;

	temp = t->head;
	while(temp != NULL)
		{
		free_rkan_node(temp);
		temp2 = temp;
		temp = temp->next;
		t->head = temp;
		free(temp2);
		}
	t->tail = temp;
	free(temp);
}

void free_rkan_node(struct rkan_node *t)
{
struct b_node *temp, *temp2;

	temp = t->list.head;
	while(temp != NULL)
		{
		free_b_node(temp);
		temp2 = temp;
		temp = temp->next;
		free(temp2);
		}
	t->list.tail = temp;
	free(temp);
}

void free_b_node(struct b_node *t)
{
struct pointer_node *temp, *temp2;

	temp = t->p_list.head;
	while (temp != NULL)
		{
		temp2 = temp;
		temp = temp->next;
		free(temp2);
		}
	t->p_list.tail = temp;
	free(temp);
}

