/*SUM.C*/

/*        Categories database        */
/* M.Fleming, R.Gunther, R.Rosebrugh */
/* Mount Allison University 1994-95  */

#include        <stdio.h>
#include		<string.h>
#include        "sum.h"

/*
Parameters:
	path:   the current path
	domain: the current object of the path
	codomain:       the object we are trying to find all arrows to
	objects:        an array of integers which keeps the function from exceeding
				the maximum number of loops
	tree:   a pointer to the first node of a tree
	A:      the category being tested
	alpha:  an injection of the sum being tested
	beta:   an injection of the sum being tested
Purpose:
	To generate the reduced hom set from one object to another and check 1-1
Design:
	Creates a tree of trees, with the first tree containing the reduced form
	of all the gamma_alphas and each gamma_alpha having a tree which contains
	all of the gamma_betas paired with it.
*/

bool check_oneone(int path[MAXWORD], int domain, int codomain, int objects[MAXWORD], struct nodetype* tree, category *A,int alpha[MAXWORD],int beta[MAXWORD], struct nodetype* gammatree, int maxloop)
{

/*
	number: stores the number of arrows from an object
	current_arrow:  stores the number of the current arrow in the path
	codom:  stores the codomain of the current arrow
	arrow_array:    contains all of the arrows from an object
	gamma:  an arrow from the object of the sum being tested to codomain
*/

int i,j, number, current_arrow, codom;
int arrow_array[MAXARR];
int gamma[MAXWORD], reduced_gamma[MAXWORD], reduced_gammabeta[MAXWORD], reduced_gammaalpha[MAXWORD];
struct nodetype *temptree, *pointer, *tp;
struct listnodetype *list, *gammalist;

if(!(objects[domain] == maxloop))
	{
	number = all_arr(A,domain,arrow_array); /*get all arrows from current object*/
	for(i=0;i<number;i++)   /*find path for each arrow out of object*/
		{
		objects[domain]++;              /*increment number of times object has been visited*/
		current_arrow = arrow_array[i];
		for((j=path_len(path)+1);j>0;j--)       /*get current arrow and add to path*/
			{
			path[j] = path[j-1];
			}
		path[0]=current_arrow;


		if(((codom = A->arr[current_arrow][1]) == codomain) && (objects[codom] != maxloop))   /*if codomain is destination*/
			{                                                                                               /*object then check 1-1 and store*/
			copy_path(gamma,path);                                                  /*paths in tree of trees*/
			reduce(reduced_gamma,gamma,*A);
			if (path_len(reduced_gamma) == 0)               /* if gamma reduces to identity */
				{
				reduced_gamma[0] = -2;                          /* make a path containing -2 */
				reduced_gamma[1] = -1;
				}
			make_list(&gammalist,reduced_gamma);
			if(!(search_and_add(gammatree,gammalist,&tp)))  /* search for and add path with -2 to gammatree */
				{
				append_path(gamma,alpha);
				reduce(reduced_gammaalpha,gamma,*A);
				if (path_len(reduced_gammaalpha) == 0)  /* if gamma alpha reduces to identity */
					{
					reduced_gammaalpha[0] = -2;             /* make path with -2 to represent identity */
					reduced_gammaalpha[1] = -1;
					}
				copy_path(gamma,path);
				append_path(gamma,beta);
				reduce(reduced_gammabeta,gamma,*A);
				if (path_len(reduced_gammabeta) == 0)   /* if gamma beta reduces to identity */
					{
					reduced_gammabeta[0] = -2;
					reduced_gammabeta[1] = -1;
					}
				make_list(&list,reduced_gammaalpha);    /* make list from reduced gamma alpha */
				search_and_add(tree,list,&pointer);             /* and add it to tree  */
				free_list(list);
				if (pointer->tree == NULL)      /* if no subtree for gamma alpha*/
					{                               /* then create one */
					pointer->tree = create_tree();
					}
				make_list(&list,reduced_gammabeta);     /* make list from reduced gamma beta */
				temptree = create_tree();
				if (search_and_add(pointer->tree,list,&temptree))       /* add it to subtree */
					{
					free_list(list);
					return(false);  /*if pair of paths already in tree, not 1-1*/
					}                               /*exit function and return false*/
				free_list(list);
				}
			free_list(gammalist);
			}

			/*keep going along path with codomain of current arrow as new domain*/

			if(!(check_oneone(path,codom,codomain,objects,tree,A,alpha,beta,gammatree,maxloop)))
				{
				return(false);  /*if not 1-1 exit function*/
				}
			for(j=1;j<=path_len(path);j++)
				{
				path[j-1] = path[j];    /*arrow no longer in path, remove it*/
				}
			objects[domain]--;      /*object no longer part of path, decrement number*/
		}                       /*of times it has been visited*/
	}
	return(true);   /*1-1, return true*/
}


/*
Parameters:
	path:   the current path
	domain: the current object of the path
	codomain:       the object we are trying to find all arrows to
	objects:        an array of integers which keeps the function from exceeding
				the maximum number of loops
	tree:   a pointer to the first node of a tree
	A:      the category being tested
Purpose:
	To generate the reduced hom set from one object to another
Design:
	Creates a tree of all the reduced paths from one object to another.
*/



void make_rhotree(int path[], int domain, int codomain, int objects[],struct nodetype* tree, category *A, int maxloop)
{

/*
	number: stores the number of arrows from an object
	current_arrow:  stores the number of the current arrow in the path
	codom:  stores the codomain of the current arrow
	arrow_array:    contains all of the arrows from an object
*/

int i,j, number, current_arrow, codom;
int arrow_array[MAXARR], reducedpath[MAXWORD];
struct nodetype  *pointer;
struct listnodetype *list;

if(!(objects[domain] == maxloop))
	{
	number = all_arr(A,domain,arrow_array); /*get all arrows from current object*/
	for(i=0;i<number;i++)   /*find path for each arrow out of object*/
		{
		objects[domain]++;              /*increment number of times object has been visited*/
		current_arrow = arrow_array[i];
		for((j=path_len(path)+1);j>0;j--)       /*get current arrow and add to path*/
			{
			path[j] = path[j-1];
			}
		path[0]=current_arrow;
		if(((codom = A->arr[current_arrow][1]) == codomain) && (objects[codom] != maxloop))   /*if codomain is destination*/
			{                                               /*object then check 1-1 and store*/
			reduce(reducedpath,path,*A);
			if (path_len(reducedpath) == 0) /* if path reduces to identity */
				{
				reducedpath[0] = -2;    /* make path with -2 to represent identity */
				reducedpath[1] = -1;
				}
			make_list(&list,reducedpath);           /* make list with reducedpath and */
			search_and_add(tree,list,&pointer); /* add it to the tree */
			free_list(list);
			}

			/*keep going along path with codomain of current arrow as new domain*/

			make_rhotree(path,codom,codomain,objects,tree,A,maxloop);
			for(j=1;j<=path_len(path);j++)
				{
				path[j-1] = path[j];    /*arrow no longer in path, remove it*/
				}
			objects[domain]--;      /*object no longer part of path, decrement number*/
		}                                               /*of times it has been visited*/
	}
}


/*
Parameters:
	path:   the current path
	domain: the current object of the path
	codomain:       the object we are trying to find all arrows to
	objects:        an array of integers which keeps the function from exceeding
				the maximum number of loops
	tree:   a pointer to the first node of a tree
	A:      the category being tested
	rho_tree:       a tree of all the rhos
Purpose:
	To check that a sum is onto
Design:
	Generates all of the lambdas from one object to another and uses the rho tree
	it is passed to check that all of the pairs <lambda,rho> are in our tree.
*/

bool check_onto(int path[], int domain, int codomain, int objects[],struct nodetype* tree, category *A, struct nodetype* rhotree, int maxloop)
{

/*
	number: stores the number of arrows from an object
	current_arrow:  stores the number of the current arrow in the path
	codom:  stores the codomain of the current arrow
	arrow_array:    contains all of the arrows from an object
*/

int i,j, number, current_arrow, codom;
int arrow_array[MAXARR], reducedpath[MAXWORD];
struct nodetype *pointer;
struct listnodetype *list;

if(!(objects[domain] == maxloop))
	{
	number = all_arr(A,domain,arrow_array); /*get all arrows from current object*/
	for(i=0;i<number;i++)   /*find path for each arrow out of object*/
		{
		objects[domain]++;              /*increment number of times object has been visited*/
		current_arrow = arrow_array[i];

		for((j=path_len(path)+1);j>0;j--)       /*get current arrow and add to path*/
			{
			path[j] = path[j-1];
			}
		path[0]=current_arrow;
		if(((codom = A->arr[current_arrow][1]) == codomain)     && (objects[codom] != maxloop)) /*if codomain is destination*/
			{                                                                                               /*object then check that lambda-rho in tree*/
			reduce(reducedpath,path,*A);
			if (path_len(reducedpath) == 0) /* if path reduces to identity */
				{
				reducedpath[0] = -2;            /* make path with -2 to represent identity */
				reducedpath[1] = -1;
				}
			make_list(&list,reducedpath);   /* make list with reducedpath */
			if(!(search_and_add(tree,list,&pointer)))       /* search for it in tree */
				{
				free_list(list);
				return(false);                                  /* if not in tree, not onto */
				}
			else
				{
				free_list(list);
				if(!(equal_trees(pointer->tree,rhotree)))       /* check if rhotree equals */
					{                                                                               /* subtree of reducedpath in tree */
					return(false);
					}
				}
			}

			/*keep going along path with codomain of current arrow as new domain*/

			if(!(check_onto(path,codom,codomain,objects,tree,A,rhotree,maxloop)))
				{
				return(false);  /*if not 1-1 exit function*/
				}
			for(j=1;j<=path_len(path);j++)
				{
				path[j-1] = path[j];    /*arrow no longer in path, remove it*/
				}
			objects[domain]--;      /*object no longer part of path, decrement number*/
		}                                               /*of times it has been visited*/
	}
	return(true);   /*1-1, return true*/
}

/*
Parameters:
	A:      pointer to the category being tested
	obj:    the object of the sum being tested
	alpha:  an injection of the sum
	beta:   an injection of the sum
Purpose:
	To check all objects in the category for 1-1 and onto.
Design:
	Returns true if object is a sum, and false if not.
*/

bool is_sum(category *A, int obj, int alpha[MAXWORD], int beta[MAXWORD], int maxloop)
{
/*
	path:   keeps track of the current path
	objects:        array of integers storing the number of times each object has
				been visited in current path
	id:     a path that has a value of -2 and represents the identity arrow
	T:      tree of gamma alphas and gamma betas
	rhotree:        tree of all rhos
	p, tp:  temporary pointers
	gammatree:      tree of all gammas
	l,  list:       temporary lists
*/

int i,j;
int path[MAXWORD], objects[MAXWORD], id[MAXWORD], dom;
struct nodetype *T, *rhotree, *p, *tp, *gammatree;
struct listnodetype* l,*list, *gammalist;


	for (i=0;i<A->numobj;i++)               /*test for each object in category*/
		{
		path[0] = -1;
		for (j=0;j<A->numobj;j++)
			{                       /*initialize path and objects*/
			objects[j] = 0;
			}
		id[0] = -2;
		id[1] = -1;
		T = create_tree();
		gammatree = create_tree();
		if(i == obj)                            /*if destination obj. equals test obj.*/
			{
			make_list(&list,alpha);
			search_and_add(T,list,&p);
			free_list(list);                /* identity is a gamma so add */
			p->tree = create_tree();                /* alpha and beta to T */
			make_list(&list,beta);
			search_and_add(p->tree,list,&tp);
			free_list(list);
			make_list(&gammalist,id);               /* add identity to tree */
			search_and_add(gammatree,gammalist,&p); /* of gammas */
			}

		if(!(check_oneone(path,obj,i,objects,T,A,alpha,beta,gammatree,maxloop)))
			{
			free_tree(gammatree);
			return(false);
			}
		else
			{
			free_tree(gammatree);
			path[0] = -1;                           /* re-initialize path and objects */
			for (j=0;j<A->numobj;j++)
				{
				objects[j] = 0;
				}
			rhotree = create_tree();
			if(path_len(beta) == 0)         /* if beta is the identity */
				{
				dom = obj;                              /* domain of beta is test object */
				}
			else
				{
				dom = A->arr[beta[path_len(beta)-1]][0]; /* domain of beta is domain */
									 /* of first arrow in path   */
				}
			make_rhotree(path,dom,i,objects,rhotree,A,maxloop);  /* store all rhos in tree */
			if (dom == i)   /* if domain of beta is destination object */
				{
				make_list(&l,id);
				search_and_add(rhotree,l,&tp);  /* add identity to rhotree */
				free_list(l);
				}
			path[0] = -1;                           /* re-initialize path and objects */
			for (j=0;j<A->numobj;j++)
				{
				objects[j] = 0;
				}
			if(path_len(alpha) == 0)        /* if alpha is identity */
				{
				dom = obj;                              /* domain of alpha is test object */
				}
			else
				{
				dom = A->arr[alpha[path_len(alpha)-1]][0];      /* domain of alpha is domain */
															/* of first arrow in alpha   */
				}
			if (rhotree->down != NULL)      /* if rhotree is not empty */
				{                       /* check for onto */
				if (dom == i)                   /* if domain of alpha is dest. obj. */
					{
					make_list(&l,id);               /* check to see that identity is */
					if(!(search_and_add(T,l,&tp)))  /* in T; if not, then not onto   */
						{
						free_list(l);
						return(false);
						}
					else
						{                           /* if identity is in T */
						free_list(l);
						if(!(equal_trees(rhotree,tp->tree)))  /* check that rhotree equals */
											/* subtree of identity in T  */
							{
							return(false);                                    /* if not, then not onto */
							}
						}
					}                               /* check for onto */
				if (!(check_onto(path,dom,i,objects,T,A,rhotree,maxloop)))
					{
					free_tree(T);
					free_tree(rhotree);
					return(false);                  /* if not onto, then not a sum */
					}
					free_tree(T);
					free_tree(rhotree);
				}
			}
		}
		return(true);
}


/*
Parameter:
	A:      category being tested
Purpose:
	To get user input about sum to be tested.
Design:
	Gets user input and then calls function is_sum to test sum
*/

void sum(category A, int maxloop)
{
char objchr, temp[1];
int obj;
char alphapath[MAXWORD], betapath[MAXWORD];
int alpha[MAXWORD], beta[MAXWORD];
bool done;

	done = false;
	while (!done)
		{
		printf("(-1 to cancel)\n");
		printf("Object of sum: ");
		flush();
		objchr = getchar();
		if (objchr == '-') return;
		obj = object_number(objchr,A);
		done = true;
		if(check_obj(objchr,A))
			{
			printf("\n\007****************ERROR****************\n");
			printf("      %c is not an object of %s. \n",objchr, A.name);
			printf("*************************************\n");
			done = false;
			}
		}
	done = false;
	while (!done)
		{
		done = true;
		printf("Enter alpha: ");
		flush();
		gets(alphapath);
		if (strcmp(alphapath,"-1") == 0) return;
		if (!(check_string(alphapath,A)))
			{
			done = false;
			}
		if (done)
			{
			copy_path(alpha,string_to_path(alphapath,A));
			if (!(check_path(alpha,A)))
				{
				done = false;
				}
			if ((A.arr[alpha[0]][1] != obj) && (done))
				{
				printf("\n\007Alpha must have codomain of %c\n",A.obj[obj]);
				done = false;
				}
			}
		}
		done = false;
		while (!done)
			{
			done = true;
			printf("Enter beta: ");
			flush();
			gets(betapath);
			if (strcmp(betapath,"-1") == 0) return;
			if (!(check_string(betapath,A)))
				{
				done = false;
				}
			if(done)
				{
				copy_path(beta,string_to_path(betapath,A));
				if (!(check_path(beta,A)))
					{
					done = false;
					}
				if ((A.arr[beta[0]][1] != obj) && (done))
					{
					printf("\n\007Beta must have codomain of %c\n",A.obj[obj]);
					done = false;
					}
				}
			}
		printf("\nEnter maximum number of loops (default = %d):  ",maxloop);
		flush();
		gets(temp);
		if (temp[0] != '\0')
			maxloop = atoi(temp);
		if (is_sum(&A,obj,alpha,beta,maxloop))
			{
			printf("\n\n%c is a sum.",A.obj[obj]);
			}
		else
			{
			printf("\n\n%c is not a sum.",A.obj[obj]);
			}
		getchar();
}

