/*TREE.C*/

/*        Categories database        */
/* M.Fleming, R.Gunther, R.Rosebrugh */
/* Mount Allison University 1994-95  */

#include <stdio.h>
#include <stdlib.h>
#include "tree.h"

/*
Parameters:
	None:
Purpose:
	To create the first node of a tree.
Design:
	Makes the first node of a tree with a * as the data.
*/

struct nodetype* create_tree()
{
struct nodetype* temp;

	temp = (struct nodetype*)malloc(sizeof(struct nodetype));
	temp->data = -1;
	temp->down = NULL;
	temp->across = NULL;
	temp->tree = NULL;
	temp->flag = 0;
	temp->info = 0;
	return(temp);
}

/*
Parameters:
	currentelement: a pointer to a node for which we wish to add a child to.
	datatype:       the data we wish to put in the new node.
Purpose:
	To insert a node into a tree with an alphabetical ordering.
Design:
	Make a new node which we insert as the first child of the node pointed to
	by currentelement.  Then sort the children of the node pointed to by
	currentelement in alphabetical order.  Returns a pointer to the inserted
	node.
*/

struct nodetype *add_node(struct nodetype* currentelement, datatype info)
{
struct nodetype *temp, *temp2;

	temp = (struct nodetype*)malloc(sizeof(struct nodetype));       /*allocate memory*/
	temp->flag = (currentelement->down != NULL);
	if (temp->flag)
		{
		temp->across = currentelement->down;    /*link node to other children*/
		}
	else
		{
		temp->across = currentelement;
		}
	currentelement->down = temp;            /*link node to parent*/
	while((temp->flag) && (temp->across->data < info))
		{
		temp->data = temp->across->data;
		temp->down = temp->across->down;        /*sort alphabetically*/
		temp->info = temp->across->info;
		if (temp->across->tree != temp->across)
			{
			temp->tree = temp->across->tree;
			}
		else
			{
			temp->tree = temp;
			}
		if (temp->down != NULL)
			{
			temp2 = temp->down;
			while(temp2->flag)
				{
				temp2 = temp2->across;
				}
			temp2->across = temp;
			}
		temp = temp->across;
		}
	temp->data = info;              /*insert data into new node*/
	temp->tree = NULL;
	temp->info = 0;
	temp->down = NULL;
	return(temp);
}

/*
Parameters:
	nodepointer:    a pointer to a pointer to a node on the level of the tree
					that will be searched
	info:   the data that is to be searched for
Purpose:
	To search a level of a tree for a node containing the data that is being
	searched for.
Design:
	To search the entire level of the tree a pointer to a pointer to the
	first node of the level should be passed in.  If found returns true and
	nodepointer points to a pointer to the node with the data being searched
	for,  otherwise returns false and nodepointer points to a pointer to the
	last node of the level.
*/

bool search_level(struct nodetype **nodepointer, datatype info)
{
	while(((*nodepointer)->flag) && ((*nodepointer)->data != info))
		{
		*nodepointer = (*nodepointer)->across;  /*if data not found and not at*/
		}                                       /*last node of level, move across*/
	if((*nodepointer)->data == info)
		{
		return(true);
		}
	else
		{
		return(false);
		}
}

/*
Parameters:
	tree: the tree to be searched
	list: the list to be searched for
	pointer: returns a pointer to the last node of the list in the tree
Purpose:
	Searches a tree for the list and returns true if it is found and false otherwise
Design:
	Returns true if list found in tree and false otherwise.  Sets what pointer points to
	to point to last node of list in tree if found and sets what pointer points to to NULL
	if not found.
*/

bool search_tree(struct nodetype *tree, listtype list, struct nodetype **pointer)
{
bool finished;
struct nodetype *nodepointer;
struct listnodetype *temppointer;

	finished = false;
	temppointer = list;
	/*if tree is empty, set nodepointer to point to top of tree and do not search
		tree*/

	if ((nodepointer = tree->down) == NULL)
		{
		finished = true;
		nodepointer = tree;
		}

	/*search for list until list is found or some element of list not in tree*/

	while((temppointer != NULL) && (!finished))
		{

		/*      if current element of list not found in proper level of tree
			then exit while loop and have nodepointer point to parent of level
			just searched */

		if(!search_level(&nodepointer,temppointer->data))
			{
			finished = true;
			nodepointer = nodepointer->across;
			}

		/*      if current element of list is in tree and not last element of
			list and at the end of a branch of tree then exit while loop.
			(list not in tree) */

		else
			{
			if((temppointer->next != NULL) && (nodepointer->down == NULL))
				{
				finished = true;
				}

		  /*    if current element of list in tree, and not at end of branch
				of tree then move down one level of tree and get next element
				of list */
			else
				{
				if(temppointer->next != NULL)
					{
					nodepointer = nodepointer->down;
					}
				}
			temppointer = temppointer->next;
			}
		}

	/* if got to end of list in while loop then exit and return nodepointer */

	if(!finished)
		{
		*pointer = nodepointer;
		return(true);
		}

	else
		{
		*pointer = NULL;
		return(false);
		}
}



/*
Parameters:
	tree:   a pointer to the head node of a tree
	list:   a pointer to the first node of a list
Purpose:
	To search a tree and return and determine if list is in the tree,
	if it is not then list is added to the tree.
Design:
	Returns true if list is in the tree and if it is not returns false
	and adds list to the tree.
*/

bool search_and_add(struct nodetype *tree, listtype list, struct nodetype** pointer)
{
bool finished;
struct nodetype *nodepointer;
struct listnodetype *temppointer;

	finished = false;
	temppointer = list;
	/*if tree is empty, set nodepointer to point to top of tree and do not search
		tree*/

	if ((nodepointer = tree->down) == NULL)
		{
		finished = true;
		nodepointer = tree;
		}

	/*search for list until list is found or some element of list not in tree*/

	while((temppointer != NULL) && (!finished))
		{

		/*      if current element of list not found in proper level of tree
			then exit while loop and have nodepointer point to parent of level
			just searched */

		if(!search_level(&nodepointer,temppointer->data))
			{
			finished = true;
			nodepointer = nodepointer->across;
			}

		/*      if current element of list is in tree and not last element of
			list and at the end of a branch of tree then exit while loop.
			(list not in tree) */

		else
			{
			if((temppointer->next != NULL) && (nodepointer->down == NULL))
				{
				finished = true;
				}

		  /*    if current element of list in tree, and not at end of branch
				of tree then move down one level of tree and get next element
				of list */
			else
				{
				if(temppointer->next != NULL)
					{
					nodepointer = nodepointer->down;
					}
				}
			temppointer = temppointer->next;
			}
		}

	/* if got to end of list in while loop then exit and return nodepointer */

	if(!finished)
		{
		*pointer = nodepointer;
		return(true);
		}

	/*      if entire list not in tree then add rest of list to tree and return nodepointer */

	else
		{
		while(temppointer != NULL)
			{
			nodepointer = add_node(nodepointer,temppointer->data);
			temppointer = temppointer->next;
			}
		*pointer = nodepointer;
		return(false);
		}
}


void display_tree(struct nodetype *tree)
{
struct nodetype* temp;

	temp = tree;
	printf("%d\n",temp->data);
	if (temp->tree != NULL)
		{
		display_tree(temp->tree);
		}
	if (temp->down != NULL)
		{
		display_tree(temp->down);
		}
	if (temp->flag)
		{
		display_tree(temp->across);
		}
}

/*
Parameters:
	tree1, tree2 :  the two trees to be tested for equality
Purpose:
	To compare two trees and return whether or not they are equal trees
Design:
	Traverses both trees and compares each node to see if they are equal
*/

bool equal_trees(struct nodetype *tree1, struct nodetype *tree2)

{
struct nodetype* temp1, *temp2;

	temp1 = tree1;
	temp2 = tree2;
	if(temp1->data != temp2->data)
		{
		return(false);
		}
	else
		{
		if((temp1->down == NULL) != (temp2->down == NULL))
			{
			return(false);
			}
		if ((temp1->down != NULL) && (temp2->down != NULL))
			{
			if(!(equal_trees(temp1->down,temp2->down)))
				{
				return(false);
				}
			}
		if((temp1->flag) != (temp2->flag))
			{
			return(false);
			}
		if((temp1->flag) && (temp2->flag))
			{
			if(!(equal_trees(temp1->across,temp2->across)))
				{
				return(false);
				}
			}
		}
	return(true);
}



/*
Parameters:
	list:a pointer to a pointer to the first element of the list being made
	path:an array of integers to be stored in the list
Purpose:
	To create a list of integers
Design:
	The integers are stored in the list in reverse order. The first integer
	of the array is stored at the end of the list.
*/

void    make_list(listtype *list, int path[MAXWORD])
{
struct listnodetype     *temppointer;
int i;

	*list = NULL;
	for(i=0;i<path_len(path);i++)
		{
		temppointer = (struct listnodetype*)malloc(sizeof(struct listnodetype));
		temppointer->next = *list;
		temppointer->data = path[i];
		*list = temppointer;
		}
}

void display_list(const listtype list)
{
struct listnodetype *temppointer;

	temppointer = list;
	while(temppointer != NULL)
		{
		printf("%d ",temppointer->data);
		temppointer = temppointer->next;
		}
}


/*
Parameter:
	tree:   the tree to deallocate memory for
Purpose:
	To deallocate memory which has been allocated for tree
*/

void free_tree(struct nodetype *tree)
{
/*	if ((tree->tree != NULL) && (tree->tree != tree) && (tree->info == 1))
		{
		free_tree(tree->tree);
		}                     */
	if (tree->down != NULL)
		{
		free_tree(tree->down);
		}
	if (tree->flag)
		{
		free_tree(tree->across);
		}
	free(tree);
}


/*
Parameter:
	L:      the list to deallocate memory for
Purpose:
	To deallocate memory which has been allocated for list
*/

void free_list(struct listnodetype *L)
{
	if (L != NULL)
		{
		if(L->next != NULL)
			{
			free_list(L->next);
			}
		}
	free(L);
}
